Require Import MyTactics.
Require Import LabelAlgebra.
Require Import FinSetLattice.
Require Import Principal.
Require Import Taint.
Require Import CFinSetLattice.

Set Implicit Arguments.

Definition ctaint_flowsTo (a s1 s2: CFinSet Prin.t) : Prop :=
  CFinSet_included s1 (CFinSet_union s2 a).

Instance CTaint_flowsTo_PreOrder (a: CFinSet Prin.t) :
  PreOrder (ctaint_flowsTo a).
Proof.
unfold ctaint_flowsTo. constructor.
* intro s. apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
* intros s1 s2 s3 H12 H23.
  rewrite H12.
  apply (@join_lub _ _ _ (CFinSet_JoinPreLattice _)).
  + assumption.
  + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
Qed.

Instance CTaint_flowsTo_HasBottom (a: CFinSet Prin.t):
  HasBottom _ (ctaint_flowsTo a) :=
{ bottom := CFinSet_empty _ }.
Proof.
unfold ctaint_flowsTo. intro s.
apply (@bottom_minimum _ _ (CFinSet_included_HasBottom _)).
Defined.

Instance CTaint_flowsTo_HasTop (a: CFinSet Prin.t):
  HasTop _ (ctaint_flowsTo a) :=
{ top := CFinSet_full _ }.
Proof.
unfold ctaint_flowsTo. intro s.
apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
Defined.

Instance CTaint_JoinPrelattice (a: CFinSet Prin.t) :
  JoinPreLattice (CFinSet Prin.t) (ctaint_flowsTo a) :=
{ join := join }.
Proof.
* intros s1 s2. unfold ctaint_flowsTo.
  transitivity (@join _ CFinSet_included _ _ s1 s2).
  apply join_ub1.
  transitivity (@join _ CFinSet_included _ _ (@join _ CFinSet_included _ _ s1 s2) a);
    try reflexivity.
  apply join_ub1.
* intros s1 s2. unfold ctaint_flowsTo.
  transitivity (@join _ CFinSet_included _ _ s1 s2).
  apply join_ub2.
  transitivity (@join _ CFinSet_included _ _ (@join _ CFinSet_included _ _ s1 s2) a);
    try reflexivity.
  apply join_ub1.
* intros s s1 s2 H1 H2. unfold ctaint_flowsTo in *.
  transitivity (@join _ CFinSet_included _ _ s1 s2); try reflexivity.
  apply join_lub; assumption.
Defined.

Instance CTaint_MeetPrelattice (a: CFinSet Prin.t) :
  MeetPreLattice (CFinSet Prin.t) (ctaint_flowsTo a) :=
{ meet := meet }.
Proof.
* intros s1 s2. unfold ctaint_flowsTo.
  transitivity s1.
  transitivity (@meet _ CFinSet_included _ _ s1 s2). reflexivity. apply meet_lb1.
  transitivity (@join _ CFinSet_included _ _ s1 a).
  apply join_ub1. reflexivity.
* intros s1 s2. unfold ctaint_flowsTo.
  transitivity s2.
  transitivity (@meet _ CFinSet_included _ _ s1 s2). reflexivity. apply meet_lb2.
  transitivity (@join _ CFinSet_included _ _ s2 a).
  apply join_ub1. reflexivity.
* intros s s1 s2 H1 H2. unfold ctaint_flowsTo in *.
  transitivity (@meet _ CFinSet_included _ _
                      (@join _ CFinSet_included _ _ s1 a)
                      (@join _ CFinSet_included _ _ s2 a)).
  apply meet_glb; assumption.
  apply CFinSet_distrib1.
Defined.

Instance CTaint :
  LabelAlgebra (CFinSet Prin.t) (CFinSet Prin.t) :=
{ Ldef := CFinSet_intro (@Ldef _ _ Taint)
; labLeq := CFinSet_included
; authLeq := CFinSet_included
; flowsTo := ctaint_flowsTo
}.
Proof.
* unfold ctaint_flowsTo; split; intros s1 s2 H.
  + rewrite H. apply (@join_lub _ _ _ (CFinSet_JoinPreLattice _)).
    - reflexivity.
    - apply bottom_minimum.
  + rewrite H. apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
* auto.
* auto.
* intros a1 a2 H12 s s' H. unfold ctaint_flowsTo in *.
  rewrite H. apply (@join_lub _ _ _ (CFinSet_JoinPreLattice _)).
  + apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
  + rewrite H12. apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
Defined.
