Require Import PreLattice.

Set Implicit Arguments.

Inductive Dual (A: Type) : Type := Dual_intro : A -> Dual A.

Definition Dual_leq {A: Type} (leq: A -> A -> Prop) (x1 x2: Dual A): Prop :=
match x1, x2 with
| Dual_intro a1, Dual_intro a2 => leq a2 a1
end.

Instance Dual_leq_PreOrder {A: Type} (leq: A -> A -> Prop) {H: PreOrder leq} :
  PreOrder (Dual_leq leq).
Proof.
constructor.
* intros [a]; simpl. reflexivity.
* intros [a1] [a2] [a3]; simpl; auto; intros.
  transitivity a2; trivial.
Qed.

Instance Dual_HasBottom (A: Type) (leq: A -> A -> Prop) {H: HasTop A leq}:
  HasBottom (Dual A) (Dual_leq leq) :=
{ bottom := Dual_intro top }.
Proof.
intros [a]; simpl. apply top_maximum.
Defined.

Instance Dual_HasTop (A: Type) (leq: A -> A -> Prop) {H: HasBottom A leq}:
  HasTop (Dual A) (Dual_leq leq) :=
{ top := Dual_intro bottom }.
Proof.
intros [a]; simpl. apply bottom_minimum.
Defined.

Instance Dual_JoinPreLattice A (leq: A -> A -> Prop)
         {P: PreOrder leq} {M: MeetPreLattice A leq} :
  JoinPreLattice (Dual A) (Dual_leq leq) :=
{ join x1 x2 :=
    match x1, x2 with
      | Dual_intro a1, Dual_intro a2 => Dual_intro (meet a1 a2)
    end
}.
Proof.
* intros [a1] [a2]; simpl. apply meet_lb1.
* intros [a1] [a2]; simpl. apply meet_lb2.
* intros [a] [a1] [a2]; simpl. apply meet_glb.
Defined.

Instance Dual_MeetPreLattice A (leq: A -> A -> Prop)
         {P: PreOrder leq} {J: JoinPreLattice A leq} :
  MeetPreLattice (Dual A) (Dual_leq leq) :=
{ meet x1 x2 :=
    match x1, x2 with
      | Dual_intro a1, Dual_intro a2 => Dual_intro (join a1 a2)
    end
}.
Proof.
* intros [a1] [a2]; simpl. apply join_ub1.
* intros [a1] [a2]; simpl. apply join_ub2.
* intros [a] [a1] [a2]; simpl. apply join_lub.
Defined.
