Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CEndorsement.
Require Export CDistrust.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.
Require Import Constructive_sets.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_CE_CD: ~(CEndorsement ==>[boolean_sem] CDistrust).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists S, Amap m (CFinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (CFinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm1 (CFinSet_empty _)
                    (CFinSet_empty _)
                    (CFinSet_singleton (Principal.Prin.enum 0))).
    simpl in Hm1.
    unfold CTaint.ctaint_flowsTo in Hm1.
    rewrite Heq in Hm1. destruct Hm1 as [_ Hm1].
    assert
      (CFinSet_included (Lmap m (CFinSet_empty Principal.Prin.t))
                        (CFinSet_union
                           (Lmap m (CFinSet_singleton (Principal.Prin.enum 0)))
                           (CFinSet_full Principal.Prin.t))) as H.
    { transitivity (CFinSet_full Principal.Prin.t).
      + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm1 H). clear H.
    assert (In _ (proj1_sig (FinSet_singleton (Principal.Prin.enum 0)))
               (Principal.Prin.enum 0))
      as H.
    { simpl. auto with sets. }
    specialize (Hm1 _ H). clear H. simpl in Hm1.
    apply Union_inv in Hm1. destruct Hm1 as [[] | []].
}
apply (cprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). destruct Hm2 as [_ Hm2].
  simpl in Hm2. unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intros x y H. specialize (Hm1 (CFinSet_empty _) y x).
  rewrite <- HSA0. apply Hm1.
  simpl. unfold flip, CTaint.ctaint_flowsTo.
  rewrite CFinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_boolean_embeds_CE_CD.

Lemma not_eval_embeds_CE_CD: ~(CEndorsement ==>[eval_sem] CDistrust).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm4]]]].
assert (exists S, Amap m (CFinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (CFinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm2 (CFinSet_empty _)
                    (CFinSet_empty _)
                    (CFinSet_empty _)
                    (CFinSet_singleton (Principal.Prin.enum 0))).
    simpl in Hm2.
    unfold CTaint.ctaint_flowsTo, flip in Hm2.
    rewrite CFinSet_union_with_empty_set_included1 in Hm2.
    assert (FinSet_included
              (FinSet_empty Principal.Prin.t)
              (FinSet_empty Principal.Prin.t))
      as H0 by reflexivity.
    specialize (Hm2 H0). clear H0.
    rewrite Heq in Hm2. destruct Hm2 as [_ Hm2].
    assert
      (CFinSet_included (Lmap m (CFinSet_empty Principal.Prin.t))
                        (CFinSet_union
                           (Lmap m (CFinSet_singleton (Principal.Prin.enum 0)))
                           (CFinSet_full Principal.Prin.t))) as H.
    { transitivity (CFinSet_full Principal.Prin.t).
      + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm2 H). clear H.
    assert (In _ (proj1_sig (FinSet_singleton (Principal.Prin.enum 0)))
               (Principal.Prin.enum 0))
      as H.
    { simpl. auto with sets. }
    specialize (Hm2 _ H). clear H. simpl in Hm2.
    apply Union_inv in Hm2. destruct Hm2 as [[] | []].
}
apply (cprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 y (CFinSet_empty _) y x).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip in Hm2.
    rewrite CFinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _) (CFinSet_empty _) y).
    - rewrite <- H.
      apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
  + specialize (Hm2 x (CFinSet_empty _) x y).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip in Hm2.
    rewrite CFinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _) (CFinSet_empty _) x).
    - rewrite H.
      apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
* intros x y H. specialize (Hm2 y (CFinSet_empty _) y x).
  rewrite <- HSA0. apply Hm2.
  + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
  + simpl. unfold flip, CTaint.ctaint_flowsTo.
    rewrite CFinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_eval_embeds_CE_CD.

Lemma not_boolean_embeds_CD_CE: ~(CDistrust ==>[boolean_sem] CEndorsement).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists S, Amap m (CFinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (CFinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm1 (CFinSet_empty _)
                    (CFinSet_singleton (Principal.Prin.enum 0))
                    (CFinSet_empty _)).
    simpl in Hm1.
    unfold flip, CTaint.ctaint_flowsTo in Hm1.
    rewrite Heq in Hm1. destruct Hm1 as [_ Hm1].
    assert
      (CFinSet_included (Lmap m (CFinSet_empty Principal.Prin.t))
                        (CFinSet_union
                           (Lmap m (CFinSet_singleton (Principal.Prin.enum 0)))
                           (CFinSet_full Principal.Prin.t))) as H.
    { transitivity (CFinSet_full Principal.Prin.t).
      + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm1 H). clear H.
    assert (In _ (proj1_sig (FinSet_singleton (Principal.Prin.enum 0)))
               (Principal.Prin.enum 0))
      as H.
    { simpl. auto with sets. }
    specialize (Hm1 _ H). clear H. simpl in Hm1.
    apply Union_inv in Hm1. destruct Hm1 as [[] | []].
}
apply (cprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 y x). destruct Hm2 as [_ Hm2].
  simpl in Hm2. unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intros x y H. specialize (Hm1 (CFinSet_empty _) x y).
  rewrite <- HSA0. apply Hm1.
  simpl. unfold CTaint.ctaint_flowsTo.
  rewrite CFinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_boolean_embeds_CD_CE.

Lemma not_eval_embeds_CD_CE: ~(CDistrust ==>[eval_sem] CEndorsement).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm4]]]].
assert (exists S, Amap m (CFinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (CFinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm2 (CFinSet_full _)
                    (CFinSet_empty _)
                    (CFinSet_singleton (Principal.Prin.enum 0))
                    (CFinSet_empty _) I).
    simpl in Hm2.
    unfold CTaint.ctaint_flowsTo, flip in Hm2.
    rewrite FinSet_union_with_empty_set_included1 in Hm2.
    rewrite Heq in Hm2. destruct Hm2 as [_ Hm2].
    assert
      (CFinSet_included (Lmap m (CFinSet_empty Principal.Prin.t))
                        (CFinSet_union
                           (Lmap m (CFinSet_singleton (Principal.Prin.enum 0)))
                           (CFinSet_full Principal.Prin.t))) as H.
    { transitivity (CFinSet_full Principal.Prin.t).
      + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm2 H). clear H.
    assert (In _ (proj1_sig (FinSet_singleton (Principal.Prin.enum 0)))
               (Principal.Prin.enum 0))
      as H.
    { simpl. auto with sets. }
    specialize (Hm2 _ H). destruct Hm2.
}
apply (cprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 (CFinSet_full _) (CFinSet_empty _) x y).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip in Hm2.
    rewrite CFinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - destruct x; exact I.
    - rewrite <- H.
      apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
  + specialize (Hm2 (CFinSet_full _) (CFinSet_empty _) y x).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip in Hm2.
    rewrite CFinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - destruct y; exact I.
    - rewrite H.
      apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
* intros x y H. specialize (Hm2 (CFinSet_full _) (CFinSet_empty _) x y).
  rewrite <- HSA0. apply Hm2.
  + destruct x; exact I.
  + simpl. unfold CTaint.ctaint_flowsTo.
    rewrite CFinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_eval_embeds_CD_CE.

(** * Summary *)
Open Scope table.

Theorem injects_CD_CE_summary :
  property_square_table injects_in
     [| CDistrust; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CD_CE_summary :
  embed_table boolean_sem
     [| CDistrust; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CD_CE_summary :
  embed_table eval_sem
     [| CDistrust; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
