Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Endorsement.
Require Export CDistrust.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.
Require Import Constructive_sets.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_E_CD: ~(Endorsement ==>[boolean_sem] CDistrust).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists S, Amap m (FinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (FinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm1 (FinSet_empty _)
                    (FinSet_empty _)
                    (FinSet_singleton (Principal.Prin.enum 0))).
    simpl in Hm1. unfold CTaint.ctaint_flowsTo, flip in Hm1.
    rewrite Heq in Hm1. destruct Hm1 as [_ Hm1].
    assert
      (CFinSet_included (Lmap m (FinSet_empty Principal.Prin.t))
                        (CFinSet_union
                           (Lmap m (FinSet_singleton (Principal.Prin.enum 0)))
                           (CFinSet_full Principal.Prin.t))) as H.
    { transitivity (CFinSet_full Principal.Prin.t).
      + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm1 H). clear H.
    assert (In _ (proj1_sig (FinSet_singleton (Principal.Prin.enum 0)))
               (Principal.Prin.enum 0))
      as H.
    { simpl. auto with sets. }
    specialize (Hm1 _ H). clear H. simpl in Hm1.
    apply Union_inv in Hm1. destruct Hm1 as [[] | []].
}
apply (fprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). destruct Hm2 as [_ Hm2].
  simpl in Hm2.
  unfold CTaint.ctaint_flowsTo, flip in Hm2.
  simpl in H. simpl. tauto.
* intros x y H. specialize (Hm1 (FinSet_empty _) y x).
  rewrite <- HSA0. apply Hm1.
  simpl. unfold flip, Taint.taint_flowsTo.
  rewrite FinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_boolean_embeds_E_CD.

Lemma not_eval_embeds_E_CD: ~(Endorsement ==>[eval_sem] CDistrust).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm4]]]].
assert (exists S, Amap m (FinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (FinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm2 (FinSet_empty _)
                    (FinSet_empty _)
                    (FinSet_empty _)
                    (FinSet_singleton (Principal.Prin.enum 0))).
    simpl in Hm2.
    unfold CTaint.ctaint_flowsTo, Taint.taint_flowsTo, flip in Hm2.
    rewrite FinSet_union_with_empty_set_included1 in Hm2.
    assert (FinSet_included
              (FinSet_empty Principal.Prin.t)
              (FinSet_empty Principal.Prin.t))
      as H0 by reflexivity.
    specialize (Hm2 H0). clear H0.
    rewrite Heq in Hm2. destruct Hm2 as [_ Hm2].
    assert
      (CFinSet_included (Lmap m (FinSet_empty Principal.Prin.t))
                        (CFinSet_union
                           (Lmap m (FinSet_singleton (Principal.Prin.enum 0)))
                           (CFinSet_full Principal.Prin.t))) as H.
    { transitivity (CFinSet_full Principal.Prin.t).
      + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm2 H). clear H.
    assert (In _ (proj1_sig (FinSet_singleton (Principal.Prin.enum 0)))
               (Principal.Prin.enum 0))
      as H.
    { simpl. auto with sets. }
    specialize (Hm2 _ H). clear H. simpl in Hm2.
    apply Union_inv in Hm2. destruct Hm2 as [[] | []].
}
apply (fprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 y (FinSet_empty _) y x).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip, Taint.taint_flowsTo in Hm2.
    rewrite FinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
    - rewrite <- H.
      apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
  + specialize (Hm2 x (FinSet_empty _) x y).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip, Taint.taint_flowsTo in Hm2.
    rewrite FinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
    - rewrite H.
      apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
* intros x y H. specialize (Hm2 y (FinSet_empty _) y x).
  rewrite <- HSA0. apply Hm2.
  + apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
  + simpl. unfold flip, Taint.taint_flowsTo.
    rewrite FinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_eval_embeds_E_CD.

Lemma not_boolean_embeds_CD_E: ~(CDistrust ==>[boolean_sem] Endorsement).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (cprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (CFinSet_empty _)) (Amap m (CFinSet_empty _)))
         (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). destruct Hm2 as [_ Hm2].
  simpl in Hm2. unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intro x. specialize (Hm1 (CFinSet_empty _) (CFinSet_empty _) x).
  apply Hm1. simpl.
  destruct x; simpl; auto.
  apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
Qed.
Hint Resolve not_boolean_embeds_CD_E.

Lemma not_eval_embeds_CD_E: ~(CDistrust ==>[eval_sem] Endorsement).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm4]]]].
apply (cprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (CFinSet_full _)) (Amap m (CFinSet_full _)))
         (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 (CFinSet_full _) (CFinSet_empty _) x y).
    assert (CFinSet_included (CFinSet_full _) (CFinSet_union x (CFinSet_full _)))
    as Hincl.
    { destruct x; exact I. }
    specialize (Hm2 Hincl).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip, Taint.taint_flowsTo in Hm2.
    rewrite CFinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2. rewrite <- H.
    apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
  + specialize (Hm2 (CFinSet_full _) (CFinSet_empty _) y x).
    assert (CFinSet_included (CFinSet_full _) (CFinSet_union y (CFinSet_full _)))
    as Hincl.
    { destruct y; exact I. }
    specialize (Hm2 Hincl).
    simpl in Hm2. unfold CTaint.ctaint_flowsTo, flip, Taint.taint_flowsTo in Hm2.
    rewrite CFinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2. rewrite H.
    apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
* intro x.
  specialize (Hm2 (CFinSet_full _) (CFinSet_full _) (CFinSet_full _) x).
  destruct Hm2 as [Hm2 _].
  + destruct x; simpl; auto.
  + apply Hm2. destruct x; simpl; auto.
Qed.
Hint Resolve not_eval_embeds_CD_E.

(** * Summary *)
Open Scope table.

Theorem injects_CD_E_summary :
  property_square_table injects_in
     [| CDistrust; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CD_E_summary :
  embed_table boolean_sem
     [| CDistrust; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CD_E_summary :
  embed_table eval_sem
     [| CDistrust; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
