Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CDistrust.
Require Export UEndorsement.
Require Import EmbedUDUE.
Require Import EmbedCDUD.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma injects_CD_UE: CDistrust ↪ UEndorsement.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve injects_CD_UE.

Lemma not_boolean_embeds_UE_CD: ~(UEndorsement ==>[boolean_sem] CDistrust).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UDistrust)
     (LA2 := UEndorsement)
     (LA3 := CDistrust);
solve_embed.
Qed.
Hint Resolve not_boolean_embeds_UE_CD.

Lemma not_eval_embeds_UE_CD: ~(UEndorsement ==>[eval_sem] CDistrust).
Proof.
apply embeds_in_trans_not1
with (sem := eval_sem)
     (LA1 := UDistrust)
     (LA2 := UEndorsement)
     (LA3 := CDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UE_CD.

(** * Summary *)
Open Scope table.

Theorem injects_CD_UE_summary :
  property_square_table injects_in
     [| CDistrust; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CD_UE_summary :
  embed_table boolean_sem
     [| CDistrust; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CD_UE_summary :
  embed_table eval_sem
     [| CDistrust; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
