Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CEndorsement.
Require Export UEndorsement.
Require Import CFinSetLattice.
Require Import SetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.
Require Import Finite_sets_facts.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma injects_CE_UE: CEndorsement ↪ UEndorsement.
Proof.
pose (f (s: CFinSet Principal.Prin.t) :=
        match s with
          | CFinSet_intro s => proj1_sig s
          | CFinSet_full => Full_set _
        end).
exists {| Lmap := f ; Amap := f |}.
rewrite LAinjection_iff. unfold LAinjection_prop.
splits; try reflexivity.
* intros a1 a2.
  destruct a1; destruct a2; try reflexivity; split; intro; try reflexivity; auto.
  + simpl. auto using Full_intro with sets.
  + destruct f0 as [S HS]. simpl. simpl in H.
    apply Finite_downward_closed in H; auto using Prin_not_finite.
* intros a1 a2. destruct a1; destruct a2; auto. reflexivity.
* intros l1 l2. destruct l1; destruct l2; auto.
* intros l1 l2.
  destruct l1; destruct l2; simpl; unfold flip; auto using Full_intro with sets.
* intros a l1 l2.
  destruct a; destruct l1; destruct l2; simpl;
  unfold flip, UTaint.taint_flowsTo, CTaint.ctaint_flowsTo; simpl;
  split; intro H; auto using Full_intro with sets.
  + destruct f0 as [S0 H0]. destruct f1 as [S1 H1]. simpl in H.
    apply Finite_downward_closed in H;
      auto using Prin_not_finite, Union_preserves_Finite.
Qed.
Hint Resolve injects_CE_UE.

Lemma not_boolean_embeds_UE_CE: ~(UEndorsement ==>[boolean_sem] CEndorsement).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists SA0, Amap m (Empty_set _) = CFinSet_intro SA0) as [SA0 HSA0].
{ case_eq (Amap m (Empty_set _)).
  * intros S Heq. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize
      (Hm1 (Empty_set _) (Empty_set _) (Singleton _ (Principal.Prin.enum 0))).
    destruct Hm1 as [_ Hm1].
    assert (CFinSet_included
              (Lmap m (Singleton Principal.Prin.t (Principal.Prin.enum 0)))
              (CFinSet_union
                 (Lmap m (Empty_set Principal.Prin.t))
                 (Amap m (Empty_set Principal.Prin.t))))
      as H.
    { rewrite Heq. transitivity (CFinSet_full Principal.Prin.t).
      * apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      * apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm1 H (Principal.Prin.enum 0) (In_singleton _ _)).
    apply Union_inv in Hm1. destruct Hm1 as [[] | []].
}
apply (prinset_to_cprinset_inj_quasi_monotone SA0 (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). simpl in *. tauto.
* intros x y H. specialize (Hm1 (Empty_set _) y x). simpl in Hm1.
  unfold flip, CTaint.ctaint_flowsTo, UTaint.taint_flowsTo in Hm1.
  rewrite HSA0 in Hm1. apply Hm1.
  rewrite Union_with_empty_set_Included1.
  assumption.
Qed.
Hint Resolve not_boolean_embeds_UE_CE.

Lemma not_eval_embeds_UE_CE: ~(UEndorsement ==>[eval_sem] CEndorsement).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm5]]]].
assert (exists SA0, Amap m (Empty_set _) = CFinSet_intro SA0) as [SA0 HSA0].
{ case_eq (Amap m (Empty_set _)).
  * intros S HS. exists S. reflexivity.
  * intro H. exfalso.
    specialize (Hm2 (Full_set _) (Empty_set _) (Empty_set _) (Full_set _)).
    rewrite H in Hm2. simpl in Hm2.
    unfold flip, CTaint.ctaint_flowsTo, UTaint.taint_flowsTo in Hm2.
    destruct Hm2 as [_ Hm2].
    + apply (@bottom_minimum _ _ SetHasBottom).
    + assert
        (forall x y,
           CFinSet_included x (CFinSet_union y (CFinSet_full Principal.Prin.t)))
        as Hincl.
      { intros x y. destruct x; destruct y; simpl; auto. }
      specialize (Hm2 (Hincl _ _)).
      specialize (Hm2 (Principal.Prin.enum 0) (Full_intro _ _)).
      apply Union_inv in Hm2. destruct Hm2 as [ [] | [] ].
}
assert (forall x, Included _ x (Union _ (Empty_set _) (Full_set Principal.Prin.t)))
  as Hincl by auto using Full_intro with sets.
apply (prinset_to_cprinset_inj_quasi_monotone SA0 (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 (Full_set _) (Empty_set _) y x (Hincl _)).
    rewrite <- Union_with_empty_set_Included1. apply Hm2.
    simpl. unfold flip, CTaint.ctaint_flowsTo. rewrite H.
    apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
  + specialize (Hm2 (Full_set _) (Empty_set _) x y (Hincl _)).
    rewrite <- Union_with_empty_set_Included1. apply Hm2.
    simpl. unfold flip, CTaint.ctaint_flowsTo. rewrite <- H.
    apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
* intros x y H.
  specialize (Hm2 (Full_set _) (Empty_set _) y x (Hincl _)).
  rewrite HSA0 in Hm2. apply Hm2.
  simpl. unfold flip, UTaint.taint_flowsTo.
  rewrite Union_with_empty_set_Included1. assumption.
Qed.
Hint Resolve not_eval_embeds_UE_CE.

(** * Summary *)
Open Scope table.

Theorem injects_CE_UE_summary :
  property_square_table injects_in
     [| CEndorsement; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_E_CE_summary :
  embed_table boolean_sem
     [| CEndorsement; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_E_CE_summary :
  embed_table eval_sem
     [| CEndorsement; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
