Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CReader.
Require Export CDistrust.
Require Import EmbedCDCE.
Require Import EmbedCRCE.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_CR_CD: ~(CReader ==>[boolean_sem] CDistrust).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := CEndorsement)
     (LA2 := CReader)
     (LA3 := CDistrust);
auto.
Qed.
Hint Resolve not_boolean_embeds_CR_CD.

Lemma not_boolean_embeds_CD_CR: ~(CDistrust ==>[boolean_sem] CReader).
Proof.
apply embeds_in_trans_not2
with (sem := boolean_sem)
     (LA1 := CDistrust)
     (LA2 := CReader)
     (LA3 := CEndorsement);
auto.
Qed.
Hint Resolve not_boolean_embeds_CD_CR.

Lemma not_eval_embeds_CD_CR: ~(CDistrust ==>[eval_sem] CReader).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CDistrust CReader _ _) as H'.
apply H' with (l := bottom) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CD_CR.

Lemma not_eval_embeds_CR_CD: ~(CReader ==>[eval_sem] CDistrust).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CReader CDistrust _ _) as H'.
apply H' with (l := bottom) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CR_CD.

(** * Summary *)
Open Scope table.

Theorem injects_CR_CD_summary :
  property_square_table injects_in
     [| CReader; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CR_CD_summary :
  embed_table boolean_sem
     [| CReader; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CR_CD_summary :
  embed_table eval_sem
     [| CReader; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
