Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CReader.
Require Export CEndorsement.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_CR_CE: CReader ==>[boolean_sem] CEndorsement.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_CR_CE.

Lemma boolean_embeds_CE_CR: CEndorsement ==>[boolean_sem] CReader.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_CE_CR.

Lemma not_eval_embeds_CR_CE: ~(CReader ==>[eval_sem] CEndorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CReader CEndorsement _ _) as H'.
apply H' with (l := bottom) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CR_CE.

Lemma not_eval_embeds_CE_CR: ~(CEndorsement ==>[eval_sem] CReader).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CEndorsement CReader _ _) as H'.
apply H' with (l := top) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CE_CR.

(** * Summary *)
Open Scope table.

Theorem injects_CR_CE_summary :
  property_square_table injects_in
     [| CReader; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CR_CE_summary :
  embed_table boolean_sem
     [| CReader; CEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CR_CE_summary :
  embed_table eval_sem
     [| CReader; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
