Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Endorsement.
Require Export CReader.
Require Import FinSetLattice.
Require Import EmbedECE.
Require Import EmbedCRCE.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_E_CR: Endorsement ==>[boolean_sem] CReader.
Proof.
transitivity_embeds CEndorsement; solve_embed.
Qed.
Hint Resolve boolean_embeds_E_CR.

Lemma not_boolean_embeds_CR_E: ~(CReader ==>[boolean_sem] Endorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := CEndorsement)
     (LA2 := CReader)
     (LA3 := Endorsement);
auto.
Qed.
Hint Resolve not_boolean_embeds_CR_E.

Lemma not_eval_embeds_E_CR: ~(Endorsement ==>[eval_sem] CReader).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ Endorsement CReader _ _) as H'.
apply H' with (l := FinSet_singleton (Principal.Prin.enum 0)) in H.
* simpl in H. unfold flip in H. destruct H as [_ H].
  assert (In _
             (Add Principal.Prin.t (Empty_set Principal.Prin.t)
                  (Principal.Prin.enum 0))
             (Principal.Prin.enum 0)) as HIn by auto with sets.
  specialize (H _ HIn). destruct H.
* reflexivity.
* reflexivity.
Qed.
Hint Resolve not_eval_embeds_E_CR.

Lemma not_eval_embeds_CR_E: ~(CReader ==>[eval_sem] Endorsement).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := CReader)
     (LA2 := Endorsement)
     (LA3 := CEndorsement);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_CR_E.

(** * Summary *)
Open Scope table.

Theorem injects_CR_E_summary :
  property_square_table injects_in
     [| CReader; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_E_CR_summary :
  embed_table boolean_sem
     [| Endorsement; CReader |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_E_CR_summary :
  embed_table eval_sem
     [| Endorsement; CReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
