Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CReader.
Require Export UDistrust.
Require Import EmbedURUD.
Require Import EmbedCRUR.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_CR_UD: CReader ==>[boolean_sem] UDistrust.
Proof.
transitivity_embeds UReader; solve_embed.
Qed.
Hint Resolve boolean_embeds_CR_UD.

Lemma not_boolean_embeds_UD_CR: ~(UDistrust ==>[boolean_sem] CReader).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UReader)
     (LA2 := UDistrust)
     (LA3 := CReader);
auto.
Qed.
Hint Resolve not_boolean_embeds_UD_CR.

Lemma not_eval_embeds_UD_CR: ~(UDistrust ==>[eval_sem] CReader).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UDistrust)
     (LA2 := CReader)
     (LA3 := UReader);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UD_CR.

Lemma not_eval_embeds_CR_UD: ~(CReader ==>[eval_sem] UDistrust).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CReader UDistrust _ _) as H'.
apply H' with (l := bottom) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CR_UD.

(** * Summary *)
Open Scope table.

Theorem injects_CR_UD_summary :
  property_square_table injects_in
     [| CReader; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CR_UD_summary :
  embed_table boolean_sem
     [| CReader; UDistrust |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CR_UD_summary :
  embed_table eval_sem
     [| CReader; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
