Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CReader.
Require Export UEndorsement.
Require Import EmbedURUE.
Require Import EmbedCRUR.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_CR_UE: CReader ==>[boolean_sem] UEndorsement.
Proof.
transitivity_embeds UReader; solve_embed.
Qed.
Hint Resolve boolean_embeds_CR_UE.

Lemma not_boolean_embeds_UE_CR: ~(UEndorsement ==>[boolean_sem] CReader).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UReader)
     (LA2 := UEndorsement)
     (LA3 := CReader);
auto.
Qed.
Hint Resolve not_boolean_embeds_UE_CR.

Lemma not_eval_embeds_CR_UE: ~(CReader ==>[eval_sem] UEndorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CReader UEndorsement _ _) as H'.
apply H' with (l := bottom) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CR_UE.

Lemma not_eval_embeds_UE_CR: ~(UEndorsement ==>[eval_sem] CReader).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UEndorsement)
     (LA2 := CReader)
     (LA3 := UReader);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UE_CR.

(** * Summary *)
Open Scope table.

Theorem injects_CR_UE_summary :
  property_square_table injects_in
     [| CReader; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CR_UE_summary :
  embed_table boolean_sem
     [| CReader; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CR_UE_summary :
  embed_table eval_sem
     [| CReader; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
