Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CReader.
Require Export UReader.
Require Import SetLattice.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.
Require Import Finite_sets_facts.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma injects_CR_UR: CReader ↪ UReader.
Proof.
pose (f (s: CFinSet Principal.Prin.t) :=
        match s with
          | CFinSet_intro s => proj1_sig s
          | CFinSet_full => Full_set _
        end).
exists {| Lmap := f ; Amap := f |}.
rewrite LAinjection_iff. unfold LAinjection_prop.
splits; try reflexivity.
* intros a1 a2.
  destruct a1; destruct a2; try reflexivity; split; intro; try reflexivity; auto.
  + simpl. auto using Full_intro with sets.
  + destruct f0 as [S HS]. simpl. simpl in H.
    apply Finite_downward_closed in H; auto using Prin_not_finite.
* intros a1 a2. destruct a1; destruct a2; auto. reflexivity.
* intros l1 l2. destruct l1; destruct l2; auto.
* intros l1 l2.
  destruct l1; destruct l2; simpl; unfold flip; auto using Full_intro with sets.
* intros a l1 l2.
  destruct a; destruct l1; destruct l2; simpl;
  unfold flip, UTaint.taint_flowsTo, CTaint.ctaint_flowsTo; simpl;
  split; intro H; auto using Full_intro with sets.
  + destruct f0 as [S0 H0]. destruct f1 as [S1 H1]. simpl in H.
    apply Finite_downward_closed in H;
      auto using Prin_not_finite, Union_preserves_Finite.
Qed.
Hint Resolve injects_CR_UR.

Lemma not_boolean_embeds_UR_CR: ~(UReader ==>[boolean_sem] CReader).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists SA0, Amap m (Empty_set _) = CFinSet_intro SA0) as [SA0 HSA0].
{ case_eq (Amap m (Empty_set _)).
  * intros S Heq. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize
      (Hm1 (Empty_set _) (Empty_set _) (Singleton _ (Principal.Prin.enum 0))).
    destruct Hm1 as [_ Hm1].
    assert (CFinSet_included
              (Lmap m (Singleton Principal.Prin.t (Principal.Prin.enum 0)))
              (CFinSet_union (Lmap m (Empty_set Principal.Prin.t))
                 (Amap m (Empty_set Principal.Prin.t)))) as H.
    { rewrite Heq. transitivity (CFinSet_full Principal.Prin.t).
      * apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      * apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm1 H (Principal.Prin.enum 0) (In_singleton _ _)).
    apply Union_inv in Hm1. destruct Hm1 as [[] | []].
}
apply (prinset_to_cprinset_inj_quasi_monotone SA0 (fun x => Lmap m x)).
* intros x y Heq. specialize (Hm2 x y). simpl in *. unfold flip in *. tauto.
* intros x y H.
  specialize (Hm1 (Empty_set _) y x). simpl in Hm1.
  unfold flip, UTaint.taint_flowsTo, CTaint.ctaint_flowsTo in Hm1.
  rewrite Union_with_empty_set_Included1 in Hm1.
  rewrite HSA0 in Hm1.
  apply Hm1. assumption.
Qed.
Hint Resolve not_boolean_embeds_UR_CR.

(** * Summary *)
Open Scope table.

Theorem injects_CR_UR_summary :
  property_square_table injects_in
     [| CReader; UReader |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CR_UR_summary :
  embed_table boolean_sem
     [| CReader; UReader |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CR_UR_summary :
  embed_table eval_sem
     [| CReader; UReader |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
