Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CTaint.
Require Export CDistrust.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_CT_CD: CTaint ==>[boolean_sem] CDistrust.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_CT_CD.

Lemma boolean_embeds_CD_CT: CDistrust ==>[boolean_sem] CTaint.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_CD_CT.

Lemma not_eval_embeds_CT_CD: ~(CTaint ==>[eval_sem] CDistrust).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CTaint CDistrust _ _) as H'.
apply H' with (l := top) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CT_CD.

Lemma not_eval_embeds_CD_CT: ~(CDistrust ==>[eval_sem] CTaint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CDistrust CTaint _ _) as H'.
apply H' with (l := bottom) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CD_CT.

(** * Summary *)
Open Scope table.

Theorem injects_CT_CD_summary :
  property_square_table injects_in
     [| CTaint; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_CD_summary :
  embed_table boolean_sem
     [| CTaint; CDistrust |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_CD_summary :
  embed_table eval_sem
     [| CTaint; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
