Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CTaint.
Require Export CEndorsement.
Require Import EmbedCTCD.
Require Import EmbedCDCE.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_CT_CE: ~(CTaint ==>[boolean_sem] CEndorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := CDistrust)
     (LA2 := CTaint)
     (LA3 := CEndorsement);
auto.
Qed.
Hint Resolve not_boolean_embeds_CT_CE.

Lemma not_boolean_embeds_CE_CT: ~(CEndorsement ==>[boolean_sem] CTaint).
Proof.
apply embeds_in_trans_not2
with (sem := boolean_sem)
     (LA1 := CEndorsement)
     (LA2 := CTaint)
     (LA3 := CDistrust);
auto.
Qed.
Hint Resolve not_boolean_embeds_CE_CT.

Lemma not_eval_embeds_CE_CT: ~(CEndorsement ==>[eval_sem] CTaint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CEndorsement CTaint _ _) as H'.
apply H' with (l := top) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CE_CT.

Lemma not_eval_embeds_CT_CE: ~(CTaint ==>[eval_sem] CEndorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CTaint CEndorsement _ _) as H'.
apply H' with (l := top) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CT_CE.

(** * Summary *)
Open Scope table.

Theorem injects_CT_CE_summary :
  property_square_table injects_in
     [| CTaint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_CE_summary :
  embed_table boolean_sem
     [| CTaint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_CE_summary :
  embed_table eval_sem
     [| CTaint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
