Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CTaint.
Require Export CReader.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_CT_CR: ~(CTaint ==>[boolean_sem] CReader).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists S, Amap m (CFinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (CFinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm1 (CFinSet_empty _) (CFinSet_full _) (CFinSet_empty _)).
    simpl in Hm1. apply Hm1. rewrite Heq.
    unfold flip, ctaint_flowsTo.
    transitivity (CFinSet_full Principal.Prin.t).
    + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
    + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
}
apply (cprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). destruct Hm2 as [_ Hm2].
  simpl in Hm2. unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intros x y H. specialize (Hm1 (CFinSet_empty _) x y).
  rewrite <- HSA0. apply Hm1.
  simpl. unfold ctaint_flowsTo.
  rewrite CFinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_boolean_embeds_CT_CR.

Lemma not_boolean_embeds_CR_CT: ~(CReader ==>[boolean_sem] CTaint).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists S, Amap m (CFinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (CFinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm1 (CFinSet_empty _) (CFinSet_empty _) (CFinSet_full _)).
    simpl in Hm1. apply Hm1. rewrite Heq.
    unfold ctaint_flowsTo.
    transitivity (CFinSet_full Principal.Prin.t).
    + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
    + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
}
apply (cprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). destruct Hm2 as [_ Hm2].
  simpl in Hm2. unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intros x y H. specialize (Hm1 (CFinSet_empty _) y x).
  rewrite <- HSA0. apply Hm1.
  simpl. unfold flip, ctaint_flowsTo.
  rewrite CFinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_boolean_embeds_CR_CT.

(** * Summary *)
Open Scope table.

Theorem injects_CT_CR_summary :
  property_square_table injects_in
     [| CTaint; CReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_CR_summary :
  embed_table boolean_sem
     [| CTaint; CReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_CR_summary :
  embed_table eval_sem
     [| CTaint; CReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
