Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CTaint.
Require Export Endorsement.
Require Import EmbedCTCD.
Require Import EmbedCDE.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import PrincipalFacts.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_CT_E: ~(CTaint ==>[boolean_sem] Endorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := CDistrust)
     (LA2 := CTaint)
     (LA3 := Endorsement);
auto.
Qed.
Hint Resolve not_boolean_embeds_CT_E.

Lemma not_boolean_embeds_E_CT: ~(Endorsement ==>[boolean_sem] CTaint).
Proof.
apply embeds_in_trans_not2
with (sem := boolean_sem)
     (LA1 := Endorsement)
     (LA2 := CTaint)
     (LA3 := CDistrust);
auto.
Qed.
Hint Resolve not_boolean_embeds_E_CT.

Lemma not_eval_embeds_E_CT: ~(Endorsement ==>[eval_sem] CTaint).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm4]]]].
assert (exists SA0, Amap m (FinSet_empty _) = CFinSet_intro SA0) as [SA0 HSA0].
{ case_eq (Amap m (FinSet_empty _)).
  * intros SA0 H. exists SA0. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm2 (FinSet_empty _) (FinSet_empty _)
                    (FinSet_empty _) (FinSet_singleton (Principal.Prin.enum 0))).
    simpl in Hm2.
    unfold flip, Taint.taint_flowsTo, ctaint_flowsTo in Hm2.
    repeat rewrite FinSet_union_with_empty_set_included1 in Hm2.
    rewrite Heq in Hm2.
    destruct Hm2 as [_ Hm2].
    + reflexivity.
    + assert
        (CFinSet_included
           (Lmap m (FinSet_empty _))
           (CFinSet_union (Lmap m (FinSet_singleton (Principal.Prin.enum 0)))
                          (CFinSet_full _))) as H.
      { transitivity (CFinSet_full Principal.Prin.t).
        * apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
        * apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
      }
      specialize (Hm2 H). clear H.
      unfold FinSet_singleton, FinSet_included in Hm2. simpl in Hm2.
      assert (In _ (Add _ (Empty_set _) (Principal.Prin.enum 0))
                 (Principal.Prin.enum 0))
        as H by auto with sets.
      specialize (Hm2 _ H). clear H.
      destruct Hm2.
}
apply (fprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 y (FinSet_empty _) y x).
    simpl in Hm2. unfold flip, Taint.taint_flowsTo, ctaint_flowsTo in Hm2.
    repeat rewrite FinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
    - rewrite <- H. apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
  + specialize (Hm2 x (FinSet_empty _) x y).
    simpl in Hm2. unfold flip, Taint.taint_flowsTo, ctaint_flowsTo in Hm2.
    repeat rewrite FinSet_union_with_empty_set_included1 in Hm2.
    apply Hm2.
    - apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
    - rewrite H. apply (@join_ub1 _ _ _ (CFinSet_JoinPreLattice _)).
* intros x y H.
  specialize (Hm2 y (FinSet_empty _) y x). rewrite HSA0 in Hm2.
  apply Hm2.
  + apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
  + simpl. unfold flip, Taint.taint_flowsTo.
    rewrite FinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_eval_embeds_E_CT.

Lemma not_eval_embeds_CT_E: ~(CTaint ==>[eval_sem] Endorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CTaint Endorsement _ _) as H'.
apply H' with (l := top) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CT_E.

(** * Summary *)
Open Scope table.

Theorem injects_CT_E_summary :
  property_square_table injects_in
     [| CTaint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_E_summary :
  embed_table boolean_sem
     [| CTaint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_E_summary :
  embed_table eval_sem
     [| CTaint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
