Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CTaint.
Require Export UDistrust.
Require Import EmbedUTUD.
Require Import EmbedCTUT.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_CT_UD: CTaint ==>[boolean_sem] UDistrust.
Proof.
transitivity_embeds UTaint; solve_embed.
Qed.
Hint Resolve boolean_embeds_CT_UD.

Lemma not_boolean_embeds_UD_CT: ~(UDistrust ==>[boolean_sem] CTaint).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UTaint)
     (LA2 := UDistrust)
     (LA3 := CTaint);
auto.
Qed.
Hint Resolve not_boolean_embeds_UD_CT.

Lemma not_eval_embeds_CT_UD: ~(CTaint ==>[eval_sem] UDistrust).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CTaint UDistrust _ _) as H'.
apply H' with (l := top) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CT_UD.

Lemma not_eval_embeds_UD_CT: ~(UDistrust ==>[eval_sem] CTaint).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UDistrust)
     (LA2 := CTaint)
     (LA3 := UTaint);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UD_CT.

(** * Summary *)
Open Scope table.

Theorem injects_CT_UD_summary :
  property_square_table injects_in
     [| CTaint; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_UD_summary :
  embed_table boolean_sem
     [| CTaint; UDistrust |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_UD_summary :
  embed_table eval_sem
     [| CTaint; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
