Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CTaint.
Require Export UEndorsement.
Require Import EmbedURUE.
Require Import EmbedCTUR.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_CT_UE: CTaint ==>[boolean_sem] UEndorsement.
Proof.
transitivity_embeds UReader; solve_embed.
Qed.
Hint Resolve boolean_embeds_CT_UE.

Lemma not_boolean_embeds_UE_CT: ~(UEndorsement ==>[boolean_sem] CTaint).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UReader)
     (LA2 := UEndorsement)
     (LA3 := CTaint);
solve_embed.
Qed.
Hint Resolve not_boolean_embeds_UE_CT.

Lemma not_eval_embeds_UE_CT: ~(UEndorsement ==>[eval_sem] CTaint).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UEndorsement)
     (LA2 := CTaint)
     (LA3 := UReader);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UE_CT.

Lemma not_eval_embeds_CT_UE: ~(CTaint ==>[eval_sem] UEndorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ CTaint UEndorsement _ _) as H'.
apply H' with (l := top) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CT_UE.

(** * Summary *)
Open Scope table.

Theorem injects_CT_UE_summary :
  property_square_table injects_in
     [| CTaint; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_UE_summary :
  embed_table boolean_sem
     [| CTaint; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_UE_summary :
  embed_table eval_sem
     [| CTaint; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
