Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Endorsement.
Require Export CEndorsement.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import FinSetLattice.
Require Import PrincipalFacts.

(** * Intermediate lemmas. *)
Lemma injects_E_CE: Endorsement ↪ CEndorsement.
Proof.
exists {| Lmap l := CFinSet_intro l ; Amap a := CFinSet_intro a |}.
rewrite LAinjection_iff. unfold LAinjection_prop.
splits; reflexivity.
Qed.
Hint Resolve injects_E_CE.

Lemma not_boolean_embeds_CE_E: ~(CEndorsement ==>[boolean_sem] Endorsement).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (cprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (CFinSet_full _)) (Amap m (CFinSet_full _)))
         (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). simpl in *. tauto.
* intro x.
  specialize (Hm1 (CFinSet_full _) (CFinSet_full _) x).
  destruct Hm1 as [Hm1 _].
  apply Hm1. destruct x; exact I.
Qed.
Hint Resolve not_boolean_embeds_CE_E.

Lemma not_eval_embeds_CE_E: ~(CEndorsement ==>[eval_sem] Endorsement).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm4]]]].
assert (forall x, CFinSet_included x (CFinSet_full Principal.Prin.t)) as Hincl.
{ intro z. simpl. destruct z; simpl; auto. }
apply (cprinset_to_fprinset_inj_bounded_absurd
         (join Ldef (Amap m (CFinSet_full _)))
         (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 (CFinSet_full _) (CFinSet_empty _) y x (Hincl _)).
    apply (@flowsToBottom _ _ CTaint.CTaint). apply Hm2.
    simpl. unfold flip. unfold Taint.taint_flowsTo. rewrite H.
    apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
  + specialize (Hm2 (CFinSet_full _) (CFinSet_empty _) x y (Hincl _)).
    apply (@flowsToBottom _ _ CTaint.CTaint). apply Hm2.
    simpl. unfold flip. unfold Taint.taint_flowsTo. rewrite <- H.
    apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
* intro x.
  simpl in Hm1. unfold flip in Hm1. destruct Hm1 as [Hm1 Hm1'].
  specialize (Hm2 (CFinSet_full _) (CFinSet_full _) Ldef x).
  destruct Hm2 as [Hm2 _].
  + simpl. unfold flip, CTaint.ctaint_flowsTo.
    transitivity (CFinSet_full Principal.Prin.t); auto.
  + simpl in Hm2. unfold flip, CTaint.ctaint_flowsTo, Taint.taint_flowsTo in Hm2.
    rewrite Hm2; auto.
    simpl. apply (@join_lub _ _ _ (FinSet_JoinPreLattice _)).
    - rewrite Hm1'. apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
    - apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
  Grab Existential Variables.
  exact (FinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CE_E.

(** * Summary *)
Open Scope table.

Theorem injects_E_CE_summary :
  property_square_table injects_in
     [| Endorsement; CEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_E_CE_summary :
  embed_table boolean_sem
     [| Endorsement; CEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_E_CE_summary :
  embed_table eval_sem
     [| Endorsement; CEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
