Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Endorsement.
Require Export UEndorsement.
Require Import EmbedECE.
Require Import EmbedCEUE.
Require Import FinSetLattice.
Require Import SetLattice.
Require Import EmbedTableCheck.
Require Import FinSetLattice.
Require Import PrincipalFacts.
Require Import Ensembles.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma injects_E_UE: Endorsement ↪ UEndorsement.
Proof.
transitivity_embeds CEndorsement; auto.
Qed.
Hint Resolve injects_E_UE.

Lemma not_boolean_embeds_UE_E: ~(UEndorsement ==>[boolean_sem] Endorsement).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (prinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (Empty_set _)) (Amap m (Full_set _)))
         (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). simpl in *. tauto.
* intro x.
  specialize (Hm1 (Full_set _) (Empty_set _) x).
  destruct Hm1 as [Hm1 _]. apply Hm1.
  simpl. unfold flip, UTaint.taint_flowsTo.
  auto using Full_intro with sets.
Qed.
Hint Resolve not_boolean_embeds_UE_E.

Lemma not_eval_embeds_UE_E: ~(UEndorsement ==>[eval_sem] Endorsement).
Proof.
rewrite eval_embeds_iff.
intros [m [Hm1 [Hm2 [Hm3 Hm4]]]].
assert (forall x,
          Included _ x (Union _ (Empty_set _) (Full_set Principal.Prin.t)))
  as Hincl by auto using Full_intro with sets.
apply (prinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (FinSet_empty _) (Amap m (Full_set _)))
         (fun x => Lmap m x)).
* intros x y H. split.
  + specialize (Hm2 (Full_set _) (Empty_set _) y x (Hincl _)).
    rewrite <- Union_with_empty_set_Included1. apply Hm2.
    simpl. unfold flip, Taint.taint_flowsTo. rewrite H.
    apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
  + specialize (Hm2 (Full_set _) (Empty_set _) x y (Hincl _)).
    rewrite <- Union_with_empty_set_Included1. apply Hm2.
    simpl. unfold flip, Taint.taint_flowsTo. rewrite <- H.
    apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
* intro x.
  simpl in Hm1. unfold flip in Hm1. destruct Hm1 as [Hm1 Hm1'].
  specialize (Hm2 (Full_set _) (Full_set _) (Empty_set _) x).
  destruct Hm2 as [Hm2 _].
  + simpl. unfold flip, UTaint.taint_flowsTo.
    transitivity (Full_set Principal.Prin.t); auto.
    apply (@top_maximum _ _ SetHasTop).
  + simpl in Hm2. unfold flip, UTaint.taint_flowsTo, Taint.taint_flowsTo in Hm2.
    rewrite Hm2; auto. apply (@join_lub _ _ _ (FinSet_JoinPreLattice _)).
    - rewrite Hm1'. apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
    - apply (@join_ub2 _ _ _ (FinSet_JoinPreLattice _)).
Qed.
Hint Resolve not_eval_embeds_UE_E.

(** * Summary *)
Open Scope table.

Theorem injects_E_UE_summary :
  property_square_table injects_in
     [| Endorsement; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_E_CE_summary :
  embed_table boolean_sem
     [| Endorsement; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_E_CE_summary :
  embed_table eval_sem
     [| Endorsement; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
