(** val fst : ('a1*'a2) -> 'a1 **)

let fst = function
| x,y -> x

(** val snd : ('a1*'a2) -> 'a2 **)

let snd = function
| x,y -> y

type string =
| EmptyString
| String of char * string

(** val append : string -> string -> string **)

let rec append s1 s2 =
  match s1 with
  | EmptyString -> s2
  | String (c, s1') -> String (c, (append s1' s2))

(** val combine : ('a1 -> 'a2 -> 'a3) -> 'a1 list -> 'a2 list -> 'a3 list **)

let rec combine f la lb =
  match la with
  | [] -> []
  | a::la0 ->
    (match lb with
     | [] -> []
     | b::lb0 -> (f a b)::(combine f la0 lb0))

type answer =
| Yes
| No
| Unknown

module ExportTo = 
 struct 
  (** val print_string : string -> unit **)
  
  let print_string = let rec print_string = function EmptyString -> print_newline () | String (c, s) -> begin print_char c; print_string s end in print_string
  
  (** val list_to_string :
      string -> string -> string -> ('a1 -> string) -> 'a1 list -> string **)
  
  let list_to_string open0 del close f l =
    let loop =
      let rec loop = function
      | [] -> EmptyString
      | a::l1 ->
        (match l1 with
         | [] -> f a
         | a0::l2 -> append (f a) (append del (loop l1)))
      in loop
    in
    append open0 (append (loop l) close)
  
  (** val table_to_string :
      string -> string -> string -> ('a1 -> string) -> 'a1 list -> string ->
      string -> string -> ('a2 -> string) -> 'a2 list -> string -> string ->
      string -> ('a3 -> string) -> 'a3 list list -> string **)
  
  let table_to_string ropen rdel rclose rf rows copen cdel cclose cf cols topen tdel tclose tf table =
    list_to_string (append topen (list_to_string copen cdel cclose cf cols))
      tdel tclose (fun r ->
      list_to_string (append ropen (rf (fst r))) rdel rclose tf (snd r))
      (combine (fun x x0 -> x,x0) rows table)
  
  module HTML = 
   struct 
    (** val export_answer : answer -> string **)
    
    let export_answer = function
    | Yes ->
      String ('\226', (String ('\156', (String ('\147', EmptyString)))))
    | No ->
      String ('\226', (String ('\156', (String ('\151', EmptyString)))))
    | Unknown -> String ('?', EmptyString)
    
    (** val export :
        ('a1 -> string) -> ('a2 -> string) -> 'a1 list -> 'a2 list -> answer
        list list -> string **)
    
    let export rf cf rows cols table =
      table_to_string (String ('<', (String ('t', (String ('r', (String ('>',
        EmptyString)))))))) EmptyString (String ('<', (String ('/', (String
        ('t', (String ('r', (String ('>', EmptyString)))))))))) (fun r ->
        append (String ('<', (String ('t', (String ('h', (String ('>',
          EmptyString))))))))
          (append (rf r) (String ('<', (String ('/', (String ('t', (String
            ('h', (String ('>', EmptyString)))))))))))) rows (String ('<',
        (String ('t', (String ('h', (String ('e', (String ('a', (String ('d',
        (String ('>', (String ('<', (String ('t', (String ('r', (String ('>',
        (String ('<', (String ('t', (String ('d', (String ('>', (String ('<',
        (String ('/', (String ('t', (String ('d', (String ('>',
        EmptyString)))))))))))))))))))))))))))))))))))))))) EmptyString
        (String ('<', (String ('/', (String ('t', (String ('r', (String ('>',
        (String ('<', (String ('/', (String ('t', (String ('h', (String ('e',
        (String ('a', (String ('d', (String ('>', (String ('<', (String ('t',
        (String ('b', (String ('o', (String ('d', (String ('y', (String ('>',
        EmptyString)))))))))))))))))))))))))))))))))))))))) (fun c ->
        append (String ('<', (String ('t', (String ('h', (String ('>',
          EmptyString))))))))
          (append (cf c) (String ('<', (String ('/', (String ('t', (String
            ('h', (String ('>', EmptyString)))))))))))) cols (String ('<',
        (String ('t', (String ('a', (String ('b', (String ('l', (String ('e',
        (String ('>', EmptyString)))))))))))))) EmptyString (String ('<',
        (String ('/', (String ('t', (String ('b', (String ('o', (String ('d',
        (String ('y', (String ('>', (String ('<', (String ('/', (String ('t',
        (String ('a', (String ('b', (String ('l', (String ('e', (String ('>',
        EmptyString)))))))))))))))))))))))))))))))) (fun a ->
        append (String ('<', (String ('t', (String ('d', (String ('>',
          EmptyString))))))))
          (append (export_answer a) (String ('<', (String ('/', (String ('t',
            (String ('d', (String ('>', EmptyString)))))))))))) table
    
    (** val export_simple :
        string list -> string list -> answer list list -> string **)
    
    let export_simple =
      export (fun s -> s) (fun s -> s)
    
    (** val print :
        ('a1 -> string) -> ('a2 -> string) -> 'a1 list -> 'a2 list -> answer
        list list -> unit **)
    
    let print rf cf rows cols table =
      print_string (export rf cf rows cols table)
    
    (** val print_simple :
        string list -> string list -> answer list list -> unit **)
    
    let print_simple =
      print (fun s -> s) (fun s -> s)
   end
  
  module TeX = 
   struct 
    (** val newline : string **)
    
    let newline =
      String ('\n', EmptyString)
    
    (** val export_answer : answer -> string **)
    
    let export_answer = function
    | Yes ->
      String ('\\', (String ('d', (String ('i', (String ('n', (String ('g',
        (String ('{', (String ('5', (String ('1', (String ('}',
        EmptyString)))))))))))))))))
    | No ->
      String ('\\', (String ('d', (String ('i', (String ('n', (String ('g',
        (String ('{', (String ('5', (String ('5', (String ('}',
        EmptyString)))))))))))))))))
    | Unknown -> String ('?', EmptyString)
    
    (** val export :
        ('a1 -> string) -> ('a2 -> string) -> 'a1 list -> 'a2 list -> answer
        list list -> string **)
    
    let export rf cf rows cols table =
      let format =
        list_to_string (String ('{', (String ('c', (String ('|',
          EmptyString)))))) EmptyString
          (append (String ('}', EmptyString)) newline) (fun x -> String ('c',
          EmptyString)) cols
      in
      table_to_string EmptyString (String (' ', (String ('&', (String (' ',
        EmptyString))))))
        (append (String (' ', (String ('\\', (String ('\\', EmptyString))))))
          newline) (fun r ->
        append (rf r) (String (' ', (String ('&', (String (' ',
          EmptyString))))))) rows (String (' ', (String ('&', (String (' ',
        EmptyString)))))) (String (' ', (String ('&', (String (' ',
        EmptyString))))))
        (append (String (' ', (String ('\\', (String ('\\', (String (' ',
          (String ('\\', (String ('h', (String ('l', (String ('i', (String
          ('n', (String ('e', EmptyString)))))))))))))))))))) newline) cf
        cols
        (append (String ('\\', (String ('b', (String ('e', (String ('g',
          (String ('i', (String ('n', (String ('{', (String ('t', (String
          ('a', (String ('b', (String ('u', (String ('l', (String ('a',
          (String ('r', (String ('}',
          EmptyString)))))))))))))))))))))))))))))) format) EmptyString
        (String ('\\', (String ('e', (String ('n', (String ('d', (String
        ('{', (String ('t', (String ('a', (String ('b', (String ('u', (String
        ('l', (String ('a', (String ('r', (String ('}',
        EmptyString)))))))))))))))))))))))))) export_answer table
    
    (** val export_simple :
        string list -> string list -> answer list list -> string **)
    
    let export_simple =
      export (fun s -> s) (fun s -> s)
    
    (** val print :
        ('a1 -> string) -> ('a2 -> string) -> 'a1 list -> 'a2 list -> answer
        list list -> unit **)
    
    let print rf cf rows cols table =
      print_string (export rf cf rows cols table)
    
    (** val print_simple :
        string list -> string list -> answer list list -> unit **)
    
    let print_simple =
      print (fun s -> s) (fun s -> s)
   end
 end

(** val setLAnames : string list **)

let setLAnames =
  (String ('T', EmptyString))::((String ('C', (String ('T',
    EmptyString))))::((String ('U', (String ('T', EmptyString))))::((String
    ('E', EmptyString))::((String ('C', (String ('E',
    EmptyString))))::((String ('U', (String ('E', EmptyString))))::((String
    ('C', (String ('R', EmptyString))))::((String ('U', (String ('R',
    EmptyString))))::((String ('C', (String ('D', EmptyString))))::((String
    ('U', (String ('D', EmptyString))))::[])))))))))

(** val eval_table : answer list list **)

let eval_table =
  (Yes::(Yes::(Yes::(No::(No::(No::(No::(Yes::(No::(No::[]))))))))))::((No::(Yes::(Yes::(No::(No::(No::(No::(Yes::(No::(No::[]))))))))))::((No::(No::(Yes::(No::(No::(No::(No::(Yes::(No::(No::[]))))))))))::((No::(No::(No::(Yes::(Yes::(Yes::(No::(No::(No::(Yes::[]))))))))))::((No::(No::(No::(No::(Yes::(Yes::(No::(No::(No::(Yes::[]))))))))))::((No::(No::(No::(No::(No::(Yes::(No::(No::(No::(Yes::[]))))))))))::((No::(No::(Yes::(No::(No::(No::(Yes::(Yes::(No::(No::[]))))))))))::((No::(No::(Yes::(No::(No::(No::(No::(Yes::(No::(No::[]))))))))))::((No::(No::(No::(No::(No::(Yes::(No::(No::(Yes::(Yes::[]))))))))))::((No::(No::(No::(No::(No::(Yes::(No::(No::(No::(Yes::[]))))))))))::[])))))))))

(** val print_eval_table : unit **)

let print_eval_table =
  ExportTo.TeX.print_simple setLAnames setLAnames eval_table

