Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Taint.
Require Export CDistrust.
Require Import FinSetLattice.
Require Import EmbedTCT.
Require Import EmbedCTCD.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_T_CD: Taint ==>[boolean_sem] CDistrust.
Proof.
transitivity_embeds CTaint; solve_embed.
Qed.
Hint Resolve boolean_embeds_T_CD.

Lemma not_boolean_embeds_CD_T: ~(CDistrust ==>[boolean_sem] Taint).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := CTaint)
     (LA2 := CDistrust)
     (LA3 := Taint);
auto.
Qed.
Hint Resolve not_boolean_embeds_CD_T.

Lemma not_eval_embeds_T_CD: ~(Taint ==>[eval_sem] CDistrust).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ Taint CDistrust _ _) as H'.
apply H' with (l := FinSet_singleton (Principal.Prin.enum 0)) in H.
* simpl in H. unfold taint_flowsTo in H. destruct H as [H _].
  repeat rewrite FinSet_union_with_empty_set_included1 in H.
  assert (In _
             (Add Principal.Prin.t (Empty_set Principal.Prin.t)
                  (Principal.Prin.enum 0))
             (Principal.Prin.enum 0)) as HIn by auto with sets.
  specialize (H _ HIn). destruct H.
* reflexivity.
* reflexivity.
Qed.
Hint Resolve not_eval_embeds_T_CD.

Lemma not_eval_embeds_CD_T: ~(CDistrust ==>[eval_sem] Taint).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := CDistrust)
     (LA2 := Taint)
     (LA3 := CTaint);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_CD_T.

(** * Summary *)
Open Scope table.

Theorem injects_T_CD_summary :
  property_square_table injects_in
     [| Taint; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_T_CD_summary :
  embed_table boolean_sem
     [| Taint; CDistrust |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_T_CD_summary :
  embed_table eval_sem
     [| Taint; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
