Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CEndorsement.
Require Export Taint.
Require Import EmbedCRCE.
Require Import EmbedTCR.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import PrincipalFacts.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_CE_T : ~(CEndorsement ==>[boolean_sem] Taint).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := CReader)
     (LA2 := CEndorsement)
     (LA3 := Taint);
auto.
Qed.
Hint Resolve not_boolean_embeds_CE_T.

Lemma not_boolean_embeds_T_CE: ~(Taint ==>[boolean_sem] CEndorsement).
Proof.
apply embeds_in_trans_not2
with (sem := boolean_sem)
     (LA1 := Taint)
     (LA2 := CEndorsement)
     (LA3 := CReader);
auto.
Qed.
Hint Resolve not_boolean_embeds_T_CE.

Lemma not_eval_embeds_T_CE: ~(Taint ==>[eval_sem] CEndorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ Taint CEndorsement _ _) as H'.
apply H' with (l := FinSet_singleton (Principal.Prin.enum 0)) in H.
* simpl in H. destruct H as [H _].
  assert (In _ (Add _ (Empty_set _) (Principal.Prin.enum 0))
             (Principal.Prin.enum 0)) as Hin by auto with sets.
  specialize (H _ Hin). destruct H.
* reflexivity.
* reflexivity.
Qed.
Hint Resolve not_eval_embeds_T_CE.

Lemma not_eval_embeds_CE_T: ~(CEndorsement ==>[eval_sem] Taint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CEndorsement Taint _ _) as H'.
apply H' with (l := top) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CE_T.

(** * Summary *)
Open Scope table.

Theorem injects_T_CE_summary :
  property_square_table injects_in
     [| Taint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_T_CE_summary :
  embed_table boolean_sem
     [| Taint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_T_CE_summary :
  embed_table eval_sem
     [| Taint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
