Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Taint.
Require Export CReader.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.
Require Import Constructive_sets.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_T_CR: ~(Taint ==>[boolean_sem] CReader).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
assert (exists S, Amap m (FinSet_empty _) = CFinSet_intro S) as [SA0 HSA0].
{ case_eq (Amap m (FinSet_empty _)).
  * intros S HS. exists S. reflexivity.
  * intro Heq. exfalso.
    specialize (Hm1 (FinSet_empty _)
                    (FinSet_singleton (Principal.Prin.enum 0))
                    (FinSet_empty _)).
    simpl in Hm1.
    unfold flip, taint_flowsTo in Hm1.
    rewrite Heq in Hm1. destruct Hm1 as [_ Hm1].
    assert
      (CFinSet_included (Lmap m (FinSet_empty Principal.Prin.t))
                        (CFinSet_union
                           (Lmap m (FinSet_singleton (Principal.Prin.enum 0)))
                           (CFinSet_full Principal.Prin.t))) as H.
    { transitivity (CFinSet_full Principal.Prin.t).
      + apply (@top_maximum _ _ (CFinSet_included_HasTop _)).
      + apply (@join_ub2 _ _ _ (CFinSet_JoinPreLattice _)).
    }
    specialize (Hm1 H). clear H.
    assert (In _ (proj1_sig (FinSet_singleton (Principal.Prin.enum 0)))
               (Principal.Prin.enum 0))
      as H.
    { simpl. auto with sets. }
    specialize (Hm1 _ H). clear H. simpl in Hm1.
    apply Union_inv in Hm1. destruct Hm1 as [[] | []].
}
apply (fprinset_to_cprinset_inj_quasi_antitone SA0 (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). destruct Hm2 as [_ Hm2].
  simpl in Hm2. unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intros x y H. specialize (Hm1 (FinSet_empty _) x y).
  rewrite <- HSA0. apply Hm1.
  simpl. unfold taint_flowsTo.
  rewrite FinSet_union_with_empty_set_included1. assumption.
Qed.
Hint Resolve not_boolean_embeds_T_CR.

Lemma not_boolean_embeds_CR_T: ~(CReader ==>[boolean_sem] Taint).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (cprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (CFinSet_empty _)) (Amap m (CFinSet_empty _)))
         (fun x => Lmap m x)).
* intros x y H. specialize (Hm2 x y). destruct Hm2 as [_ Hm2].
  simpl in Hm2. unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intro x. specialize (Hm1 (CFinSet_empty _) x (CFinSet_empty _)).
  apply Hm1. simpl in *. unfold flip in *.
  destruct x; simpl; auto.
  apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
Qed.
Hint Resolve not_boolean_embeds_CR_T.

(** * Summary *)
Open Scope table.

Theorem injects_T_CR_summary :
  property_square_table injects_in
     [| Taint; CReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_T_CR_summary :
  embed_table boolean_sem
     [| Taint; CReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_T_CR_summary :
  embed_table eval_sem
     [| Taint; CReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
