Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Taint.
Require Export CTaint.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.

(** * Intermediate lemmas. *)
Lemma injects_T_CT: Taint ↪ CTaint.
Proof.
exists {| Lmap l := CFinSet_intro l ; Amap a := CFinSet_intro a |}.
rewrite LAinjection_iff. unfold LAinjection_prop.
splits; reflexivity.
Qed.
Hint Resolve injects_T_CT.

Lemma not_boolean_embeds_CT_T: ~(CTaint ==>[boolean_sem] Taint).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (cprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (CFinSet_full _)) (Amap m (CFinSet_empty _)))
         (fun x => Lmap m x)).
* intros x y H.
  assert (x =L y) as Heq.
  { apply Hm2. simpl; unfold taint_flowsTo.
    repeat rewrite FinSet_union_with_empty_set_included1.
    assumption.
  }
  simpl in Heq. unfold ctaint_flowsTo in Heq.
  repeat rewrite CFinSet_union_with_empty_set_included1 in Heq.
  assumption.
* intro x.
  specialize (Hm1 (CFinSet_empty _) x (CFinSet_full _)).
  destruct Hm1 as [Hm1 _].
  apply Hm1. simpl. unfold ctaint_flowsTo.
  destruct x; simpl; auto.
Qed.
Hint Resolve not_boolean_embeds_CT_T.

(** * Summary *)
Open Scope table.

Theorem injects_T_CT_summary :
  property_square_table injects_in
     [| Taint; CTaint |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_T_CT_summary :
  embed_table boolean_sem
     [| Taint; CTaint |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_T_CT_summary :
  embed_table eval_sem
     [| Taint; CTaint |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
