Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Taint.
Require Export Endorsement.
Require Import EmbedTableCheck.
Require Import FinSetLattice.
Require Import PrincipalFacts.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_T_E: ~(Taint ==>[boolean_sem] Endorsement).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (fprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (FinSet_empty _)) (Amap m (FinSet_empty _)))
         (fun x => Lmap m x)).
* intros x y H.
  specialize (Hm2 x y). destruct Hm2 as [_ Hm2]. simpl in Hm2.
  unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intro x. apply (Hm1 (FinSet_empty _) (FinSet_empty _) x).
  simpl. apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
Qed.
Hint Resolve not_boolean_embeds_T_E.

Lemma not_boolean_embeds_E_T: ~(Endorsement ==>[boolean_sem] Taint).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (fprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (FinSet_empty _)) (Amap m (FinSet_empty _)))
         (fun x => Lmap m x)).
* intros x y H.
  specialize (Hm2 x y). destruct Hm2 as [_ Hm2]. simpl in Hm2.
  unfold flip in Hm2.
  simpl in H. simpl. tauto.
* intro x. apply (Hm1 (FinSet_empty _) x (FinSet_empty _)).
  simpl. apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
Qed.
Hint Resolve not_boolean_embeds_E_T.

Lemma not_eval_embeds_E_T: ~(Endorsement ==>[eval_sem] Taint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ Endorsement Taint _ _) as H'.
apply H' with (l := FinSet_singleton (Principal.Prin.enum 0)) in H.
* destruct H as [_ H]. simpl in H.
  assert (In _ (Add _ (Empty_set _) (Principal.Prin.enum 0))
             (Principal.Prin.enum 0)) as Hin by auto with sets.
  specialize (H _ Hin). destruct H.
* reflexivity.
* reflexivity.
Qed.
Hint Resolve not_eval_embeds_E_T.

Lemma not_eval_embeds_T_E: ~(Taint ==>[eval_sem] Endorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ Taint Endorsement _ _) as H'.
apply H' with (l := FinSet_singleton (Principal.Prin.enum 0)) in H.
* destruct H as [H _]. simpl in H.
  assert (In _ (Add _ (Empty_set _) (Principal.Prin.enum 0))
             (Principal.Prin.enum 0)) as Hin by auto with sets.
  specialize (H _ Hin). destruct H.
* reflexivity.
* reflexivity.
Qed.
Hint Resolve not_eval_embeds_T_E.

(** * Summary *)
Open Scope table.

Theorem injects_T_E_summary :
  property_square_table injects_in
     [| Taint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_T_E_summary :
  embed_table boolean_sem
     [| Taint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_T_E_summary :
  embed_table eval_sem
     [| Taint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
