Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UEndorsement.
Require Export Taint.
Require Import EmbedUDUE.
Require Import EmbedTUD.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import FinSetLattice.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_UE_T : ~(UEndorsement ==>[boolean_sem] Taint).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UDistrust)
     (LA2 := UEndorsement)
     (LA3 := Taint);
solve_embed.
Qed.
Hint Resolve not_boolean_embeds_UE_T.

Lemma boolean_embeds_T_UE: Taint ==>[boolean_sem] UEndorsement.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve boolean_embeds_T_UE.

Lemma not_eval_embeds_T_UE: ~(Taint ==>[eval_sem] UEndorsement).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := Taint)
     (LA2 := UEndorsement)
     (LA3 := UDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_T_UE.

Lemma not_eval_embeds_UE_T: ~(UEndorsement ==>[eval_sem] Taint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ UEndorsement Taint _ _) as H'.
apply H' with (l := top) in H.
* simpl in H. destruct H as [_ H].
  specialize (H _ (Full_intro _ (Principal.Prin.enum 0))).
  destruct H.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (Empty_set _).
Qed.
Hint Resolve not_eval_embeds_UE_T.

(** * Summary *)
Open Scope table.

Theorem injects_T_UE_summary :
  property_square_table injects_in
     [| Taint; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_T_UE_summary :
  embed_table boolean_sem
     [| Taint; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_T_UE_summary :
  embed_table eval_sem
     [| Taint; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
