Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Taint.
Require Export UTaint.
Require Import EmbedTCT.
Require Import EmbedCTUT.
Require Import FinSetLattice.
Require Import SetLattice.
Require Import EmbedTableCheck.
Require Import PrincipalFacts.
Require Import Ensembles.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma injects_T_UT: Taint ↪ UTaint.
Proof.
transitivity_embeds CTaint; auto.
Qed.
Hint Resolve injects_T_UT.

Lemma not_boolean_embeds_UT_T: ~(UTaint ==>[boolean_sem] Taint).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
apply (prinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (Lmap m (Full_set _)) (Amap m (Empty_set _)))
         (fun x => Lmap m x)).
* intros x y H.
  assert (x =L y) as Heq.
  { apply Hm2. simpl. unfold Taint.taint_flowsTo.
    repeat rewrite FinSet_union_with_empty_set_included1.
    assumption.
  }
  simpl in Heq. unfold taint_flowsTo in Heq.
  repeat rewrite Union_with_empty_set_Included1 in Heq.
  assumption.
* intros x.
  specialize (Hm1 bottom x top).
  destruct Hm1 as [Hm1 _].
  apply Hm1. apply top_maximum.
  Grab Existential Variables.
  + exact (Empty_set _).
Qed.
Hint Resolve not_boolean_embeds_UT_T.

(** * Summary *)
Open Scope table.

Theorem injects_T_UT_summary :
  property_square_table injects_in
     [| Taint; UTaint |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_T_UT_summary :
  embed_table boolean_sem
     [| Taint; UTaint |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_T_UT_summary :
  embed_table eval_sem
     [| Taint; UTaint |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
