Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Import TwoA.
Require Import CDistrust.
Require Import Ensembles.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma injects_twoA_CD : TwoA ↪ CDistrust.
Proof.
pose (f x := match x return CFinSet Principal.Prin.t with
               | Bottom2 => CFinSet_empty _
               | Top2 => CFinSet_full _
             end).
exists {| Lmap := f; Amap := f |}.
rewrite LAinjection_iff. unfold LAinjection_prop. splits; simpl.
* split; reflexivity.
* intros a1 a2; destruct a1; destruct a2; simpl; try tauto.
  split; intro H; try tauto. reflexivity.
* intros a1 a2; destruct a1; destruct a2; simpl; try tauto.
  apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
* tauto.
* intros l1 l2; destruct l1; destruct l2; simpl; try tauto.
  apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
* intros l1 l2; destruct l1; destruct l2; simpl; try tauto; try reflexivity.
  apply (@meet_lb1 _ _ _ (FinSet_MeetPreLattice _)).
* intros a l1 l2; destruct a; destruct l1; destruct l2;
  simpl; try tauto.
  split; intro H; trivial.
  apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
Qed.
Hint Resolve injects_twoA_CD.

Lemma not_boolean_embeds_CD_twoA : ~(CDistrust ==>[boolean_sem] TwoA).
Proof.
intro H. apply boolean_embeds_iff in H.
destruct H as [m [Hm1 Hm2]].
case_eq (Lmap m (CFinSet_full _)); intro HeqL;
case_eq (Lmap m (CFinSet_empty _)); intro HeqL'.
* specialize (Hm2 (CFinSet_full _) (CFinSet_empty _)).
  rewrite HeqL in Hm2. rewrite HeqL' in Hm2. simpl in Hm2. tauto.
* specialize (Hm1 (CFinSet_empty _) (CFinSet_full _) (CFinSet_empty _)).
  rewrite HeqL in Hm1. rewrite HeqL' in Hm1. simpl in Hm1.
  apply Hm1. destruct ((Amap m (CFinSet_empty _))); exact I.
* case_eq (Lmap m (CFinSet_singleton (Principal.Prin.enum 0))); intro HeqL''.
  + specialize (Hm2 (CFinSet_empty _) (CFinSet_singleton (Principal.Prin.enum 0))).
    rewrite HeqL' in Hm2. rewrite HeqL'' in Hm2. simpl in Hm2.
    destruct Hm2 as [_ Hm2]. specialize (Hm2 (conj I I)).
    destruct Hm2 as [_ Hm2].
    assert (In _ (Add _ (Empty_set _)
                      (Principal.Prin.enum 0)) (Principal.Prin.enum 0))
      as Hin by auto with sets.
    specialize (Hm2 _ Hin). destruct Hm2.
  + specialize (Hm2 (CFinSet_full _) (CFinSet_singleton (Principal.Prin.enum 0))).
    rewrite HeqL in Hm2. rewrite HeqL'' in Hm2. simpl in Hm2. tauto.
* specialize (Hm2 (CFinSet_full _) (CFinSet_empty _)).
  rewrite HeqL in Hm2. rewrite HeqL' in Hm2. simpl in Hm2. tauto.
Qed.
Hint Resolve not_boolean_embeds_CD_twoA.

Lemma not_eval_embeds_CD_twoA : ~(CDistrust ==>[eval_sem] TwoA).
Proof.
rewrite eval_embeds_iff. intros [m [_ [Hm2 [_ _]]]].
assert (forall auth l1 l2 : CFinSet Principal.Prin.t,
          (l1 ⊑[ auth]l2 <-> Lmap m l1 ⊑[Amap m auth]Lmap m l2)) as H.
{ intros a l1 l2.
  specialize (Hm2 (CFinSet_full _) a l1 l2). apply Hm2.
  simpl. destruct l1; exact I.
}
clear Hm2.
case_eq (Amap m (CFinSet_empty _)); intro HeqA;
case_eq (Lmap m (CFinSet_full _)); intro HeqL;
case_eq (Lmap m (CFinSet_empty _)); intro HeqL';
try solve
    [ specialize (H (CFinSet_empty _) (CFinSet_full _) (CFinSet_empty _));
      rewrite HeqA in H; rewrite HeqL in H; rewrite HeqL' in H; simpl in H;
      tauto ].
case_eq (Lmap m (CFinSet_singleton (Principal.Prin.enum 0))); intro HeqL''.
* specialize (H (CFinSet_empty _)
                (CFinSet_singleton (Principal.Prin.enum 0)) (CFinSet_empty _)).
  rewrite HeqA in H; rewrite HeqL' in H; rewrite HeqL'' in H; simpl in H.
  destruct H as [_ H]. specialize (H I).
  rewrite FinSet_union_with_empty_set_included1 in H.
  assert (In _ (Add _ (Empty_set _)
                    (Principal.Prin.enum 0)) (Principal.Prin.enum 0))
    as Hin by auto with sets.
  specialize (H _ Hin). destruct H.
* specialize (H (CFinSet_empty _)
                (CFinSet_full _) (CFinSet_singleton (Principal.Prin.enum 0))).
  rewrite HeqA in H; rewrite HeqL in H; rewrite HeqL'' in H; simpl in H.
  tauto.
Qed.
Hint Resolve not_eval_embeds_CD_twoA.

(** * Summary *)
Open Scope table.

Theorem boolean_embed_twoA_CD_summary :
  embed_table boolean_sem
     [| TwoA; CDistrust |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_twoA_CD_summary :
  embed_table eval_sem
     [| TwoA; CDistrust |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
