Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Import Two.
Require Import TwoA.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_two_twoA : Two ==>[boolean_sem] TwoA.
Proof.
apply boolean_embeds_iff.
exists {| Lmap l := l; Amap a := Bottom2 |}. splits; simpl; tauto.
Qed.
Hint Resolve boolean_embeds_two_twoA.

Lemma not_boolean_embeds_twoA_two : ~(TwoA ==>[boolean_sem] Two).
Proof.
intro H. apply boolean_embeds_iff in H.
destruct H as [m [Hm1 Hm2]].
pose proof (Hm1 Bottom2 Top2 Bottom2) as H1. destruct (Amap m Bottom2). simpl in H1.
pose proof (Hm1 Top2 Top2 Bottom2) as H2. destruct (Amap m Top2). simpl in H2.
tauto.
Qed.
Hint Resolve not_boolean_embeds_twoA_two.

Lemma not_eval_embeds_two_twoA : ~(Two ==>[eval_sem] TwoA).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ Two TwoA _ _) as H'.
apply H' with (l := Top2) in H; simpl in *; tauto.
Qed.
Hint Resolve not_eval_embeds_two_twoA.

Lemma not_eval_embeds_twoA_two : ~(TwoA ==>[eval_sem] Two).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ TwoA Two _ _) as H'.
apply H' with (l := Bottom2) in H; simpl in *; tauto.
Qed.
Hint Resolve not_eval_embeds_twoA_two.

(** * Summary *)
Open Scope table.

Theorem boolean_embed_two_twoA_summary :
  embed_table boolean_sem
     [| Two; TwoA |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_two_twoA_summary :
  embed_table eval_sem
     [| Two; TwoA |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
