Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export CEndorsement.
Require Export UDistrust.
Require Import EmbedUDUE.
Require Import EmbedCEUE.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma injects_CE_UD: CEndorsement ↪ UDistrust.
Proof.
transitivity_embeds UEndorsement; solve_embed.
Qed.
Hint Resolve injects_CE_UD.

Lemma not_boolean_embeds_UD_CE: ~(UDistrust ==>[boolean_sem] CEndorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UEndorsement)
     (LA2 := UDistrust)
     (LA3 := CEndorsement);
solve_embed.
Qed.
Hint Resolve not_boolean_embeds_UD_CE.

Lemma not_eval_embeds_UD_CE: ~(UDistrust ==>[eval_sem] CEndorsement).
Proof.
apply embeds_in_trans_not1
with (sem := eval_sem)
     (LA1 := UEndorsement)
     (LA2 := UDistrust)
     (LA3 := CEndorsement);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UD_CE.

(** * Summary *)
Open Scope table.

Theorem injects_UD_CE_summary :
  property_square_table injects_in
     [| UDistrust; CEndorsement |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UD_CE_summary :
  embed_table boolean_sem
     [| UDistrust; CEndorsement |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UD_CE_summary :
  embed_table eval_sem
     [| UDistrust; CEndorsement |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
