Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Endorsement.
Require Export UDistrust.
Require Import EmbedUDUE.
Require Import EmbedEUE.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma injects_E_UD: Endorsement ↪ UDistrust.
Proof.
transitivity_embeds UEndorsement; solve_embed.
Qed.
Hint Resolve injects_E_UD.

Lemma not_boolean_embeds_UD_E: ~(UDistrust ==>[boolean_sem] Endorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UEndorsement)
     (LA2 := UDistrust)
     (LA3 := Endorsement);
solve_embed.
Qed.
Hint Resolve not_boolean_embeds_UD_E.

Lemma not_eval_embeds_UD_E: ~(UDistrust ==>[eval_sem] Endorsement).
Proof.
apply embeds_in_trans_not1
with (sem := eval_sem)
     (LA1 := UEndorsement)
     (LA2 := UDistrust)
     (LA3 := Endorsement);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UD_E.

(** * Summary *)
Open Scope table.

Theorem injects_UD_E_summary :
  property_square_table injects_in
     [| UDistrust; Endorsement |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_E_UD_summary :
  embed_table boolean_sem
     [| Endorsement; UDistrust |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_E_UD_summary :
  embed_table eval_sem
     [| Endorsement; UDistrust |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
