Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UDistrust.
Require Export UEndorsement.
Require Import SetLattice.
Require Import EmbedTableCheck.
Require Import Classical_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma injects_UD_UE: UDistrust ↪ UEndorsement.
Proof.
exists {| Lmap := Complement Principal.Prin.t
        ; Amap a := a |}.
rewrite LAinjection_iff.
unfold LAinjection_prop.
splits; simpl; unfold flip, UTaint.taint_flowsTo;
try split; intros; auto with sets; try reflexivity;
try solve [ rewrite Complement_Full_set; reflexivity ];
try solve [ rewrite Complement_Union; reflexivity];
try solve [ rewrite Complement_Intersection; reflexivity].
* apply Complement_adjoint. rewrite <- Complement_Union.
  rewrite <- Complement_Included. assumption.
* apply Complement_adjoint in H. rewrite <- Complement_Union in H.
  rewrite Complement_Included. assumption.
Qed.
Hint Resolve injects_UD_UE.

Lemma injects_UE_UD: UEndorsement ↪ UDistrust.
Proof.
exists {| Lmap := Complement Principal.Prin.t
        ; Amap a := a |}.
rewrite LAinjection_iff.
unfold LAinjection_prop.
splits; simpl; unfold flip, UTaint.taint_flowsTo;
try split; intros; auto with sets; try reflexivity;
try solve [ rewrite Complement_Empty_set; reflexivity ];
try solve [ rewrite Complement_Union; reflexivity];
try solve [ rewrite Complement_Intersection; reflexivity].
* apply Complement_adjoint. rewrite <- Complement_Union.
  rewrite <- Complement_Included. assumption.
* apply Complement_adjoint in H. rewrite <- Complement_Union in H.
  rewrite Complement_Included. assumption.
Qed.
Hint Resolve injects_UE_UD.

(** * Summary *)
Open Scope table.

Theorem injects_UD_UE_summary :
  property_square_table injects_in
     [| UDistrust; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UD_UE_summary :
  embed_table boolean_sem
     [| UDistrust; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UD_UD_summary :
  embed_table eval_sem
     [| UDistrust; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
