Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UReader.
Require Export CDistrust.
Require Import EmbedURUD.
Require Import EmbedCDUD.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_UR_CD: ~(UReader ==>[boolean_sem] CDistrust).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UDistrust)
     (LA2 := UReader)
     (LA3 := CDistrust);
auto.
Qed.
Hint Resolve not_boolean_embeds_UR_CD.

Lemma boolean_embeds_CD_UR: CDistrust ==>[boolean_sem] UReader.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve boolean_embeds_CD_UR.

Lemma not_eval_embeds_UR_CD: ~(UReader ==>[eval_sem] CDistrust).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UReader)
     (LA2 := CDistrust)
     (LA3 := UDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UR_CD.

Lemma not_eval_embeds_CD_UR: ~(CDistrust ==>[eval_sem] UReader).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CDistrust UReader _ _) as H'.
apply H' with (l := bottom) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CD_UR.

(** * Summary *)
Open Scope table.

Theorem injects_UR_CD_summary :
  property_square_table injects_in
     [| UReader; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UR_CD_summary :
  embed_table boolean_sem
     [| UReader; CDistrust |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UR_CD_summary :
  embed_table eval_sem
     [| UReader; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
