Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UReader.
Require Export CEndorsement.
Require Import EmbedCEUE.
Require Import EmbedURUE.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_UR_CE: ~(UReader ==>[boolean_sem] CEndorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UEndorsement)
     (LA2 := UReader)
     (LA3 := CEndorsement);
auto.
Qed.
Hint Resolve not_boolean_embeds_UR_CE.

Lemma boolean_embeds_CE_UR: CEndorsement ==>[boolean_sem] UReader.
Proof.
transitivity_embeds UEndorsement; solve_embed.
Qed.
Hint Resolve boolean_embeds_CE_UR.

Lemma not_eval_embeds_UR_CE: ~(UReader ==>[eval_sem] CEndorsement).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UReader)
     (LA2 := CEndorsement)
     (LA3 := UEndorsement);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UR_CE.

Lemma not_eval_embeds_CE_UR: ~(CEndorsement ==>[eval_sem] UReader).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CEndorsement UReader _ _) as H'.
apply H' with (l := top) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CE_UR.

(** * Summary *)
Open Scope table.

Theorem injects_UR_CE_summary :
  property_square_table injects_in
     [| UReader; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UR_UR_summary :
  embed_table boolean_sem
     [| UReader; CEndorsement |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UR_UR_summary :
  embed_table eval_sem
     [| UReader; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
