Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Endorsement.
Require Export UReader.
Require Import FinSetLattice.
Require Import EmbedECE.
Require Import EmbedURCE.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_E_UR: Endorsement ==>[boolean_sem] UReader.
Proof.
transitivity_embeds CEndorsement; solve_embed.
Qed.
Hint Resolve boolean_embeds_E_UR.

Lemma not_boolean_embeds_UR_E: ~(UReader ==>[boolean_sem] Endorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := CEndorsement)
     (LA2 := UReader)
     (LA3 := Endorsement);
auto.
Qed.
Hint Resolve not_boolean_embeds_UR_E.

Lemma not_eval_embeds_E_UR: ~(Endorsement ==>[eval_sem] UReader).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ Endorsement UReader _ _) as H'.
apply H' with (l := FinSet_singleton (Principal.Prin.enum 0)) in H.
* simpl in H. unfold flip in H. destruct H as [_ H].
  repeat rewrite FinSet_union_with_empty_set_included1 in H.
  assert (In _
             (Add Principal.Prin.t (Empty_set Principal.Prin.t)
                  (Principal.Prin.enum 0))
             (Principal.Prin.enum 0)) as HIn by auto with sets.
  specialize (H _ HIn). destruct H.
* reflexivity.
* reflexivity.
Qed.
Hint Resolve not_eval_embeds_E_UR.

Lemma not_eval_embeds_UR_E: ~(UReader ==>[eval_sem] Endorsement).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UReader)
     (LA2 := Endorsement)
     (LA3 := CEndorsement);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UR_E.

(** * Summary *)
Open Scope table.

Theorem injects_UR_E_summary :
  property_square_table injects_in
     [| UReader; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_E_UR_summary :
  embed_table boolean_sem
     [| Endorsement; UReader |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_E_UR_summary :
  embed_table eval_sem
     [| Endorsement; UReader |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
