Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UReader.
Require Export UDistrust.
Require Import EmbedUTUR.
Require Import EmbedUTUD.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_UR_UD: UReader ==>[boolean_sem] UDistrust.
Proof.
transitivity_embeds UTaint; solve_embed.
Qed.
Hint Resolve boolean_embeds_UR_UD.

Lemma boolean_embeds_UD_UR: UDistrust ==>[boolean_sem] UReader.
Proof.
transitivity_embeds UTaint; solve_embed.
Qed.
Hint Resolve boolean_embeds_UD_UR.

Lemma not_eval_embeds_UR_UD: ~(UReader ==>[eval_sem] UDistrust).
Proof.
apply embeds_in_trans_not1
with (sem := eval_sem)
     (LA1 := UTaint)
     (LA2 := UReader)
     (LA3 := UDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UR_UD.

Lemma not_eval_embeds_UD_UR: ~(UDistrust ==>[eval_sem] UReader).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UDistrust)
     (LA2 := UReader)
     (LA3 := UTaint);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UD_UR.

(** * Summary *)
Open Scope table.

Theorem injects_UR_UD_summary :
  property_square_table injects_in
     [| UReader; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UR_UD_summary :
  embed_table boolean_sem
     [| UReader; UDistrust |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UR_UD_summary :
  embed_table eval_sem
     [| UReader; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
