Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UReader.
Require Export UEndorsement.
Require Import SetLattice.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_UR_UE: UReader ==>[boolean_sem] UEndorsement.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_UR_UE.

Lemma boolean_embeds_UE_UR: UEndorsement ==>[boolean_sem] UReader.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_UE_UR.

Lemma not_eval_embeds_UR_UE: ~(UReader ==>[eval_sem] UEndorsement).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ UReader UEndorsement _ _) as H'.
apply H' with (l := bottom) in H.
* destruct H as [H _].
  specialize (H (Principal.Prin.enum 0) (Full_intro _ _)).
  destruct H.
* reflexivity.
* simpl. unfold flip. split; reflexivity.
  Grab Existential Variables.
  + exact (Empty_set _).
Qed.
Hint Resolve not_eval_embeds_UR_UE.

Lemma not_eval_embeds_UE_UR: ~(UEndorsement ==>[eval_sem] UReader).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ UEndorsement UReader _ _) as H'.
apply H' with (l := top) in H.
* destruct H as [_ H].
  specialize (H (Principal.Prin.enum 0) (Full_intro _ _)).
  destruct H.
* reflexivity.
* simpl. unfold flip. split; reflexivity.
  Grab Existential Variables.
  + exact (Empty_set _).
Qed.
Hint Resolve not_eval_embeds_UE_UR.

(** * Summary *)
Open Scope table.

Theorem injects_UR_UE_summary :
  property_square_table injects_in
     [| UReader; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_UR_summary :
  embed_table boolean_sem
     [| UReader; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_UR_summary :
  embed_table eval_sem
     [| UReader; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
