Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UTaint.
Require Export CDistrust.
Require Import EmbedCDUD.
Require Import EmbedUTUD.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_UT_CD: ~(UTaint ==>[boolean_sem] CDistrust).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UDistrust)
     (LA2 := UTaint)
     (LA3 := CDistrust);
auto.
Qed.
Hint Resolve not_boolean_embeds_UT_CD.

Lemma boolean_embeds_CD_UT: CDistrust ==>[boolean_sem] UTaint.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve boolean_embeds_CD_UT.

Lemma not_eval_embeds_UT_CD: ~(UTaint ==>[eval_sem] CDistrust).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UTaint)
     (LA2 := CDistrust)
     (LA3 := UDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UT_CD.

Lemma not_eval_embeds_CD_UT: ~(CDistrust ==>[eval_sem] UTaint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CDistrust UTaint _ _) as H'.
apply H' with (l := bottom) in H.
* simpl in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CD_UT.

(** * Summary *)
Open Scope table.

Theorem injects_UT_CD_summary :
  property_square_table injects_in
     [| UTaint; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UT_UT_summary :
  embed_table boolean_sem
     [| UTaint; CDistrust |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UT_UT_summary :
  embed_table eval_sem
     [| UTaint; CDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
