Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UReader.
Require Export CEndorsement.
Require Import EmbedUTUD.
Require Import EmbedUDCE.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_UT_CE: ~(UTaint ==>[boolean_sem] CEndorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UDistrust)
     (LA2 := UTaint)
     (LA3 := CEndorsement);
solve_embed.
Qed.
Hint Resolve not_boolean_embeds_UT_CE.

Lemma boolean_embeds_CE_UT: CEndorsement ==>[boolean_sem] UTaint.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve boolean_embeds_CE_UT.

Lemma not_eval_embeds_UT_CE: ~(UTaint ==>[eval_sem] CEndorsement).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UTaint)
     (LA2 := CEndorsement)
     (LA3 := UDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UT_CE.

Lemma not_eval_embeds_CE_UT: ~(CEndorsement ==>[eval_sem] UTaint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ CEndorsement UTaint _ _) as H'.
apply H' with (l := top) in H.
* compute in H. tauto.
* reflexivity.
* reflexivity.
  Grab Existential Variables.
  + exact (CFinSetLattice.CFinSet_empty _).
Qed.
Hint Resolve not_eval_embeds_CE_UT.

(** * Summary *)
Open Scope table.

Theorem injects_UT_CE_summary :
  property_square_table injects_in
     [| UTaint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UT_CE_summary :
  embed_table boolean_sem
     [| UTaint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UT_CE_summary :
  embed_table eval_sem
     [| UTaint; CEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
