Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UReader.
Require Export Endorsement.
Require Import EmbedUTUD.
Require Import EmbedUDE.
Require Import EmbedTableCheck.
Require Import FinSetLattice.
Require Import SetFacts.
Require Import Constructive_sets.

(** * Intermediate lemmas. *)
Lemma not_boolean_embeds_UT_E: ~(UTaint ==>[boolean_sem] Endorsement).
Proof.
apply embeds_in_trans_not1
with (sem := boolean_sem)
     (LA1 := UDistrust)
     (LA2 := UTaint)
     (LA3 := Endorsement);
solve_embed.
Qed.
Hint Resolve not_boolean_embeds_UT_E.

Lemma boolean_embeds_E_UT: Endorsement ==>[boolean_sem] UTaint.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve boolean_embeds_E_UT.

Lemma not_eval_embeds_UT_E: ~(UTaint ==>[eval_sem] Endorsement).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UTaint)
     (LA2 := Endorsement)
     (LA3 := UDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UT_E.

Lemma not_eval_embeds_E_UT: ~(Endorsement ==>[eval_sem] UTaint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ Endorsement UTaint _ _) as H'.
apply H' with (l := FinSet_singleton (Principal.Prin.enum 0)) in H.
* simpl in H. destruct H as [_ H].
  assert (In _ (Add _ (Empty_set _) (Principal.Prin.enum 0))
             (Principal.Prin.enum 0)) as Hin by auto with sets.
  specialize (H _ Hin). destruct H.
* reflexivity.
* reflexivity.
Qed.
Hint Resolve not_eval_embeds_E_UT.

(** * Summary *)
Open Scope table.

Theorem injects_UT_E_summary :
  property_square_table injects_in
     [| UTaint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UT_E_summary :
  embed_table boolean_sem
     [| UTaint; Endorsement |]
     [ [ ✓; × ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UT_E_summary :
  embed_table eval_sem
     [| UTaint; Endorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
