Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UTaint.
Require Export UDistrust.
Require Import SetLattice.
Require Import EmbedTableCheck.
Require Import Constructive_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_UT_UD: UTaint ==>[boolean_sem] UDistrust.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_UT_UD.

Lemma boolean_embeds_UD_UT: UDistrust ==>[boolean_sem] UTaint.
Proof.
rewrite boolean_embeds_iff.
* exists {| Lmap x := x ; Amap x := x |}.
  splits.
  + reflexivity.
  + reflexivity.
Qed.
Hint Resolve boolean_embeds_UD_UT.

Lemma not_eval_embeds_UT_UD: ~(UTaint ==>[eval_sem] UDistrust).
Proof.
intro H.
pose proof (@embeds_bottom_top _ _ _ _ UTaint UDistrust _ _) as H'.
apply H' with (l := top) in H.
* destruct H as [H _]. simpl in H.
  specialize (H (Principal.Prin.enum 0) (Full_intro _ _)).
  destruct H.
* reflexivity.
* simpl. unfold taint_flowsTo.
  repeat rewrite Union_with_empty_set_Included1. auto with sets.
  Grab Existential Variables.
  + exact (Empty_set _).
Qed.
Hint Resolve not_eval_embeds_UT_UD.

Lemma not_eval_embeds_UD_UT: ~(UDistrust ==>[eval_sem] UTaint).
Proof.
intro H.
pose proof (@embeds_top_bottom _ _ _ _ UDistrust UTaint _ _) as H'.
apply H' with (l := bottom) in H.
* destruct H as [_ H]. simpl in H. unfold taint_flowsTo in H.
  specialize (H (Principal.Prin.enum 0) (Full_intro _ _)).
  destruct H.
* reflexivity.
* simpl. unfold taint_flowsTo.
  repeat rewrite Union_with_empty_set_Included1. auto with sets.
  Grab Existential Variables.
  + exact (Empty_set _).
Qed.
Hint Resolve not_eval_embeds_UD_UT.

(** * Summary *)
Open Scope table.

Theorem injects_UT_UD_summary :
  property_square_table injects_in
     [| UTaint; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_CT_UT_summary :
  embed_table boolean_sem
     [| UTaint; UDistrust |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_CT_UT_summary :
  embed_table eval_sem
     [| UTaint; UDistrust |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
