Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UTaint.
Require Export UEndorsement.
Require Import EmbedUTUD.
Require Import EmbedUDUE.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma boolean_embeds_UT_UE: UTaint ==>[boolean_sem] UEndorsement.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve boolean_embeds_UT_UE.

Lemma boolean_embeds_UE_UT: UEndorsement ==>[boolean_sem] UTaint.
Proof.
transitivity_embeds UDistrust; solve_embed.
Qed.
Hint Resolve boolean_embeds_UE_UT.

Lemma not_eval_embeds_UT_UE: ~(UTaint ==>[eval_sem] UEndorsement).
Proof.
apply embeds_in_trans_not2
with (sem := eval_sem)
     (LA1 := UTaint)
     (LA2 := UEndorsement)
     (LA3 := UDistrust);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UT_UE.

Lemma not_eval_embeds_UE_UT: ~(UEndorsement ==>[eval_sem] UTaint).
Proof.
apply embeds_in_trans_not1
with (sem := eval_sem)
     (LA1 := UDistrust)
     (LA2 := UEndorsement)
     (LA3 := UTaint);
solve_embed.
Qed.
Hint Resolve not_eval_embeds_UE_UT.

(** * Summary *)
Open Scope table.

Theorem injects_UT_UE_summary :
  property_square_table injects_in
     [| UTaint; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UT_UE_summary :
  embed_table boolean_sem
     [| UTaint; UEndorsement |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UT_UE_summary :
  embed_table eval_sem
     [| UTaint; UEndorsement |]
     [ [ ✓; × ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
