Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export UTaint.
Require Export UReader.
Require Import SetLattice.
Require Import EmbedTableCheck.
Require Import Classical_sets.
Require Import SetFacts.

(** * Intermediate lemmas. *)
Lemma injects_UT_UR: UTaint ↪ UReader.
Proof.
exists {| Lmap := Complement Principal.Prin.t
        ; Amap a := a |}.
rewrite LAinjection_iff.
unfold LAinjection_prop.
splits; simpl; unfold flip, taint_flowsTo;
try split; intros; auto with sets; try reflexivity;
try solve [ rewrite Complement_Empty_set; reflexivity ];
try solve [ rewrite Complement_Union; reflexivity];
try solve [ rewrite Complement_Intersection; reflexivity].
* apply Complement_adjoint. rewrite <- Complement_Union.
  rewrite <- Complement_Included. assumption.
* apply Complement_adjoint in H. rewrite <- Complement_Union in H.
  rewrite Complement_Included. assumption.
Qed.
Hint Resolve injects_UT_UR.

Lemma injects_UR_UT: UReader ↪ UTaint.
Proof.
exists {| Lmap := Complement Principal.Prin.t
        ; Amap a := a |}.
rewrite LAinjection_iff.
unfold LAinjection_prop.
splits; simpl; unfold flip, taint_flowsTo;
try split; intros; auto with sets; try reflexivity;
try solve [ rewrite Complement_Full_set; reflexivity ];
try solve [ rewrite Complement_Union; reflexivity];
try solve [ rewrite Complement_Intersection; reflexivity].
* apply Complement_adjoint. rewrite <- Complement_Union.
  rewrite <- Complement_Included. assumption.
* apply Complement_adjoint in H. rewrite <- Complement_Union in H.
  rewrite Complement_Included. assumption.
Qed.
Hint Resolve injects_UR_UT.

(** * Summary *)
Open Scope table.

Theorem injects_UT_UR_summary :
  property_square_table injects_in
     [| UTaint; UReader |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_UT_UR_summary :
  embed_table boolean_sem
     [| UTaint; UReader |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_UT_UT_summary :
  embed_table eval_sem
     [| UTaint; UReader |]
     [ [ ✓; ✓ ]
     ; [ ✓; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
