Require Import Lambda.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import LowerBound.
Require Import TermEquiv.

Lemma eval_injection_preserved {A1 A2 L1 L2}
      {LA1: LabelAlgebra A1 L1} {LA2: LabelAlgebra A2 L2}
      (m: LA1 ==> LA2) :
  LAinjection m ->
  forall pc t a,
    pc; nil ⊢ t ⇓ a ->
    exists a',
      atom_equiv (apply m a) a' /\
      Lmap m pc; nil ⊢ apply m t ⇓ a'.
Proof.
intros Hm pc t a Heval.
destruct Hm as [[_ [_ [_ [_ [_ [Hjoin [_ Hmon]]]]]]] _].
apply lower_bound_eval_preserved; trivial.
* intros auth l1 l2 _ _. symmetry. auto.
* intros auth auth' l1 l2 _ H. apply Hmon. assumption.
Qed.

Lemma eval_injection_reflected_strong {A1 A2 L1 L2}
      {LA1: LabelAlgebra A1 L1} {LA2: LabelAlgebra A2 L2}
      (m: LA1 ==> LA2) :
  LAinjection m ->
  forall pc e t a,
    pc; e ⊢ t ⇓ a ->
    forall pc' e' t',
      pc =L Lmap m pc' ->
      env_equiv e (apply m e') ->
      term_equiv t (apply m t') ->
      exists a',
        atom_equiv a (apply m a') /\
        pc'; e' ⊢ t' ⇓ a'.
Proof.
intros Hm pc e t a Heval.
induction Heval; intros pc' e' t' Hpc Henv Ht;
destruct t'; try solve [compute in Ht; tauto]; simpl in Ht; simpl.
* subst b0. exists (Atom (VBool b) pc'). split; auto. split; auto.
* subst.
  pose proof (list_forall2_atIndex_fun _ _ _ _ _ Henv H)
    as [[v' l'] [Heq [Hv' Hl']]]. fold value_equiv in Hv'.
  pose proof (atIndexMapSome_inv _ _ _ _ Heq) as [[v'' l''] [Hindex Ha'']].
  unfold LA_apply_atom in Ha''. fold (LA_apply_value m v'') in Ha''.
  exists (Atom v'' (l'' ⊔ pc')). split; auto.
  split; auto.
  + fold value_equiv. fold (LA_apply_value m).
    transitivity v'; auto. inversion Ha''; subst. reflexivity.
  + destruct Hm as [[_ [_ [_ [_ [_ [Hjoin _]]]]]] _].
    rewrite <- Hjoin. rewrite Hpc. rewrite Hl'.
    inversion Ha''; subst. reflexivity.
* exists (Atom (VClos e' t') pc'). split; auto. split; auto.
* destruct Ht as [Ht1 Ht2].
  specialize (IHHeval1 _ _ _ Hpc Henv Ht1).
  destruct IHHeval1 as [[v1' l1'] [[Hv1' Hl1'] Heval1']].
  fold (LA_apply_value m) in Hv1'.
  fold atom_equiv in Hv1'. fold env_equiv in Hv1'.
  destruct v1'; simpl in Hv1'; try tauto. destruct Hv1' as [He1' Ht1'].
  specialize (IHHeval2 _ _ _ Hpc Henv Ht2).
  destruct IHHeval2 as [a2' [Ha2' Heval2']].
  specialize (IHHeval3 l1' (a2' :: l) t).
  destruct IHHeval3 as [a3' [Ha3' Heval3']]; auto.
  + split; auto.
  + exists a3'. eauto.
* destruct Ht as [Ht [Hauth Hl0]].
  specialize (IHHeval _ _ _ Hpc Henv Ht).
  destruct IHHeval as [[v' l'] [[Hv' Hl'] Heval']].
  fold value_equiv in Hv'. fold (LA_apply_value m) in Hv'.
  exists (Atom v' l0). split.
  + split; auto.
  + apply @Eval_relabel with (l1 := l'); auto.
    destruct Hm as [_ [_ HflowsTo]]. apply HflowsTo.
    apply (flowsToIncluded a).
    - tauto.
    - transitivity l1; auto. transitivity l; auto.
Qed.

Lemma eval_injection_reflected {A1 A2 L1 L2}
      {LA1: LabelAlgebra A1 L1} {LA2: LabelAlgebra A2 L2}
      (m: LA1 ==> LA2) :
  LAinjection m ->
  forall pc pc' t a,
    pc; nil ⊢ apply m t ⇓ a ->
    pc =L Lmap m pc' ->
    exists a',
      atom_equiv a (apply m a') /\
      pc'; nil ⊢ t ⇓ a'.
Proof.
intros Hm pc pc' t a Heval Hpc.
apply @eval_injection_reflected_strong
  with (pc := pc) (e := nil) (t := apply m t); auto.
Qed.
