Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import Lambda.
Require Import Auth.
Require Import BooleanIndistinguishability.
Require Import TermEquiv.
Require Import LowerBound.
Require Import LambdaNI.
Require Import MyClassical.
Require Import BooleanSemantics.
Require Import EvaluationInjection.

Set Implicit Arguments.

Definition eval_sem : SEM term :=
  fun A L LA auth lab t1 t2 =>
    exists a1 a2,
         Ldef; nil ⊢ t1 ⇓ a1
      /\ Ldef; nil ⊢ t2 ⇓ a2
      /\ bool_Lequiv auth lab a1 a2.

Lemma good_LASemantics_eval_sem : good_LASemantics eval_sem.
Proof.
intros A L LA. splits.
* intros a a' l Hleq t1 t2 [a1 [a2 [Heval1 [Heval2 Hatom_Lequiv]]]].
  exists a1. exists a2.
  splits; eauto using bool_Lequiv_auth_antimonotone.
* intros a l l' Hleq t1 t2 [a1 [a2 [Heval1 [Heval2 Hatom_Lequiv]]]].
  exists a1. exists a2.
  splits; eauto using bool_Lequiv_label_antimonotone.
* intros A' L' LA' m Hm a l t1 t2.
  assert (Lmap m Ldef =L Ldef) as HLdef.
  { destruct Hm as [[_ [_ [_ [_ [? _]]]]] _]. assumption. }
  split; intros [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  + apply (eval_injection_preserved m) in Heval1; trivial.
    destruct Heval1 as [a1' [Ha1' Heval1']].
    apply eval_nil_change_pc with (pc2 := Ldef) in Heval1'; trivial.
    destruct Heval1' as [a1'' [Heval1'' Ha1'']].
    apply (eval_injection_preserved m) in Heval2; trivial.
    destruct Heval2 as [a2' [Ha2' Heval2']].
    apply eval_nil_change_pc with (pc2 := Ldef) in Heval2'; trivial.
    destruct Heval2' as [a2'' [Heval2'' Ha2'']].
    exists a1'', a2''. splits; trivial.
    transitivity (apply m a1).
    - { apply @bool_Lequiv_equiv_r with (a1 := a1'').
        + destruct a1 as [v1 l1]. destruct a2 as [v2 l2].
          destruct v1; destruct v2; try solve [compute in Hatom; tauto].
          destruct a1' as [v1' l1']; destruct v1';
          try solve [compute in Ha1'; tauto].
          destruct a1'' as [v1'' l1'']; destruct v1'';
          try solve [compute in Ha1''; tauto].
          simpl. intro; splits; reflexivity.
        + symmetry. transitivity a1'; auto.
      }
    - { transitivity (apply m a2).
        * apply bool_Lequiv_injection; trivial.
        * apply @bool_Lequiv_equiv_l with (a3 := a2'').
          + transitivity a2'; auto.
          + destruct a1 as [v1 l1]. destruct a2 as [v2 l2].
            destruct v1; destruct v2; try solve [compute in Hatom; tauto].
            destruct a2' as [v2' l2']; destruct v2';
            try solve [compute in Ha2'; tauto].
            destruct a2'' as [v2'' l2'']; destruct v2'';
            try solve [compute in Ha2''; tauto].
            simpl. intro; splits; reflexivity.
      }
  + symmetry in HLdef.
    apply eval_nil_change_pc with (pc2 := Lmap m Ldef) in Heval1; trivial.
    destruct Heval1 as [a1' [Heval1' Ha1']].
    apply (eval_injection_reflected m) with (pc' := Ldef) in Heval1'; auto.
    destruct Heval1' as [a1'' [Ha1'' Heval1'']].
    apply eval_nil_change_pc with (pc2 := Lmap m Ldef) in Heval2; trivial.
    destruct Heval2 as [a2' [Heval2' Ha2']].
    apply (eval_injection_reflected m) with (pc' := Ldef) in Heval2'; auto.
    destruct Heval2' as [a2'' [Ha2'' Heval2'']].
    exists a1'', a2''. splits; trivial.
    apply (bool_Lequiv_injection Hm); trivial.
    transitivity a1.
    - { apply @bool_Lequiv_equiv_l with (a3 := a1).
        * symmetry. transitivity a1'; trivial.
        * destruct a1 as [v1 l1]. destruct a2 as [v2 l2].
          destruct v1; destruct v2; try solve [compute in Hatom; tauto].
          destruct a1' as [v1' l1']; destruct v1';
          try solve [compute in Ha1'; tauto].
          simpl. intro; splits; reflexivity.
      }
    - { transitivity a2; trivial.
        apply @bool_Lequiv_equiv_l with (a3 := apply m a2'').
        * transitivity a2'; trivial.
        * destruct a1 as [v1 l1]. destruct a2 as [v2 l2].
          destruct v1; destruct v2; try solve [compute in Hatom; tauto].
          destruct a2' as [v2' l2']; destruct v2';
          try solve [compute in Ha2'; tauto].
          destruct a2'' as [v2'' l2'']; destruct v2'';
          try solve [compute in Ha2''; tauto].
          simpl. intro; splits; reflexivity.
      }
Qed.
Hint Resolve good_LASemantics_eval_sem.

(** * Characterization of sound LA-maps. *)
Section Soundness_characterization.

Context {A1 L1 A2 L2}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2).

Definition LB_Lequiv_preservation_prop (auth0: A1) :=
  (forall auth l1 l2,
     Ldef ⊑[auth0] l1 ->
     Lmap m l1 ⊑[Amap m auth] Lmap m l2 ->
     l1 ⊑[auth] l2)
  /\ (forall l1 l2,
        Ldef ⊑[auth0] l1 ->
        l1 =L l2 ->
        Lmap m l1 =L Lmap m l2).

Lemma LB_atom_Lequiv_preservation_sufficient :
  forall auth0 a1 a2 auth lab,
  LB_Lequiv_preservation_prop auth0 ->
  atom_lower_bound auth0 Ldef a1 ->
  atom_lower_bound auth0 Ldef a2 ->
  bool_Lequiv auth lab a1 a2 ->
  bool_Lequiv (Amap m auth) (Lmap m lab) (apply m a1) (apply m a2).
Proof.
intros auth0 [v1 l1] [v2 l2] auth lab [HflowsTo Heq]
       HLdef1 HLdef2 Hatom.
destruct v1; destruct v2; try tauto.
destruct HLdef1 as [_ Hl1]. destruct HLdef2 as [_ Hl2].
intro Hlab'.
assert (l1 ⊑[auth] lab \/ l2 ⊑[auth] lab) as Hlab.
{ destruct Hlab'; [left|right]; auto. }
specialize (Hatom Hlab). destruct Hatom. auto.
Qed.

Definition soundness_prop :=
     (Lmap m Ldef =L Ldef)
  /\ (forall auth auth' l1 l2,
        Ldef ⊑[auth] l1 ->
        l1 ⊑[auth'] l2 ->
        Lmap m l1 ⊑[Amap m auth'] Lmap m l2)
  /\ (forall auth auth' l1 l2,
        Ldef ⊑[auth] l1 ->
        Lmap m l1 ⊑[Amap m auth'] Lmap m l2 ->
        l1 ⊑[auth'] l2)
  /\ (forall auth l1 l2,
        Ldef ⊑[auth] l1 ->
        l1 =L l2 ->
        Lmap m l1 =L Lmap m l2)
  /\ (forall auth l1 l2,
        Ldef ⊑[auth] l1 ->
        Ldef ⊑[auth] l2 ->
        Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2).

Lemma soundness_necessary : LAMapSound m eval_sem -> soundness_prop.
Proof.
intro Hsound. unfold soundness_prop. splits.
* assert (@eval_sem _ _ LA1 bottom Ldef
                    (TRelabel (TBool true) bottom Ldef) (TBool true)) as Hsem.
  { pose (a := @Atom A1 L1 (VBool true) Ldef).
    exists a, a. unfold a. splits; eauto. split; auto.
  }
  specialize (Hsound _ _ _ _ Hsem).
  destruct Hsound as [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  simpl in Heval1. inversion Heval1; subst.
  inversion H5; subst.
  simpl in Heval2. inversion Heval2; subst.
  assert (Lmap m Ldef ⊑[Amap m bottom] Lmap m Ldef \/
          Ldef ⊑[Amap m bottom] Lmap m Ldef) as Hlab by auto.
  specialize (Hatom Hlab).
  tauto.
* intros auth auth' l1 l2 HLdef H.
  pose (t := TRelabel (TRelabel (TBool true) auth l1) auth' l2).
  assert (eval_sem _ auth' l2 t t) as Hsem.
  { pose (a := @Atom A1 L1 (VBool true) l2).
    exists a, a. unfold t. unfold a. splits; eauto. split; auto.
  }
  specialize (Hsound _ _ _ _ Hsem).
  destruct Hsound as [a1 [a2 [Heval1 [Heval2 Hatom]]]]. unfold t in *|-.
  inversion Heval1; subst. inversion H6; subst.
  assumption.
* intros auth auth' l1 l2 HLdef H. apply NNPP. intro Hnleq.
  assert (eval_sem _ auth' l2
                   (TRelabel (TBool true) auth l1)
                   (TRelabel (TBool false) auth l1))
    as Hsem.
  { exists (Atom (VBool true) l1).
    exists (Atom (VBool false) l1). splits; eauto.
    intro; tauto.
  }
  specialize (Hsound _ _ _ _ Hsem). simpl in Hsound.
  destruct Hsound as [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  inversion Heval1; subst. inversion H6; subst.
  inversion Heval2; subst. inversion H8; subst.
  specialize (Hatom (or_introl H)).
  destruct Hatom. congruence.
* intros auth l1 l2 HLdef H.
  assert (eval_sem _ auth l1
                   (TRelabel (TBool true) auth l1)
                   (TRelabel (TBool true) auth l2)) as Hsem.
  { exists (Atom (VBool true) l1).
    exists (Atom (VBool true) l2). splits.
    + eapply Eval_relabel; auto.
    + eapply Eval_relabel; auto. transitivity l1; auto.
    + split; auto.
  }
  specialize (Hsound _ _ _ _ Hsem).
  destruct Hsound as [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  inversion Heval1; subst. inversion H6; subst.
  inversion Heval2; subst. inversion H8; subst.
  assert (Lmap m l1 ⊑[Amap m auth] Lmap m l1 \/
          Lmap m l2 ⊑[Amap m auth] Lmap m l1) as Hlab by auto.
  specialize (Hatom Hlab). tauto.
* intros auth l1 l2 HLdef1 HLdef2.
  assert (eval_sem _ auth (l1 ⊔ l2)
                   (TRelabel (TBool true) auth (l1 ⊔ l2))
                   (TApp (TRelabel (TLam (TVar 0)) auth l2)
                         (TRelabel (TBool true) auth l1))) as Hsem.
  { exists (Atom (VBool true) (l1 ⊔ l2)).
    exists (Atom (VBool true) (l1 ⊔ l2)). splits.
    + eapply Eval_relabel; auto. transitivity l1; auto.
    + apply Eval_app with
        (e1' := nil)
        (t1' := TVar 0)
        (a2 := Atom (VBool true) l1)
        (l3 := l2); eauto.
    + intro; auto.
  }
  specialize (Hsound _ _ _ _ Hsem).
  destruct Hsound as [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  inversion Heval1; subst. inversion H5; subst.
  inversion Heval2; subst. inversion H1; subst. inversion H7; subst.
  inversion H4; subst. inversion H11; subst. inversion H8; subst.
  simpl in H3. inversion H3; subst.
  assert (Lmap m (l1 ⊔ l2) ⊑[Amap m auth] Lmap m (l1 ⊔ l2) \/
          Lmap m l1 ⊔ Lmap m l2 ⊑[Amap m auth] Lmap m (l1 ⊔ l2))
    as Hlab by auto.
  specialize (Hatom Hlab). tauto.
Qed.

Lemma soundness_sufficient : soundness_prop -> LAMapSound m eval_sem.
Proof.
intros [HLdef [HflowsTo1 [HflowsTo2 [Heq Hjoin]]]]
       auth lab t1 t2 [a1 [a2 [Heval1 [Heval2 Hatom]]]].
pose proof (lower_bound_eval_preserved m Hjoin HflowsTo1) as H1.
pose proof H1 as H2.
specialize (H1 _ _ Heval1). destruct H1 as [a1' [Ha1' Heval1']].
destruct (eval_nil_change_pc HLdef Heval1') as [a1'' [Heval1'' Hatom1'']].
specialize (H2 _ _ Heval2). destruct H2 as [a2' [Ha2' Heval2']].
destruct (eval_nil_change_pc HLdef Heval2') as [a2'' [Heval2'' Hatom2'']].
exists a1'', a2''. splits; auto.
transitivity (apply m a1).
* assert (bool_Lequiv (Amap m auth) (Lmap m lab) (apply m a1) a1').
  { destruct a1 as [v1 l1]; destruct v1; try solve [ compute in Hatom; tauto ].
    destruct a1' as [v1' l1']; destruct v1'; try solve [ compute in Ha1'; tauto ].
    destruct Ha1'. split; auto.
  }
  transitivity a1'; symmetry; auto.
  apply @bool_Lequiv_equiv_r with (a1 := apply m a1); auto.
* transitivity (apply m a2).
  + apply LB_atom_Lequiv_preservation_sufficient
    with (auth0 := term_auth t1 ∨ term_auth t2); auto.
    - { unfold LB_Lequiv_preservation_prop. split; eauto. }
    - apply lower_bound_eval_lower_bound with (pc := Ldef) (e := nil) (t := t1);
      simpl; auto.
    - apply lower_bound_eval_lower_bound with (pc := Ldef) (e := nil) (t := t2);
      simpl; auto.
  + assert (bool_Lequiv (Amap m auth) (Lmap m lab) (apply m a2) a2').
    { destruct a1 as [v1 l1]; destruct v1; try solve [ compute in Hatom; tauto ];
      destruct a2 as [v2 l2]; destruct v2; try solve [ compute in Hatom; tauto ].
      destruct a2' as [v2' l2']; destruct v2'; try solve [ compute in Ha2'; tauto ].
      destruct Ha2'. split; eauto.
    }
    transitivity a2'; auto.
    apply @bool_Lequiv_equiv_r with (a1 := apply m a2); auto.
Qed.

(** Characterization of sound maps. *)
Lemma soundness_iff : LAMapSound m eval_sem <-> soundness_prop.
Proof.
split.
* apply soundness_necessary.
* apply soundness_sufficient.
Qed.

End Soundness_characterization.

(** * Characterization of complete LA-maps. *)
Section Completeness_characterization.

Context {A1 L1 A2 L2}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2).

Lemma splitH1 (A: Type) (P Q: A -> Prop):
  (forall x, P x /\ Q x) ->
  (forall x, P x)
  /\ (forall x, Q x).
Proof.
intro H. split; intros x; destruct (H x); assumption.
Qed.

Lemma LB_Lequiv_reflection_sufficient :
  (forall auth l,
     Ldef ⊑[auth] l ->
     Ldef ⊑[Amap m auth] Lmap m l) ->
  (forall (auth auth' : A1) (l1 l2 : L1),
     Ldef ⊑[ Amap m auth]Lmap m l1 ->
     l1 ⊑[auth']l2 ->
     Lmap m l1 ⊑[Amap m auth']Lmap m l2) ->
  forall auth lab auth1 auth2 a1' a2' a1 a2,
    atom_lower_bound auth1 Ldef a1 ->
    atom_lower_bound auth2 Ldef a2 ->
    bool_Lequiv (Amap m auth) (Lmap m lab) a1' a2' ->
    (forall auth', atom_below (Amap m auth') a1' (apply m a1)) ->
    atom_best m a1' a1 ->
    (forall auth', atom_below (Amap m auth') a2' (apply m a2)) ->
    atom_best m a2' a2 ->
    bool_Lequiv auth lab a1 a2.
Proof.
intros HLdef HflowsTo.
intros auth lab auth1 auth2 a1' a2' a1 a2
       HLB1 HLB2 Hatom Hbelow1 Hbest1 Hbelow2 Hbest.
destruct a1' as [v1' l1']. destruct a2' as [v2' l2'].
destruct v1'; destruct v2'; try solve [compute in Hatom; tauto].
rename b into b1'. rename b0 into b2'.
destruct a1 as [v1 l1].
destruct v1; try solve [destruct (Hbelow1 bottom); auto].
rename b into b1.
destruct a2 as [v2 l2].
destruct v2; try solve [destruct (Hbelow2 bottom); auto].
rename b into b2.
destruct (splitH1 _ _ Hbelow1) as [Hb1 Hl1]. clear Hbelow1.
specialize (Hb1 bottom). subst b1'.
destruct (splitH1 _ _ Hbelow2) as [Hb2 Hl2]. clear Hbelow2.
specialize (Hb2 bottom). subst b2'.
destruct Hbest1 as [_ Hbest1].
destruct Hbest as [_ Hbest].
destruct HLB1 as [_ HLBl1].
destruct HLB2 as [_ HLBl2].
intros Hlab.
assert (l1' ⊑[Amap m auth] Lmap m lab \/
        l2' ⊑[Amap m auth] Lmap m lab) as Hlab'.
{ destruct Hlab; [left|right].
  * transitivity (Lmap m l1); auto.
    apply (HflowsTo auth1); auto.
  * transitivity (Lmap m l2); auto.
    apply (HflowsTo auth2); auto.
}
specialize (Hatom Hlab'). destruct Hatom as [Hb Hl]. split; auto.
split.
* apply flowsToBottom. apply Hbest1. transitivity l2'; auto.
* apply flowsToBottom. apply Hbest. transitivity l1'; auto.
Qed.

Definition completeness_prop :=
  (forall auth, Ldef ⊑[Amap m auth] Lmap m Ldef)
  /\ (forall auth l,
        Ldef ⊑[auth] l ->
        Ldef ⊑[Amap m auth] Lmap m l)
  /\ (forall auth l,
       Ldef ⊑[Amap m auth] Lmap m l ->
       Ldef ⊑[auth] l)
  /\ (forall (auth0 auth' : A1) (l1 l2 : L1),
        Ldef ⊑[ Amap m auth0]Lmap m l1 ->
        Ldef ⊑[ Amap m auth0]Lmap m l2 ->
        Lmap m l1 ⊔ Lmap m l2 ⊑[ Amap m auth']Lmap m (l1 ⊔ l2))
  /\ (forall auth auth' l1 l2,
        Ldef ⊑[Amap m auth] Lmap m l1 ->
        Lmap m l1 ⊑[Amap m auth'] Lmap m l2 ->
        l1 ⊑[auth'] l2)
  /\ (forall auth auth' l1 l2,
        Ldef ⊑[Amap m auth] Lmap m l1 ->
        l1 ⊑[auth'] l2 ->
        Lmap m l1 ⊑[Amap m auth'] Lmap m l2).

Lemma completeness_necessary :
  LAMapComplete m eval_sem -> completeness_prop.
Proof.
intro Hcomplete. unfold completeness_prop. splits.
* intro auth. apply NNPP. intro Hnleq.
  assert (eval_sem
            _ (Amap m auth) (Lmap m Ldef)
            (apply m (TBool true)) (apply m (TBool false))) as Hsem.
  { exists (Atom (VBool true) Ldef). exists (Atom (VBool false) Ldef).
    splits; try solve [simpl; auto]. intro Hlab. tauto.
  }
  specialize (Hcomplete _ _ _ _ Hsem).
  destruct Hcomplete as [a1 [a2 [Heval1 [Heval2 HLequiv]]]].
  inversion Heval1; subst. inversion Heval2; subst.
  assert (Ldef ⊑[auth] Ldef \/ Ldef ⊑[auth] Ldef) as Hlab by auto.
  specialize (HLequiv Hlab). destruct HLequiv. congruence.
* intros auth l H. apply NNPP. intro Hnleq.
  assert (eval_sem _ (Amap m auth) (Lmap m l)
                   (apply m (TBool true))
                   (apply m (TBool false))) as Hsem.
  { exists (Atom (VBool true) Ldef).
    exists (Atom (VBool false) Ldef).
    splits; try solve [simpl; auto].
    intro Hlab. tauto.
  }
  specialize (Hcomplete _ _ _ _ Hsem).
  destruct Hcomplete as [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  inversion Heval1; subst.
  inversion Heval2; subst.
  assert (Ldef ⊑[ auth]l \/ Ldef ⊑[ auth]l) as Hlab by auto.
  specialize (Hatom Hlab). destruct Hatom. congruence.
* intros auth l Hleq.
  pose (t := apply m (TRelabel (TBool true) auth l)).
  assert (eval_sem _ (Amap m auth) (Lmap m l) t t) as Hsem.
  { pose (a := @Atom A2 L2 (VBool true) (Lmap m l)).
    exists a, a. unfold t. unfold a. splits; solve [simpl; eauto].
  }
  unfold t in *. clear t.
  specialize (Hcomplete _ _ _ _ Hsem).
  destruct Hcomplete as [a1 [a2 [Heval1 _]]].
  inversion Heval1; subst. inversion H5; subst.
  assumption.
* intros auth auth' l1 l2 Hl1 Hl2. apply NNPP. intro Hnleq.
  pose (f b :=
          apply m
          (TApp
             (TRelabel (TLam (TVar 0)) auth l2)
             (TRelabel (TBool b) auth l1))).
  assert (eval_sem _ (Amap m auth') (Lmap m (l1 ⊔ l2)) (f true) (f false)) as Hsem.
  { pose (g b := @Atom A2 L2 (VBool b) (Lmap m l1 ⊔ Lmap m l2)).
    assert (forall b, Ldef; nil ⊢ f b ⇓ g b) as Heval.
    { intro b. unfold f. unfold g.
      apply @Eval_app
      with (e1' := nil)
           (t1' := TVar 0)
           (l1 := Lmap m l2)
           (a2 := Atom (VBool b) (Lmap m l1)); auto.
      * apply @Eval_relabel with (l1 := Ldef); auto.
      * apply @Eval_relabel with (l1 := Ldef); auto.
    }
    exists (g true), (g false). splits; auto.
    clear Heval. subst f. subst g.
    intro Hlab. tauto.
  }
  subst f.
  specialize (Hcomplete _ _ _ _ Hsem).
  destruct Hcomplete as [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  inversion Heval1; subst. inversion H1; subst. inversion H5; subst.
  inversion H6; subst. inversion H3; subst. inversion H4; subst.
  inversion H8; subst.
  clear H1 H5 H3 H6 H4 H8 H10 H13.
  inversion Heval2; subst. inversion H1; subst. inversion H5; subst.
  inversion H6; subst. inversion H3; subst. inversion H4; subst.
  inversion H8; subst.
  clear H1 H5 H3 H6 H4 H8 H10 H13.
  assert (l0 ⊔ l1 ⊑[auth'] l0 ⊔ l1 \/ l0 ⊔ l1 ⊑[auth'] l0 ⊔ l1) as Hlab by auto.
  specialize (Hatom Hlab). destruct Hatom. congruence.
* intros auth auth' l1 l2 HLdef Hleq.
  pose (t := apply m (TRelabel (TRelabel (TBool true) auth l1) auth' l2)).
  assert (eval_sem _ (Amap m bottom) (Lmap m l2) t t) as Hsem.
  { pose (a := @Atom A2 L2 (VBool true) (Lmap m l2)).
    exists a, a. subst t. subst a. splits; simpl; eauto.
  }
  subst t.
  specialize (Hcomplete _ _ _ _ Hsem).
  destruct Hcomplete as [a1 [_ [Heval1 _]]].
  inversion Heval1; subst. inversion H5; subst. assumption.
* intros auth auth' l1 l2 HLB H. apply NNPP. intro Hnleq.
  assert (eval_sem _
                   (Amap m auth') (Lmap m l2)
                   (apply m (TRelabel (TBool true) auth l1))
                   (apply m (TRelabel (TBool false) auth l1))) as Hsem.
  { exists (Atom (VBool true) (Lmap m l1)).
    exists (Atom (VBool false) (Lmap m l1)). splits.
    + simpl; eauto.
    + simpl; eauto.
    + intro. tauto.
  }
  specialize (Hcomplete _ _ _ _ Hsem).
  destruct Hcomplete as [a1 [a2 [Heval1 [Heval2 Hatom]]]].
  inversion Heval1; subst. inversion H6; subst.
  inversion Heval2; subst. inversion H8; subst.
  assert (l1 ⊑[auth']l2 \/ l1 ⊑[auth']l2) as Hlab by auto.
  specialize (Hatom Hlab). destruct Hatom. congruence.
Qed.

Lemma completeness_sufficient :
  completeness_prop -> LAMapComplete m eval_sem.
Proof.
intros [HLdef [HLdefpres [HLdefrefl [Hjoin [HflowsTorefl HflowsTopres]]]]]
       auth lab t1 t2 [a1' [a2' [Heval1' [Heval2' Hatom]]]].
pose proof
     (lower_bound_eval_reflected
        m Hjoin HflowsTorefl HflowsTopres Ldef t1 Heval1' HLdef HLdefrefl)
  as [a1 [Heval1 [Hbelow1 Hbest1]]].
pose proof
     (lower_bound_eval_reflected
        m Hjoin HflowsTorefl HflowsTopres Ldef t2 Heval2' HLdef HLdefrefl)
  as [a2 [Heval2 [Hbelow2 Hbest]]].
exists a1, a2. splits; auto.
apply LB_Lequiv_reflection_sufficient
with (auth1 := term_auth t1) (auth2 := term_auth t2) (a1' := a1') (a2' := a2');
auto.
* apply lower_bound_eval_lower_bound
  with (pc := Ldef) (e := nil) (t := t1); simpl; auto.
* apply lower_bound_eval_lower_bound
  with (pc := Ldef) (e := nil) (t := t2); simpl; auto.
Qed.

(** Characterization of complete maps. *)
Lemma completeness_iff : LAMapComplete m eval_sem <-> completeness_prop.
Proof.
split.
* apply completeness_necessary.
* apply completeness_sufficient.
Qed.

End Completeness_characterization.

(** * Characterization of embeddings. *)
Section Embedding_characterization.

Context {A1 L1 A2 L2}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2).

Definition embedding_prop :=
  (Lmap m Ldef =L Ldef)
  /\ (forall auth auth' l1 l2,
        Ldef ⊑[auth] l1 ->
        (l1 ⊑[auth'] l2 <-> Lmap m l1 ⊑[Amap m auth'] Lmap m l2))
  /\ (forall auth l1 l2,
        Ldef ⊑[auth] l1 ->
        l1 =L l2 ->
        Lmap m l1 =L Lmap m l2)
  /\ (forall auth l1 l2,
        Ldef ⊑[auth] l1 ->
        Ldef ⊑[auth] l2 ->
        Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2).

Lemma embedding_prop_iff :
  embedding_prop <->
  (soundness_prop m /\ completeness_prop m).
Proof.
unfold embedding_prop.
unfold soundness_prop.
unfold completeness_prop.
split; intro H.
* destruct H as [HLdef [HflowsToiff [HLequivpres Hjoin]]].
 splits; try tauto.
  + intros auth auth' l1 l2 Hl1 Hl2.
    apply (HflowsToiff auth); auto.
  + intros auth auth' l1 l2 Hl1 Hl2.
    apply (HflowsToiff auth); auto.
  + intro auth. auto.
  + intros auth l Hl. transitivity (Lmap m Ldef); auto.
    apply (HflowsToiff auth); auto.
  + intros auth l Hl. apply (HflowsToiff auth); auto.
    transitivity Ldef; auto.
  + intros auth auth' l1 l2 Hl1 Hl2.
    apply (flowsToIncluded bottom); auto.
    apply flowsToBottom. destruct (Hjoin auth l1 l2); auto.
    - apply (HflowsToiff auth); auto. transitivity Ldef; auto.
    - apply (HflowsToiff auth); auto. transitivity Ldef; auto.
  + intros auth auth' l1 l2 Hl1 Hl2.
    apply (HflowsToiff auth); auto.
    apply (HflowsToiff auth); auto.
    transitivity Ldef; auto.
  + intros auth auth' l1 l2 Hl1 Hl2.
    apply (HflowsToiff auth); auto.
    apply (HflowsToiff auth); auto.
    transitivity Ldef; auto.
* destruct H as [Hsound Hcomplete].
  destruct Hsound as
      [HLequivLdef [HflowsTopres [HflowsTorefl [HLequivpres HLequivjoin]]]].
  destruct Hcomplete as
      [HflowsToLdef [HLdefpres [HLdefrefl [HflowsToJoin [HflowsTorefl' HflowsTopres']]]]].
  splits; try tauto.
  + intros auth auth' l1 l2 Hl1. split; intro Hleq.
    - apply (HflowsTopres auth); auto.
    - apply (HflowsTorefl auth); auto.
Qed.

(** Characterization of embeddings. *)
Theorem embedding_iff : LAEmbedding m eval_sem <-> embedding_prop.
Proof.
unfold LAEmbedding.
rewrite soundness_iff.
rewrite completeness_iff.
rewrite embedding_prop_iff.
reflexivity.
Qed.

End Embedding_characterization.

(** * Corollaries. *)
Section Corollaries.

Context {A1 L1 A2 L2}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}
        (m: LA1 ==> LA2).

(** Case where the default label of the input LA is bottom. *)
Corollary embedding_bottom1_iff :
  forall (HB1: HasBottom L1 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    (LAEmbedding m eval_sem
     <->
     ((Lmap m Ldef =L Ldef)
      /\ (forall auth l1 l2,
            (l1 ⊑[auth] l2 <-> Lmap m l1 ⊑[Amap m auth] Lmap m l2))
      /\ (forall l1 l2, l1 =L l2 -> Lmap m l1 =L Lmap m l2)
      /\ (forall l1 l2, Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2))).
Proof.
intros HB1 HLdef1.
rewrite embedding_iff. unfold embedding_prop. split; intro H.
* destruct H as [HLdef [HflowsToiff [HLequivpres Hjoin]]].
  splits; auto.
  + intros auth l1 l2. apply (HflowsToiff bottom).
    apply flowsToBottom. transitivity (@bottom _ _ HB1); auto.
    apply bottom_minimum.
  + intros l1 l2 Heq. apply (HLequivpres bottom); auto.
    apply flowsToBottom. transitivity (@bottom _ _ HB1); auto.
    apply bottom_minimum.
  + intros l1 l2. apply (Hjoin bottom).
    - apply flowsToBottom. transitivity (@bottom _ _ HB1); auto.
      apply bottom_minimum.
    - apply flowsToBottom. transitivity (@bottom _ _ HB1); auto.
      apply bottom_minimum.
* destruct H as [HLdef [HflowsToiff [HLequiv Hjoin]]]. splits; auto.
Qed.

(** If the default label of the input LA is bottom, then an evaluation embedding is also a boolean embedding. *)
Corollary embedding_bottom1_eval_boolean :
  forall (HB1: HasBottom L1 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    LAEmbedding m eval_sem ->
    LAEmbedding m boolean_sem.
Proof.
intros HB1 HLdef1 Heval_embed.
rewrite embedding_bottom1_iff in Heval_embed; eauto.
destruct Heval_embed as [HLdef [HflowsToiff [HLequivpres Hjoin]]].
rewrite BooleanSemantics.embedding_iff.
splits; auto.
* intros l1 l2. split; intro H; auto.
  split; apply flowsToBottom; apply HflowsToiff; auto.
Qed.

(** Case where the input has a bottom default and the output has a top default. *)
Corollary embedding_bottom_top :
  forall (HB1: HasBottom L1 labLeq) (HT2: HasTop L2 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    @Ldef _ _ LA2 =L top ->
    LAEmbedding m eval_sem ->
    forall (l: L1), l =L bottom.
Proof.
intros HB1 HT2 HLdef1 HLdef2 Heval_embed l.
rewrite (embedding_bottom1_iff HB1 HLdef1) in Heval_embed.
destruct Heval_embed as [HLdef [HflowsToiff [HLequivpres Hjoin]]].
split.
* apply flowsToBottom. apply (HflowsToiff bottom).
  apply (flowsToIncluded bottom); auto. apply flowsToBottom.
  transitivity Ldef.
  + transitivity top; auto. apply top_maximum.
  + transitivity (Lmap m Ldef); auto.
* apply bottom_minimum.
Qed.

(** Case where the input has a top default and the output has a bottom default. *)
Corollary embedding_top_bottom :
  forall (HT1: HasTop L1 labLeq) (HB2: HasBottom L2 labLeq),
    @Ldef _ _ LA1 =L top ->
    @Ldef _ _ LA2 =L bottom ->
    LAEmbedding m eval_sem ->
    forall (l: L1), l =L top.
Proof.
intros HT1 HB2 HLdef1 HLdef2 Heval_embed l.
rewrite embedding_iff in Heval_embed.
unfold embedding_prop in Heval_embed.
destruct Heval_embed as [HLdef [HflowsToiff [HLequivpres Hjoin]]].
split.
* apply top_maximum.
* apply flowsToBottom. apply (HflowsToiff bottom).
  + apply flowsToBottom. apply top_maximum.
  + apply (flowsToIncluded bottom); auto. apply flowsToBottom.
    transitivity Ldef.
    - transitivity (Lmap m Ldef); auto.
      assert (Lmap m Ldef =L Lmap m top) as [? ?]; auto.
      { apply (HLequivpres bottom); auto. }
    - transitivity bottom; auto. apply bottom_minimum.
Qed.

(** Case where the bottom authority is mapped to bottom. *)
Corollary embedding_auth_bottom :
  Amap m bottom =A bottom ->
  (LAEmbedding m eval_sem
   <->
   (Lmap m Ldef =L Ldef)
   /\ (forall auth auth' l1 l2,
         Ldef ⊑[auth] l1 ->
         (l1 ⊑[auth'] l2 <-> Lmap m l1 ⊑[Amap m auth'] Lmap m l2))
   /\ (forall auth l1 l2,
         Ldef ⊑[auth] l1 ->
         Ldef ⊑[auth] l2 ->
         Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2)).
Proof.
intros [Hbottom _]. rewrite embedding_iff. unfold embedding_prop.
split; intro H; splits; try tauto.
destruct H as [_ [H _]].
intros auth l1 l2 HLdef [Hl1l2 Hl2l1]. split.
* apply flowsToBottom.
  apply (flowsToIncluded (Amap m bottom)); auto.
  apply (H auth); auto.
* apply flowsToBottom.
  apply (flowsToIncluded (Amap m bottom)); auto.
  apply (H auth); auto.
  transitivity l1; auto.
Qed.

(** Case where the bottom authority is mapped to bottom and the default label of the input is bottom. *)
Corollary embedding_bottom1_auth_bottom :
  forall (HB1: HasBottom L1 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    Amap m bottom =A bottom ->
    (LAEmbedding m eval_sem
     <->
     ((Lmap m Ldef =L Ldef)
      /\ (forall auth l1 l2,
            (l1 ⊑[auth] l2 <-> Lmap m l1 ⊑[Amap m auth] Lmap m l2))
      /\ (forall l1 l2, Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2))).
Proof.
intros HB1 HLdef1 [HAbottom _].
rewrite embedding_bottom1_iff; eauto.
split; intro H; splits; try tauto.
destruct H as [_ [H _]].
intros l1 l2 [Hl1l2 Hl2l1]. split.
* apply flowsToBottom.
  apply (flowsToIncluded (Amap m bottom)); auto.
  apply (H bottom); auto.
* apply flowsToBottom.
  apply (flowsToIncluded (Amap m bottom)); auto.
  apply (H bottom); auto.
Qed.

End Corollaries.

(** * Characterization of evaluation-embeddability. *)
Section Eval_embeds.

Context {A1 L1 A2 L2}
        (LA1: LabelAlgebra A1 L1)
        (LA2: LabelAlgebra A2 L2).

Theorem eval_embeds_iff :
  LA1 ==>[ eval_sem ] LA2
  <-> exists m : LA1 ==> LA2, embedding_prop m.
Proof.
unfold embeds_in. simpl projT2. unfold embeds_in_open.
split; intros [m Hm]; exists m; apply embedding_iff; assumption.
Qed.

End Eval_embeds.

(** * Corollaries about evaluation-embeddability. *)
Section Embed_corollaries.

Context {A1 L1 A2 L2}
        {LA1: LabelAlgebra A1 L1}
        {LA2: LabelAlgebra A2 L2}.

(** Case where the default label of the input is bottom. *)
Corollary embeds_bottom1_iff :
  forall (HB1: HasBottom L1 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    (LA1 ==>[eval_sem] LA2
     <->
     (exists m : LA1 ==> LA2,
        (Lmap m Ldef =L Ldef)
        /\ (forall auth l1 l2,
              (l1 ⊑[auth] l2 <-> Lmap m l1 ⊑[Amap m auth] Lmap m l2))
        /\ (forall l1 l2, l1 =L l2 -> Lmap m l1 =L Lmap m l2)
        /\ (forall l1 l2, Lmap m (l1 ⊔ l2) =L Lmap m l1 ⊔ Lmap m l2))).
Proof.
intros HB1 HLdef1.
unfold embeds_in. simpl projT2. unfold embeds_in_open.
split; intros [m Hm]; exists m.
* rewrite <- embedding_bottom1_iff; eauto.
* rewrite -> embedding_bottom1_iff; eauto.
Qed.

(** If the default label of the input is bottom, then evaluation-embeddability implies boolean-embeddability. *)
Corollary embeds_bottom1_eval_boolean :
  forall (HB1: HasBottom L1 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    LA1 ==>[eval_sem] LA2 ->
    LA1 ==>[boolean_sem] LA2.
Proof.
intros HB1 HLdef1 [m Hm]. exists m.
eapply embedding_bottom1_eval_boolean; eauto.
Qed.

(** Contraposed version of the previous corollary. *)
Corollary embeds_bottom1_eval_boolean_contra :
  forall (HB1: HasBottom L1 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    ~(LA1 ==>[boolean_sem] LA2) ->
    ~(LA1 ==>[eval_sem] LA2).
Proof.
intros HB1 HLdef1.
apply contraposition.
eapply embeds_bottom1_eval_boolean; eassumption.
Qed.

(** Case where the input has a bottom default and the output has a top default. *)
Corollary embeds_bottom_top :
  forall (HB1: HasBottom L1 labLeq) (HT2: HasTop L2 labLeq),
    @Ldef _ _ LA1 =L bottom ->
    @Ldef _ _ LA2 =L top ->
    LA1 ==>[eval_sem] LA2 ->
    forall (l: L1), l =L bottom.
Proof.
intros HB1 HT2 HLdef1 HLdef2 [m Hm].
eapply embedding_bottom_top; eauto.
Qed.

(** Case where the input has a top default and the output has a bottom default. *)
Corollary embeds_top_bottom :
  forall (HT1: HasTop L1 labLeq) (HB2: HasBottom L2 labLeq),
    @Ldef _ _ LA1 =L top ->
    @Ldef _ _ LA2 =L bottom ->
    LA1 ==>[eval_sem] LA2 ->
    forall (l: L1), l =L top.
Proof.
intros HB1 HT2 HLdef1 HLdef2 [m Hm].
eapply embedding_top_bottom; eauto.
Qed.

End Embed_corollaries.
