Require Import LabelAlgebra.
Require Import UnitLattice.
Require Import Arith.

Inductive four := Bottom4 | LMiddle4 | UMiddle4 | Top4.

(* Injection into natural numbers. *)
Definition nat_inj t :=
  match t with
    | Bottom4 => 0
    | LMiddle4 => 1
    | UMiddle4 => 2
    | Top4 => 3
  end.

Definition four_leq (x y: four) := nat_inj x <= nat_inj y.

Instance four_PreOrder : PreOrder four_leq.
Proof.
unfold four_leq; unfold nat_inj; constructor.
* intros x. destruct x; auto.
* intros x y z Hxy Hyz.
  destruct x; destruct y; destruct z; auto.
Qed.

Instance four_JoinPreLattice : JoinPreLattice four four_leq :=
{ join x y :=
    if le_dec (nat_inj x) (nat_inj y)
    then y
    else x
}.
Proof.
* unfold four_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold four_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold four_leq. intros t1 t2 t3 H12 H23.
  destruct t1; destruct t2; destruct t3; auto.
Defined.

Instance four_MeetPreLattice : MeetPreLattice four four_leq :=
{ meet x y :=
    if le_dec (nat_inj x) (nat_inj y)
    then x
    else y
}.
Proof.
* unfold four_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold four_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold four_leq. intros t1 t2 t3 H12 H23.
  destruct t1; destruct t2; destruct t3; auto.
Defined.

Instance four_HasBottom : HasBottom four four_leq :=
{ bottom := Bottom4 }.
Proof.
unfold four_leq. auto.
Defined.

Instance four_HasTop : HasTop four four_leq :=
{ top := Top4 }.
Proof.
unfold four_leq. intros t; destruct t; compute; auto.
Defined.

Instance Four : LabelAlgebra unit four :=
{ Ldef := Bottom4
; labLeq := four_leq
; authLeq := unit_leq
; flowsTo a := four_leq
}.
Proof.
* split; intros x y; tauto.
* auto.
* auto.
* intros t1 t2 H x. tauto.
Defined.
