Require Import LabelAlgebra.
Require Import UnitLattice.

Inductive fourD := Bottom4D | LMiddle4D | RMiddle4D | Top4D.

Definition fourD_leq (x y: fourD) :=
  match x, y with
    | Bottom4D, _ => True
    | LMiddle4D, LMiddle4D | LMiddle4D, Top4D => True
    | LMiddle4D, _ => False
    | RMiddle4D, RMiddle4D | RMiddle4D, Top4D => True
    | RMiddle4D, _ => False
    | Top4D, Top4D => True
    | Top4D, _ => False
  end.

Instance fourD_PreOrder : PreOrder fourD_leq.
Proof.
unfold fourD_leq; constructor.
* intros x. destruct x; auto.
* intros x y z Hxy Hyz.
  destruct x; destruct y; destruct z; auto.
Qed.

Instance fourD_JoinPreLattice : JoinPreLattice fourD fourD_leq :=
{ join x y :=
    match x, y with
      | Top4D, _ | _, Top4D => Top4D
      | LMiddle4D, RMiddle4D | RMiddle4D, LMiddle4D => Top4D
      | LMiddle4D, _ | _, LMiddle4D => LMiddle4D
      | RMiddle4D, _ | _, RMiddle4D => RMiddle4D
      | Bottom4D, Bottom4D => Bottom4D
    end
}.
Proof.
* unfold fourD_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold fourD_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold fourD_leq. intros t1 t2 t3 H12 H23.
  destruct t1; destruct t2; destruct t3; auto.
Defined.

Instance fourD_MeetPreLattice : MeetPreLattice fourD fourD_leq :=
{ meet x y :=
    match x, y with
      | Bottom4D, _ | _, Bottom4D => Bottom4D
      | LMiddle4D, RMiddle4D | RMiddle4D, LMiddle4D => Bottom4D
      | LMiddle4D, _ | _, LMiddle4D => LMiddle4D
      | RMiddle4D, _ | _, RMiddle4D => RMiddle4D
      | Top4D, Top4D => Top4D
    end
}.
Proof.
* unfold fourD_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold fourD_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold fourD_leq. intros t1 t2 t3 H12 H23.
  destruct t1; destruct t2; destruct t3; auto.
Defined.

Instance fourD_HasBottom : HasBottom fourD fourD_leq :=
{ bottom := Bottom4D }.
Proof.
unfold fourD_leq. auto.
Defined.

Instance fourD_HasTop : HasTop fourD fourD_leq :=
{ top := Top4D }.
Proof.
unfold fourD_leq. intros t; destruct t; compute; auto.
Defined.

Instance FourD : LabelAlgebra unit fourD :=
{ Ldef := Bottom4D
; labLeq := fourD_leq
; authLeq := unit_leq
; flowsTo a := fourD_leq
}.
Proof.
* split; intros x y; tauto.
* auto.
* auto.
* intros t1 t2 H x. tauto.
Defined.
