Require Import PreLattice.

Set Implicit Arguments.

Definition pair_leq {A B: Type} (leqA: A -> A -> Prop) (leqB: B -> B -> Prop)
           (x1 x2: A * B): Prop :=
  leqA (fst x1) (fst x2) /\ leqB (snd x1) (snd x2).

Instance Pair_leq_PreOrder {A B: Type}
         (leqA: A -> A -> Prop) (leqB: B -> B -> Prop)
         {HA: PreOrder leqA} {HB: PreOrder leqB} :
  PreOrder (pair_leq leqA leqB).
Proof.
constructor.
* intros [x y]; split; simpl; reflexivity.
* intros [x1 y1] [x2 y2] [x3 y3]; unfold pair_leq; simpl; auto;
  intros [? ?] [? ?]; split.
  transitivity x2; trivial.
  transitivity y2; trivial.
Qed.

Instance Pair_HasBottom (A B: Type)
         (leqA: A -> A -> Prop) (leqB: B -> B -> Prop)
         {HA: HasBottom A leqA} {HB: HasBottom B leqB} :
  HasBottom (A * B) (pair_leq leqA leqB) :=
{ bottom := (bottom, bottom) }.
Proof.
intros [x y]; unfold pair_leq; split; simpl; apply bottom_minimum.
Defined.

Instance Pair_HasTop (A B: Type)
         (leqA: A -> A -> Prop) (leqB: B -> B -> Prop)
         {HA: HasTop A leqA} {HB: HasTop B leqB} :
  HasTop (A * B) (pair_leq leqA leqB) :=
{ top := (top, top) }.
Proof.
intros [x y]; unfold pair_leq; split; simpl; apply top_maximum.
Defined.

Instance Dual_JoinPreLattice (A B: Type)
         (leqA: A -> A -> Prop) (leqB: B -> B -> Prop)
         {PA: PreOrder leqA} {JA: JoinPreLattice A leqA}
         {PB: PreOrder leqB} {JB: JoinPreLattice B leqB} :
  JoinPreLattice (A * B) (pair_leq leqA leqB) :=
{ join x1 x2 := (join (fst x1) (fst x2), join (snd x1) (snd x2)) }.
Proof.
* intros [x1 y1] [x2 y2]; unfold pair_leq; simpl; split; apply join_ub1.
* intros [x1 y1] [x2 y2]; unfold pair_leq; simpl; split; apply join_ub2.
* intros [x1 y1] [x2 y2] [x3 y3]; unfold pair_leq; simpl;
  intros [? ?] [? ?]; split; apply join_lub; assumption.
Defined.

Instance Dual_MeetPreLattice (A B: Type)
         (leqA: A -> A -> Prop) (leqB: B -> B -> Prop)
         {PA: PreOrder leqA} {MA: MeetPreLattice A leqA}
         {PB: PreOrder leqB} {MB: MeetPreLattice B leqB} :
  MeetPreLattice (A * B) (pair_leq leqA leqB) :=
{ meet x1 x2 := (meet (fst x1) (fst x2), meet (snd x1) (snd x2)) }.
Proof.
* intros [x1 y1] [x2 y2]; unfold pair_leq; simpl; split; apply meet_lb1.
* intros [x1 y1] [x2 y2]; unfold pair_leq; simpl; split; apply meet_lb2.
* intros [x1 y1] [x2 y2] [x3 y3]; unfold pair_leq; simpl;
  intros [? ?] [? ?]; split; apply meet_glb; assumption.
Defined.
