Require Import MyList.
Require Import Arith.

(** * A principal is an abstract, opaque, name. *)
Module Type PRINCIPAL.
  Parameter t : Type.
  Parameter enum : nat -> t.
  Parameter enum_injective : forall (n1 n2: nat), enum n1 = enum n2 -> n1 = n2.
  Parameter enum_surjective : forall (p: t), { n : nat | p = enum n }.
  Parameter eq_dec : forall (p1 p2 : t), { p1 = p2 } + { p1 <> p2 }.
  Parameter fresh :
    forall l : list t, { p0 : t | list_forall (fun p => p <> p0) l }.
End PRINCIPAL.

Module Prin : PRINCIPAL.
  Definition t := nat.

  Definition enum (n: nat) := n.

  Lemma enum_injective : forall (n1 n2: nat), enum n1 = enum n2 -> n1 = n2.
  Proof.
    intros. assumption.
  Qed.

  Lemma enum_surjective : forall (p: t), { n : nat | p = enum n }.
  Proof.
    intros n. exists n. reflexivity.
  Qed.

  Definition eq_dec := eq_nat_dec.

  Definition fresh_strong (l: list t) :
    { p0 : t | list_forall (fun p => p < p0) l }.
  Proof.
    induction l.
    * exists 0. exact I.
    * destruct IHl as [p H].
      exists (S (max a p)). split.
      - apply le_lt_n_Sm. auto with arith.
      - apply list_forall_inclusion with (P1 := fun p0 : nat => p0 < p).
        + intros b Hb. apply le_lt_n_Sm. transitivity p; auto with arith.
        + assumption.
  Defined.

  Definition fresh (l: list t) :
    { p0 : t | list_forall (fun p => p <> p0) l }.
  Proof.
    pose proof (fresh_strong l) as [p H].
    exists p.
    apply list_forall_inclusion with (P1 := fun p0 : nat => p0 < p); auto.
  Defined.

End Prin.
