Require Import MyTactics.
Require Import MyList.
Require Import Principal.
Require Import Ensembles.
Require Import Finite_sets_facts.
Require Import PreLattice.
Require Import SetLattice.
Require Import SetFacts.
Require Import FinSetLattice.
Require Import CFinSetLattice.
Require Import Integers.
Require Import MyProofIrrelevance.

(** * Facts about principals. *)
Lemma Prin_inhabited : inhabited Prin.t.
Proof.
destruct (Prin.fresh nil) as [p _].
constructor. exact p.
Qed.

Lemma Finite_set_list_iff :
  forall A (S: FinSet A),
    exists l, forall x,
                In A (proj1_sig S) x <-> exists m, atIndex l m = Some x.
Proof.
intros A [S H]. simpl. induction H.
* exists nil. intro x. split.
  + intro H. destruct H.
  + intros [m H]. auto.
* destruct IHFinite as [l Hl]. exists (x :: l).
  intros y. split; intro Hy.
  + apply Union_inv in Hy. destruct Hy as [Hy | Hy].
    - specialize (Hl y). destruct Hl as [Hl _].
      destruct (Hl Hy) as [m Hm].
      exists (S m). assumption.
    - apply Singleton_inv in Hy. subst x. exists 0. reflexivity.
  + destruct Hy as [m Hm]. destruct m; simpl in Hm.
    - inversion Hm. subst. auto with sets.
    - specialize (Hl y). destruct Hl as [_ Hl].
      specialize (Hl (ex_intro _ _ Hm)). auto with sets.
Qed.

Lemma Prin_not_finite : ~ Finite _ (Full_set Prin.t).
Proof.
intro H.
assert (exists l, forall (p: Prin.t), exists m, atIndex l m = Some p) as [l Hl].
{ destruct (Finite_set_list_iff _ (exist _ _ H)) as [l Hl]. exists l.
  intro p. apply Hl. constructor.
}
destruct (Prin.fresh l) as [p Hp].
specialize (Hl p). destruct Hl as [m Hm].
pose proof (list_forall_atIndex _ _ _ _ Hp Hm) as H'.
simpl in H'. congruence.
Qed.

Lemma Prin_FinSet_not_bounded :
  ~ exists (s : FinSet Prin.t), forall s', FinSet_included s' s.
Proof.
intros [s Hincl].
pose proof (Finite_set_list_iff _ s) as Hfin.
destruct Hfin as [l Hl].
destruct (Prin.fresh l) as [p Hp].
specialize (Hincl (FinSet_singleton p)).
assert (In _ (proj1_sig s) p) as H.
{ destruct s as [s Hs].
  unfold FinSet_singleton in Hincl.
  unfold FinSet_included in Hincl.
  simpl in *. auto with sets.
}
apply Hl in H. destruct H as [m Hm].
pose proof (list_forall_atIndex _ _ _ _ Hp Hm).
simpl in H. congruence.
Qed.

Lemma Prin_chain :
  { f : nat -> Prin.t | forall n1 n2, f n1 = f n2 -> n1 = n2 }.
Proof.
pose (g := fix g (n: nat) : (Prin.t * list Prin.t) :=
        match n with
          | O => (proj1_sig (Prin.fresh nil), nil)
          | S n =>
            let (p, l) := g n in
            let l' := p :: l in
            let p' := proj1_sig (Prin.fresh l') in
            (p', l')
        end
     ).
assert (forall n, length (snd (g n)) = n) as Hmlength.
{ intro n. induction n; simpl; auto.
  case_eq (g n). intros p l Heq. simpl.
  rewrite Heq in IHn. simpl in IHn. congruence.
}
assert (forall n m,
          n > m ->
          atIndex (snd (g n)) (length (snd (g n)) - S m) = Some (fst (g m)))
  as Hgm.
{ intros n. induction n; intros m Hm; auto.
  simpl. case_eq (g n). intros p l Heq. simpl.
  assert (length l = n).
  { transitivity (length (snd (g n))); auto. rewrite Heq; reflexivity. }
  case_eq (length l - m).
  * intro Heqm. simpl. assert (m >= length l) by omega.
    replace m with n; auto.
    rewrite Heq; reflexivity.
  * intros k Heqk. simpl. rewrite <- IHn; auto.
    rewrite Heq. simpl. f_equal; auto.
}
assert (forall x y : Prin.t, x <> y -> y <> x) as F by congruence.
assert (forall n m, m < n -> fst (g n) <> fst (g m)) as Hg.
{ intro n; induction n; intros m Hmn.
  * destruct m; auto.
  * simpl. case_eq (g n). intros p l Heq. simpl.
    destruct (Prin.fresh (p :: l)) as [p' Hp']. simpl.
    pose proof
         (list_forall_atIndex _ _ (length (snd (g (S n))) - S m) (fst (g m)) Hp').
    apply F. apply H.
    replace (p :: l) with (snd (g (S n))).
    + apply Hgm; auto.
    + simpl. rewrite Heq. reflexivity.
}
pose (f n := fst (g n)).
exists f. unfold f. clear f.
intros n1 n2 H.
assert (n1 = n2 \/ n1 < n2 \/ n2 < n1) as [Hn | [Hn | Hn]] by omega; auto.
* exfalso. specialize (Hg _ _ Hn). auto.
* exfalso. specialize (Hg _ _ Hn). auto.
Qed.

Lemma Prin_FinSet_increasing_chain :
  { f : nat -> FinSet Prin.t |
    (forall n, cardinal _ (proj1_sig (f n)) n)
    /\ forall n m, n < m ->
                   FinSet_included (f n) (f m)
                   /\ ~ FinSet_included (f m) (f n)
  }.
Proof.
pose (f := fix f (n: nat) : FinSet Prin.t :=
        match n with
          | O => FinSet_empty _
          | S n => FinSet_add (proj1_sig Prin_chain n) (f n)
        end).
assert (forall m n, m <= n -> ~In _ (proj1_sig (f m)) (proj1_sig Prin_chain n))
  as HfnotIn.
{ intro m. induction m; intros.
  * simpl. auto with sets.
  * simpl. intro HIn.
    apply Union_inv in HIn. destruct HIn as [HIn | HIn].
    + apply Union_inv in HIn. destruct HIn as [[] | HIn].
      inversion HIn; subst.
      destruct Prin_chain as [U HU]. simpl in *.
      specialize (HU _ _ H1). auto.
    + apply (IHm n); auto.
}
assert (forall n m, m < n -> In _ (proj1_sig (f n)) (proj1_sig Prin_chain m))
  as HfIn.
{ intro n. induction n; intros.
  * simpl. auto with sets.
  * simpl. assert (m = n \/ m < n) as [? | ?] by omega.
    + subst. auto with sets.
    + auto with sets.
}
exists f. split.
* intro n. induction n.
  + simpl. auto with sets.
  + assert (cardinal _
              (Add Prin.t (proj1_sig (f n)) (proj1_sig Prin_chain n))
              (S n)) by auto with sets.
    assert (exists m, cardinal Prin.t (proj1_sig (f (S n))) m) as [m Hm].
    { apply finite_cardinal.
      case_eq (f (S n)). intros s Hf Heq. simpl. assumption. }
    assert (m <= S n).
    { apply (incl_card_le _ _ _ _ _ Hm H).
      clear. simpl. auto with sets.
    }
    assert (S n <= m).
    { apply (incl_card_le _ _ _ _ _ H Hm).
      clear. simpl. unfold Add. auto with sets.
    }
    assert (m = S n) by auto.
    congruence.
* { intros n m Hnm. split.
    * generalize dependent n. induction m; intros n Hnm.
      + destruct n; auto.
      + destruct n.
        - simpl. apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
        - simpl. intros p Hp.
          unfold FinSet_add, FinSet_singleton, FinSet_union in Hp. simpl in Hp.
          unfold FinSet_add, FinSet_singleton, FinSet_union. simpl.
          { apply Union_inv in Hp. destruct Hp as [Hp | Hp].
            * apply Union_inv in Hp. destruct Hp as [[] | Hp].
              inversion Hp; subst. auto with sets.
            * right. apply (IHm n); auto.
          }
    * intro H. specialize (H (proj1_sig Prin_chain n)).
      specialize (H (HfIn m n Hnm)).
      apply (HfnotIn n n); auto.
  }
Qed.

(** * A non-existence result about functions from finite sets of principals to finite sets of principals. *)
Lemma Finite_subsets2 (A: Type) :
  forall S : FinSet A,
    Finite (FinSet A) (fun S' => FinSet_included S' S).
Proof.
intros [S HS].
unfold FinSet_included. simpl.
apply (Pigeonhole_ter _ _ _ (fun S : FinSet A => proj1_sig S) 0).
* intros S1 S2 Heq. destruct S1. destruct S2. simpl in Heq.
  apply subset_eq_compat2. assumption.
* pose proof (Finite_subsets _ S HS) as H.
  apply (Finite_downward_closed _ _ H).
  intros S' HS'. apply Im_inv in HS'.
  destruct HS' as [S'' [HS'' H2]]. subst S'. assumption.
Qed.

Lemma fprinset_to_fprinset_inj_bounded_absurd :
  forall z (g : FinSet Prin.t -> FinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x, FinSet_included (g x) z) ->
    False.
Proof.
intros z g Hinj Hbounded.
apply
  (Pigeonhole_bis _ _
     (Im _ _ Integers (proj1_sig Prin_FinSet_increasing_chain)) g).
* clear z g Hinj Hbounded. intro HFinite.
  apply (Pigeonhole_bis _ _ Integers (proj1_sig Prin_FinSet_increasing_chain));
    trivial.
  + exact Integers_infinite.
  + destruct Prin_FinSet_increasing_chain as [U [HU HU']]. simpl in *.
    intros n m H.
    assert (n = m \/ n < m \/ m < n) as [Hnm | [Hnm | Hnm]] by omega; trivial.
    - exfalso. specialize (HU' _ _ Hnm). destruct HU' as [_ HU']. apply HU'.
      rewrite H. reflexivity.
    - exfalso. specialize (HU' _ _ Hnm). destruct HU' as [_ HU']. apply HU'.
      rewrite H. reflexivity.
* apply (Finite_downward_closed _ (fun S => FinSet_included S z)).
  + apply Finite_subsets2.
  + intros S HS. unfold In.
    apply Im_inv in HS. destruct HS as [S' [HS' ?]]; subst S.
    auto.
* intros x y H.
  assert (g x == g y) as Heq. { rewrite H. reflexivity. }
  specialize (Hinj x y Heq). clear H Heq z g Hbounded.
  assert (proj1_sig x = proj1_sig y) as Heq'.
  { destruct x as [x Hx]. destruct y as [y Hy]. simpl in *.
    unfold FinSet_included in Hinj. simpl in Hinj.
    auto with sets.
  }
  clear Hinj.
  destruct x as [x Hx]. destruct y as [y Hy].
  simpl in *. apply subset_eq_compat2. assumption.
Qed.

(** * A non-existence result about functions from sets of principals to finite sets of principals. *)
Lemma prinset_to_fprinset_inj_bounded_absurd :
  forall z (g : Ensemble Prin.t -> FinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x, FinSet_included (g x) z) ->
    False.
Proof.
intros z g Hinj Hbounded.
pose (f (S : FinSet Prin.t) := g (proj1_sig S)).
apply (fprinset_to_fprinset_inj_bounded_absurd z f).
* intros x y Heq. unfold f in Heq. apply Hinj in Heq.
  destruct x as [x Hx]. destruct y as [y Hy]. assumption.
* intro x. specialize (Hbounded (proj1_sig x)). assumption.
Qed.

(** * A non-existence result about functions from "completed" sets of principals to finite sets of principals. *)
Lemma cprinset_to_fprinset_inj_bounded_absurd :
  forall z (g : CFinSet Prin.t -> FinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x, FinSet_included (g x) z) ->
    False.
Proof.
intros z g Hinj Hbounded.
pose (f (S : FinSet Prin.t) := g (CFinSet_intro S)).
apply (fprinset_to_fprinset_inj_bounded_absurd z f).
* intros x y Heq. unfold f in Heq. apply Hinj in Heq. assumption.
* intro x. specialize (Hbounded (CFinSet_intro x)). assumption.
Qed.

(** * About [shift] *)
Definition shift (p: Prin.t) :=
  Prin.enum (S (proj1_sig (Prin.enum_surjective p))).

Lemma shift_injective: injective _ _ shift.
Proof.
intros p1 p2 H. unfold shift in H. apply Prin.enum_injective in H.
destruct (Prin.enum_surjective p1) as [n1 H1].
destruct (Prin.enum_surjective p2) as [n2 H2].
simpl in H. congruence.
Qed.

Lemma shift_not_0 (p: Prin.t) : shift p <> Prin.enum 0.
Proof.
unfold shift. destruct (Prin.enum_surjective p) as [n Hn]. subst p.
simpl. intro H. apply Prin.enum_injective in H. auto.
Qed.

Definition shift_set (S: Ensemble Prin.t) : Ensemble Prin.t :=
  Add _ (Im _ _ S shift) (Prin.enum 0).

Lemma shift_set_not_singleton_0 :
  forall p, ~In _ (Singleton _ (Prin.enum 0)) (shift p).
Proof.
intros p H. apply Singleton_inv in H. unfold shift in H.
apply Principal.Prin.enum_injective in H. auto.
Qed.

Lemma shift_set_Add :
  forall p (S: Ensemble Principal.Prin.t),
    In _ S p ->
    In _ (Add _ (Im _ _ S shift) (Prin.enum 0)) (shift p).
Proof.
intros p s H. left. auto with sets.
Qed.

Lemma shift_set_injective : injective _ _ shift_set.
Proof.
intros s1 s2 H. unfold shift_set in H. apply Extensionality_Ensembles.
split; intros p Hp; destruct (Prin.enum_surjective p) as [n Hn]; subst p;
apply shift_set_Add in Hp.
* rewrite H in Hp. apply Union_inv in Hp. destruct Hp as [Hp | Hp].
  + apply Im_inv in Hp. destruct Hp as [m [Hm Heq]].
    apply shift_injective in Heq. congruence.
  + exfalso. eapply shift_set_not_singleton_0; eassumption.
* rewrite <- H in Hp. apply Union_inv in Hp. destruct Hp as [Hp | Hp].
  + apply Im_inv in Hp. destruct Hp as [m [Hm Heq]].
    apply shift_injective in Heq. congruence.
  + exfalso. eapply shift_set_not_singleton_0; eassumption.
Qed.

Lemma shift_set_finite :
  forall (S: Ensemble Prin.t),
    Finite _ S ->
    Finite _ (shift_set S).
Proof.
intros S HS. unfold shift_set.
apply Add_preserves_Finite. apply finite_image. assumption.
Qed.

Definition shift_FinSet (S: FinSet Prin.t) : FinSet Prin.t :=
  exist _ (shift_set (proj1_sig S))
        (shift_set_finite (proj1_sig S) (proj2_sig S)).

Lemma shift_FinSet_injective : injective _ _ shift_FinSet.
Proof.
intros [s1 H1] [s2 H2] H.
unfold shift_FinSet in H. simpl in H.
apply subset_eq_compat2. apply shift_set_injective.
congruence.
Qed.

(** * A non-existence result about functions from finite sets of principals to "completed" finite sets of principals. *)
Lemma fprinset_to_cprinset_inj_quasi_antitone_full :
  forall z (g : FinSet Prin.t -> CFinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x y,
       FinSet_included x y ->
       CFinSet_included (g y) (CFinSet_union (g x) (CFinSet_intro z))) ->
    g (FinSet_empty Principal.Prin.t) = CFinSet_full Principal.Prin.t ->
    False.
Proof.
intros z g Hinj Hmon Heq.
assert (forall S : FinSet Principal.Prin.t,
          { S' | g (shift_FinSet S) = CFinSet_intro S' }) as Hfin.
{ intro S. case_eq (g (shift_FinSet S)).
  * intros S' Heq'. exists S'. reflexivity.
  * intro Heq'. exfalso.
    assert (equiv (g (shift_FinSet S)) (g (FinSet_empty _))) as Hmeq.
    { rewrite Heq. rewrite Heq'. reflexivity. }
    apply Hinj in Hmeq. destruct Hmeq as [Hmeq _].
    assert (In _ (proj1_sig (shift_FinSet S)) (Prin.enum 0)) as Hin.
    { right. auto with sets. }
    specialize (Hmeq (Principal.Prin.enum 0) Hin).
    destruct Hmeq.
}
pose (f S := proj1_sig (Hfin S)).
apply (fprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (f (FinSet_empty _)) z) f).
+ intros S1 S2 H12. unfold f in H12.
  destruct (Hfin S1) as [S1' H1']. destruct (Hfin S2) as [S2' H2'].
  simpl proj1_sig in H12.
  assert (equiv (g (shift_FinSet S1)) (g (shift_FinSet S2))) as Hm12.
  { rewrite H1'. rewrite H2'. simpl. assumption. }
  apply Hinj in Hm12.
  assert (shift_FinSet S1 = shift_FinSet S2) as Hm12'.
  { case_eq (shift_FinSet S1). intros S1'' HS1'' Heq1.
    case_eq (shift_FinSet S2). intros S2'' HS2'' Heq2.
      rewrite Heq1 in Hm12. rewrite Heq2 in Hm12.
      simpl in Hm12. unfold FinSet_included in Hm12. simpl in Hm12.
      apply subset_eq_compat2.
      apply Extensionality_Ensembles. assumption.
  }
  apply shift_FinSet_injective in Hm12'. rewrite Hm12'. reflexivity.
+ intros S.
  specialize (Hmon (shift_FinSet (FinSet_empty _)) (shift_FinSet S)).
  assert (FinSet_included (shift_FinSet (FinSet_empty _)) (shift_FinSet S)) as H.
  { unfold shift_FinSet. simpl.
    intros p Hp. simpl in Hp. apply Union_inv in Hp. destruct Hp as [Hp | Hp].
    * apply Im_inv in Hp. destruct Hp as [p' [Hp' ?]]. destruct Hp'.
    * apply Singleton_inv in Hp. subst p. unfold shift_set. simpl. auto with sets.
  }
  specialize (Hmon H). clear H.
  unfold f.
  case_eq (Hfin S). intros S' HS' Heq'. rewrite HS' in Hmon. simpl proj1_sig.
  case_eq (Hfin (FinSet_empty _)). intros S0 HS0 Heq0.
  rewrite HS0 in Hmon. simpl proj1_sig.
  assumption.
Qed.

Lemma fprinset_to_cprinset_inj_quasi_antitone :
  forall z (g : FinSet Prin.t -> CFinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x y,
       FinSet_included x y ->
       CFinSet_included (g y) (CFinSet_union (g x) (CFinSet_intro z))) ->
    False.
Proof.
intros z g Hinj Hmon.
case_eq (g (FinSet_empty _)).
* intros S Heq.
  assert (forall S, { S' | g S = CFinSet_intro S' }) as Hfin.
  { intro S'. case_eq (g S').
    * intros S'' Heq'. exists S''. reflexivity.
    * intro Heq'. exfalso.
      specialize (Hmon (FinSet_empty _) S').
      rewrite Heq in Hmon. rewrite Heq' in Hmon. simpl in Hmon.
      apply Hmon.
      apply (@bottom_minimum _ _ (@FinSet_included_HasBottom _)).
  }
  pose (f S := proj1_sig (Hfin S)).
  apply (fprinset_to_fprinset_inj_bounded_absurd (FinSet_union S z) f).
  + intros x y H. apply Hinj.
    unfold f in H.
    destruct (Hfin x) as [X HX]. destruct (Hfin y) as [Y HY].
    simpl in H. rewrite HX. rewrite HY. simpl. assumption.
  + intro x. specialize (Hmon (FinSet_empty _) x).
    rewrite Heq in Hmon. unfold f. destruct (Hfin x) as [X HX]. simpl.
    rewrite HX in Hmon. apply Hmon.
    apply (@bottom_minimum _ _ (FinSet_included_HasBottom _)).
* intro Heq.
  apply (fprinset_to_cprinset_inj_quasi_antitone_full z g); auto.
Qed.

(** * A non-existence result about functions from "completed" sets of principals to "completed" finite sets of principals. *)
Lemma cprinset_to_cprinset_inj_bounded_absurd :
  forall z (g : CFinSet Prin.t -> CFinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x, CFinSet_included (g x) (CFinSet_intro z)) ->
    False.
Proof.
intros z g Hinj Hbounded.
case_eq (g (CFinSet_full _)).
* intros S0 Heq.
  assert (forall S : CFinSet Prin.t,
            { S' : FinSet Principal.Prin.t | equiv (g S) (CFinSet_intro S') })
    as Hfin.
  { intro S. case_eq (g S).
    * intros S' Heq'. exists S'. reflexivity.
    * intro Heq'. exfalso.
      specialize (Hbounded S). rewrite Heq' in Hbounded. assumption.
  }
  pose (f S := proj1_sig (Hfin (CFinSet_intro S))).
  apply (fprinset_to_fprinset_inj_bounded_absurd z f).
  + intros S1 S2 HfEq. unfold f in HfEq.
    destruct (Hfin (CFinSet_intro S1)) as [S1' HS1'].
    destruct (Hfin (CFinSet_intro S2)) as [S2' HS2'].
    simpl proj1_sig in HfEq.
    assert (equiv (g (CFinSet_intro S1)) (g (CFinSet_intro S2))) as Hf12.
    { rewrite HS1'. rewrite HS2'. assumption. }
    apply Hinj in Hf12. assumption.
  + intro S. unfold f.
    destruct (Hfin (CFinSet_intro S)) as [S' HS']. simpl.
    specialize (Hbounded (CFinSet_intro S)).
    case_eq (g (CFinSet_intro S)).
    - intros S'' Heq''. rewrite Heq'' in HS'. rewrite Heq'' in Hbounded.
      simpl in Hbounded. destruct HS' as [_ HS']. simpl in HS'.
      transitivity S''; assumption.
    - intro Heq''. exfalso. rewrite Heq'' in HS'. simpl in HS'. tauto.
* intro Heq.
  specialize (Hbounded (CFinSet_full _)).
  rewrite Heq in Hbounded. assumption.
Qed.

Lemma cprinset_to_cprinset_inj_quasi_antitone :
  forall z (g : CFinSet Prin.t -> CFinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x y,
       CFinSet_included x y ->
       CFinSet_included (g y) (CFinSet_union (g x) (CFinSet_intro z))) ->
    False.
Proof.
intros z g Hinj Hmon.
pose (f (S: FinSet Principal.Prin.t) := g (CFinSet_intro S)).
apply (fprinset_to_cprinset_inj_quasi_antitone z f).
* intros x y H. specialize (Hinj _ _ H). assumption.
* intros x y H. apply Hmon. assumption.
Qed.

(** * A non-existence result about functions from sets of principals to "completed" finite sets of principals. *)
Lemma prinset_to_cprinset_inj_bounded_absurd :
  forall z (g : Ensemble Prin.t -> CFinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x, CFinSet_included (g x) (CFinSet_intro z)) ->
    False.
Proof.
intros z g Hinj Hbounded.
pose (f S :=
        match S with
          | CFinSet_intro s => g (proj1_sig s)
          | CFinSet_full => g (Full_set _)
        end).
apply (cprinset_to_cprinset_inj_bounded_absurd z f).
* intros [ [x Hx] |] [ [y Hy] |] H; simpl in H.
  + apply Hinj. assumption.
  + exfalso. specialize (Hinj _ _ H). simpl in Hinj. destruct Hinj as [_ HX].
    apply Finite_downward_closed in HX; auto using Prin_not_finite.
  + exfalso. specialize (Hinj _ _ H). simpl in Hinj. destruct Hinj as [HY _].
    apply Finite_downward_closed in HY; auto using Prin_not_finite.
  + simpl. tauto.
* intros [ S | ]; simpl; auto.
Qed.

Lemma prinset_to_cprinset_inj_quasi_monotone_full :
  forall z (g : Ensemble Prin.t -> CFinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x y,
       Included _ x y ->
       CFinSet_included (g x) (CFinSet_union (g y) (CFinSet_intro z))) ->
    g (Full_set Principal.Prin.t) = CFinSet_full Principal.Prin.t ->
    False.
Proof.
intros z g Hinj Hmon Heq.
pose (c (S: FinSet Principal.Prin.t) :=
        Ensembles.Complement _ (shift_set (proj1_sig S))).
assert (forall S : FinSet Principal.Prin.t,
          { S' | g (c S) = CFinSet_intro S' }) as Hfin.
{ intro S. case_eq (g (c S)).
  * intros S' Heq'. exists S'. reflexivity.
  * intro Heq'. exfalso.
    assert (equiv (g (c S)) (g (Full_set _))) as Hmeq.
    { rewrite Heq. rewrite Heq'. reflexivity. }
    apply Hinj in Hmeq. destruct Hmeq as [_ Hmeq].
    specialize (Hmeq (Principal.Prin.enum 0) (Full_intro _ _)).
    apply Hmeq. clear Hmeq. right. auto with sets.
}
pose (f S := proj1_sig (Hfin S)).
apply (fprinset_to_fprinset_inj_bounded_absurd
         (FinSet_union (f (FinSet_empty _)) z) f).
+ intros S1 S2 H12. unfold f in H12.
  destruct (Hfin S1) as [S1' H1']. destruct (Hfin S2) as [S2' H2'].
  simpl proj1_sig in H12.
  assert (equiv (g (c S1)) (g (c S2))) as Hm12.
  { rewrite H1'. rewrite H2'. simpl. assumption. }
  apply Hinj in Hm12.
  assert (c S1 = c S2) as Hm12'.
  { apply Extensionality_Ensembles. assumption. }
  unfold c in Hm12'. apply Complement_injective in Hm12'.
  apply shift_set_injective in Hm12'.
  destruct S1 as [S1 HS1]. destruct S2 as [S2 HS2]. simpl in Hm12'.
  simpl. unfold FinSet_included. simpl. rewrite Hm12'. auto with sets.
+ intros S.
  specialize (Hmon (c S) (c (FinSet_empty _))).
  assert (Included _ (c S) (c (FinSet_empty _))) as H.
  { unfold c. rewrite <- Complement_Included.
    intros p Hp. apply Union_inv in Hp. destruct Hp as [Hp | Hp].
    * apply Im_inv in Hp. destruct Hp as [p' [Hp' ?]]. destruct Hp'.
    * apply Singleton_inv in Hp. subst p. unfold shift_set. auto with sets.
  }
  specialize (Hmon H). clear H.
  unfold f.
  case_eq (Hfin S). intros S' HS' Heq'. rewrite HS' in Hmon. simpl proj1_sig.
  case_eq (Hfin (FinSet_empty _)). intros S0 HS0 Heq0.
  rewrite HS0 in Hmon. simpl proj1_sig.
  assumption.
Qed.

Lemma prinset_to_cprinset_inj_quasi_monotone :
  forall z (g : Ensemble Prin.t -> CFinSet Prin.t),
    (forall x y, equiv (g x) (g y) -> equiv x y) ->
    (forall x y,
       Included _ x y ->
       CFinSet_included (g x) (CFinSet_union (g y) (CFinSet_intro z))) ->
    False.
Proof.
intros z g Hinj Hmon.
case_eq (g (Full_set _)).
* intros S Heq.
  apply (prinset_to_cprinset_inj_bounded_absurd (FinSet_union S z) g); auto.
  + intro x. specialize (Hmon x (Full_set _)).
    rewrite Heq in Hmon. apply Hmon. auto using Full_intro with sets.
* intro Heq.
  apply (prinset_to_cprinset_inj_quasi_monotone_full z g); auto.
Qed.
