Require Import MyTactics.
Require Import LabelAlgebra.
Require Import FinSetLattice.
Require Import Principal.

Set Implicit Arguments.

Definition taint_flowsTo (a s1 s2 : FinSet Prin.t) : Prop :=
  FinSet_included s1 (FinSet_union s2 a).

Instance Taint_flowsTo_PreOrder (a: FinSet Prin.t) : PreOrder (taint_flowsTo a).
Proof.
unfold taint_flowsTo. constructor.
* intro s. transitivity (join s a); try reflexivity.
  apply join_ub1.
* intros s1 s2 s3 H12 H23.
  transitivity (join s3 a); try reflexivity.
  transitivity (join s2 a); try assumption.
  apply join_lub. assumption. apply join_ub2.
Qed.

Instance Taint_flowsTo_HasBottom (a: FinSet Prin.t):
  HasBottom _ (taint_flowsTo a) :=
{ bottom := FinSet_empty _ }.
Proof.
destruct a as [a Ha]. intros [s Hs].
unfold taint_flowsTo. unfold FinSet_included. simpl.
auto with sets.
Defined.

Instance Taint_JoinPrelattice (a: FinSet Prin.t) :
  JoinPreLattice (FinSet Prin.t) (taint_flowsTo a) :=
{ join := FinSet_union }.
Proof.
* intros s1 s2. unfold taint_flowsTo.
  transitivity (join s1 s2).
  apply join_ub1.
  transitivity (join (join s1 s2) a); try reflexivity.
  apply join_ub1.
* intros s1 s2. unfold taint_flowsTo.
  transitivity (join s1 s2).
  apply join_ub2.
  transitivity (join (join s1 s2) a); try reflexivity.
  apply join_ub1.
* intros s s1 s2 H1 H2. unfold taint_flowsTo in *.
  transitivity (join s1 s2); try reflexivity.
  apply join_lub; assumption.
Defined.

Instance Taint_MeetPrelattice (a: FinSet Prin.t) :
  MeetPreLattice (FinSet Prin.t) (taint_flowsTo a) :=
{ meet := FinSet_intersection }.
Proof.
* intros s1 s2. unfold taint_flowsTo.
  transitivity s1.
  transitivity (meet s1 s2). reflexivity. apply meet_lb1.
  transitivity (@join _ FinSet_included _ _ s1 a).
  apply join_ub1. reflexivity.
* intros s1 s2. unfold taint_flowsTo.
  transitivity s2.
  transitivity (meet s1 s2). reflexivity. apply meet_lb2.
  transitivity (@join _ FinSet_included _ _ s2 a).
  apply join_ub1. reflexivity.
* intros s s1 s2 H1 H2. unfold taint_flowsTo in *.
  transitivity (meet (@join _ FinSet_included _ _ s1 a)
                     (@join _ FinSet_included _ _ s2 a)).
  apply meet_glb; assumption.
  apply FinSet_distrib1.
Defined.

Instance Taint : LabelAlgebra (FinSet Prin.t) (FinSet Prin.t) :=
{ Ldef := FinSet_empty Prin.t
; labLeq := FinSet_included
; authLeq := FinSet_included
; flowsTo := taint_flowsTo
}.
Proof.
* split; unfold taint_flowsTo; intros s1 s2 H.
  + rewrite H. apply (@join_lub _ _ _ (FinSet_JoinPreLattice _)).
    - reflexivity.
    - apply bottom_minimum.
  + rewrite H. apply (@join_ub1 _ _ _ (FinSet_JoinPreLattice _)).
* auto.
* auto.
* intros a1 a2 H12 s s' H.
  unfold taint_flowsTo in *.
  transitivity (FinSet_union s' a1); auto.
  transitivity (@join _ FinSet_included _ _ s' a1); try reflexivity.
  transitivity (@join _ FinSet_included _ _ s' a2); try reflexivity.
  rewrite H12. reflexivity.
Defined.
