Require Import LabelAlgebra.
Require Import UnitLattice.
Require Import Arith.

Inductive three := Bottom3 | Middle3 | Top3.

(* Injection into natural numbers. *)
Definition nat_inj t :=
  match t with
    | Bottom3 => 0
    | Middle3 => 1
    | Top3 => 2
  end.

Definition three_leq (x y: three) := nat_inj x <= nat_inj y.

Instance three_PreOrder : PreOrder three_leq.
Proof.
unfold three_leq; unfold nat_inj; constructor.
* intros x. destruct x; auto.
* intros x y z Hxy Hyz.
  destruct x; destruct y; destruct z; auto.
Qed.

Instance three_JoinPreLattice : JoinPreLattice three three_leq :=
{ join x y :=
    if le_dec (nat_inj x) (nat_inj y)
    then y
    else x
}.
Proof.
* unfold three_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold three_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold three_leq. intros t1 t2 t3 H12 H23.
  destruct t1; destruct t2; destruct t3; auto.
Defined.

Instance three_MeetPreLattice : MeetPreLattice three three_leq :=
{ meet x y :=
    if le_dec (nat_inj x) (nat_inj y)
    then x
    else y
}.
Proof.
* unfold three_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold three_leq. intros t1 t2; destruct t1; destruct t2; auto.
* unfold three_leq. intros t1 t2 t3 H12 H23.
  destruct t1; destruct t2; destruct t3; auto.
Defined.

Instance three_HasBottom : HasBottom three three_leq :=
{ bottom := Bottom3 }.
Proof.
unfold three_leq. auto.
Defined.

Instance three_HasTop : HasTop three three_leq :=
{ top := Top3 }.
Proof.
unfold three_leq. intros t; destruct t; compute; auto.
Defined.

Instance Three : LabelAlgebra unit three :=
{ Ldef := Bottom3
; labLeq := three_leq
; authLeq := unit_leq
; flowsTo a := three_leq
}.
Proof.
* split; intros x y; tauto.
* auto.
* auto.
* intros t1 t2 H x. tauto.
Defined.
