Require Import LabelAlgebra.
Require Export Two.

Definition twoA_leq (a x y: two) :=
  match a with
    | Bottom2 => two_leq x y
    | Top2 => True
  end.

Instance twoA_PreOrder (a: two) : PreOrder (twoA_leq a).
Proof.
unfold twoA_leq; destruct a.
* exact two_PreOrder.
* constructor; auto.
Qed.

Instance twoA_JoinPreLattice (a: two) : JoinPreLattice two (twoA_leq a) :=
{ join := join }.
Proof.
* unfold twoA_leq; destruct a; auto using join_ub1.
* unfold twoA_leq; destruct a; auto using join_ub2.
* unfold twoA_leq; destruct a; auto using join_lub.
Defined.

Instance twoA_MeetPreLattice (a: two) : MeetPreLattice two (twoA_leq a) :=
{ meet := meet }.
Proof.
* unfold twoA_leq; destruct a; auto using meet_lb1.
* unfold twoA_leq; destruct a; auto using meet_lb2.
* unfold twoA_leq; destruct a; auto using meet_glb.
Defined.

Instance twoA_HasBottom (a: two) : HasBottom two (twoA_leq a) :=
{ bottom := bottom }.
Proof.
unfold twoA_leq. destruct a; auto using bottom_minimum.
Defined.

Instance twoA_HasTop (a: two) : HasTop two (twoA_leq a) :=
{ top := top }.
Proof.
unfold twoA_leq. destruct a; auto using top_maximum.
Defined.

Instance TwoA : LabelAlgebra two two :=
{ Ldef := Top2
; labLeq := two_leq
; authLeq := two_leq
; flowsTo := twoA_leq
}.
Proof.
* split; intros x y; tauto.
* auto.
* auto.
* intros t1 t2 H x. destruct t1; destruct t2; simpl in *; tauto.
Defined.
