Require Import MyTactics.
Require Import LabelAlgebra.
Require Import Principal.
Require Import Ensembles.
Require Import SetLattice.
Require Import SetFacts.

Set Implicit Arguments.

Definition taint_flowsTo (a s1 s2 : Ensemble Prin.t) : Prop :=
  Included _ s1 (Union _ s2 a).

Instance Taint_flowsTo_PreOrder (a: Ensemble Prin.t) : PreOrder (taint_flowsTo a).
Proof.
unfold taint_flowsTo. constructor.
* intro s. transitivity (join s a); try reflexivity.
  apply join_ub1.
* intros s1 s2 s3 H12 H23.
  transitivity (join s3 a); try reflexivity.
  transitivity (join s2 a); try assumption.
  apply join_lub. assumption. apply join_ub2.
Qed.

Instance Taint_flowsTo_HasTop (a: Ensemble Prin.t):
  HasTop _ (taint_flowsTo a) :=
{ top := Full_set _ }.
Proof.
unfold taint_flowsTo. intros t p H. left. apply Full_intro.
Defined.

Instance Taint_flowsTo_HasBottom (a: Ensemble Prin.t):
  HasBottom _ (taint_flowsTo a) :=
{ bottom := Empty_set _ }.
Proof.
unfold taint_flowsTo. intros t p H. destruct H.
Defined.

Instance Taint_JoinPrelattice (a: Ensemble Prin.t) :
  JoinPreLattice (Ensemble Prin.t) (taint_flowsTo a) :=
{ join := Union _ }.
Proof.
* intros s1 s2. unfold taint_flowsTo.
  transitivity (join s1 s2).
  apply join_ub1.
  transitivity (join (join s1 s2) a); try reflexivity.
  apply join_ub1.
* intros s1 s2. unfold taint_flowsTo.
  transitivity (join s1 s2).
  apply join_ub2.
  transitivity (join (join s1 s2) a); try reflexivity.
  apply join_ub1.
* intros s s1 s2 H1 H2. unfold taint_flowsTo in *.
  transitivity (join s1 s2); try reflexivity.
  apply join_lub; assumption.
Defined.

Instance Taint_MeetPrelattice (a: Ensemble Prin.t) :
  MeetPreLattice (Ensemble Prin.t) (taint_flowsTo a) :=
{ meet := Intersection _ }.
Proof.
* intros s1 s2. unfold taint_flowsTo.
  transitivity s1.
  transitivity (meet s1 s2). reflexivity. apply meet_lb1.
  transitivity (@join _ (Included _) _ _ s1 a).
  apply join_ub1. reflexivity.
* intros s1 s2. unfold taint_flowsTo.
  transitivity s2.
  transitivity (meet s1 s2). reflexivity. apply meet_lb2.
  transitivity (@join _ (Included _) _ _ s2 a).
  apply join_ub1. reflexivity.
* intros s s1 s2 H1 H2. unfold taint_flowsTo in *.
  transitivity (meet (@join _ (Included _) _ _ s1 a)
                     (@join _ (Included _) _ _ s2 a)).
  apply meet_glb; assumption.
  apply Ensemble_distrib1.
Defined.

Instance UTaint : LabelAlgebra (Ensemble Prin.t) (Ensemble Prin.t) :=
{ Ldef := Empty_set _
; labLeq := Included _
; authLeq := Included _
; flowsTo := taint_flowsTo
}.
Proof.
* unfold taint_flowsTo; split; intros x y;
  simpl; rewrite Union_with_empty_set_Included1; trivial.
* auto.
* auto.
* intros a1 a2 H12 s s' H.
  unfold taint_flowsTo in *.
  transitivity (Union _ s' a1); auto.
  transitivity (@join _ (Included _) _ _ s' a1); try reflexivity.
  transitivity (@join _ (Included _) _ _ s' a2); try reflexivity.
  rewrite H12. reflexivity.
Defined.
