Require Import PreLattice.

Set Implicit Arguments.

Inductive WithBottom (A: Type) : Type :=
| WithBottom_intro : A -> WithBottom A
| WithBottom_bottom : WithBottom A.
Hint Constructors WithBottom.

Definition WithBottom_leq {A: Type} (leq: A -> A -> Prop) (x1 x2: WithBottom A): Prop :=
match x1, x2 with
| WithBottom_bottom, _ => True
| WithBottom_intro _, WithBottom_bottom => False
| WithBottom_intro a1, WithBottom_intro a2 => leq a1 a2
end.

Instance WithBottom_leq_PreOrder {A: Type} (leq: A -> A -> Prop) {H: PreOrder leq} :
  PreOrder (WithBottom_leq leq).
Proof.
constructor.
* intros [a |]; simpl. reflexivity. trivial.
* intros [a1 |] [a2 |] [a3 |]; simpl; auto; intros.
  transitivity a2; trivial.
Qed.

Instance WithBottom_HasTop (A: Type) (leq: A -> A -> Prop) {H: HasTop A leq}:
  HasTop (WithBottom A) (WithBottom_leq leq) :=
{ top := WithBottom_intro top }.
Proof.
intros [a |]; simpl; trivial. apply top_maximum.
Defined.

Instance WithBottom_HasBottom (A: Type) (leq: A -> A -> Prop):
  HasBottom (WithBottom A) (WithBottom_leq leq) :=
{ bottom := WithBottom_bottom A }.
Proof.
intros [a |]; simpl; trivial.
Defined.

Instance WithBottom_JoinPreLattice A (leq: A -> A -> Prop)
         {P: PreOrder leq} {J: JoinPreLattice A leq} :
  JoinPreLattice (WithBottom A) (WithBottom_leq leq) :=
{ join x1 x2 :=
    match x1, x2 with
      | WithBottom_bottom, x | x, WithBottom_bottom => x
      | WithBottom_intro a1, WithBottom_intro a2 => WithBottom_intro (join a1 a2)
    end
}.
Proof.
* intros [a1 |] [a2 |]; simpl; auto. apply join_ub1. reflexivity.
* intros [a1 |] [a2 |]; simpl; auto. apply join_ub2. reflexivity.
* intros [a |] [a1 |] [a2 |]; simpl; auto. apply join_lub.
Defined.

Instance WithBottom_MeetPreLattice A (leq: A -> A -> Prop)
         {P: PreOrder leq} {M: MeetPreLattice A leq} :
  MeetPreLattice (WithBottom A) (WithBottom_leq leq) :=
{ meet x1 x2 :=
    match x1, x2 with
      | WithBottom_bottom, _ | _, WithBottom_bottom => WithBottom_bottom A
      | WithBottom_intro a1, WithBottom_intro a2 => WithBottom_intro (meet a1 a2)
    end
}.
Proof.
* intros [a1 |] [a2 |]; simpl; auto. apply meet_lb1.
* intros [a1 |] [a2 |]; simpl; auto. apply meet_lb2.
* intros [a |] [a1 |] [a2 |]; simpl; auto. apply meet_glb.
Defined.
