Require Import PreLattice.

Set Implicit Arguments.

Inductive WithTop (A: Type) : Type :=
| WithTop_intro : A -> WithTop A
| WithTop_top : WithTop A.
Hint Constructors WithTop.

Definition WithTop_leq {A: Type} (leq: A -> A -> Prop) (x1 x2: WithTop A): Prop :=
match x1, x2 with
| _, WithTop_top => True
| WithTop_top, WithTop_intro _ => False
| WithTop_intro a1, WithTop_intro a2 => leq a1 a2
end.

Instance WithTop_leq_PreOrder {A: Type} (leq: A -> A -> Prop) {H: PreOrder leq} :
  PreOrder (WithTop_leq leq).
Proof.
constructor.
* intros [a |]; simpl. reflexivity. trivial.
* intros [a1 |] [a2 |] [a3 |]; simpl; auto; intros.
  transitivity a2; trivial.
Qed.

Instance WithTop_HasTop (A: Type) (leq: A -> A -> Prop):
  HasTop (WithTop A) (WithTop_leq leq) :=
{ top := WithTop_top A }.
Proof.
intros [a |]; simpl; trivial.
Defined.

Instance WithTop_HasBottom (A: Type) (leq: A -> A -> Prop) {H: HasBottom A leq}:
  HasBottom (WithTop A) (WithTop_leq leq) :=
{ bottom := WithTop_intro bottom }.
Proof.
intros [a |]; simpl; trivial. apply bottom_minimum.
Defined.

Instance WithTop_JoinPreLattice A (leq: A -> A -> Prop)
         {P: PreOrder leq} {J: JoinPreLattice A leq} :
  JoinPreLattice (WithTop A) (WithTop_leq leq) :=
{ join x1 x2 :=
    match x1, x2 with
      | WithTop_top, _ | _, WithTop_top => WithTop_top A
      | WithTop_intro a1, WithTop_intro a2 => WithTop_intro (join a1 a2)
    end
}.
Proof.
* intros [a1 |] [a2 |]; simpl; auto. apply join_ub1.
* intros [a1 |] [a2 |]; simpl; auto. apply join_ub2.
* intros [a |] [a1 |] [a2 |]; simpl; auto. apply join_lub.
Defined.

Instance WithTop_MeetPreLattice A (leq: A -> A -> Prop)
         {P: PreOrder leq} {M: MeetPreLattice A leq} :
  MeetPreLattice (WithTop A) (WithTop_leq leq) :=
{ meet x1 x2 :=
    match x1, x2 with
      | WithTop_top, x | x, WithTop_top => x
      | WithTop_intro a1, WithTop_intro a2 => WithTop_intro (meet a1 a2)
    end
}.
Proof.
* intros [a1 |] [a2 |]; simpl; auto. apply meet_lb1. reflexivity.
* intros [a1 |] [a2 |]; simpl; auto. apply meet_lb2. reflexivity.
* intros [a |] [a1 |] [a2 |]; simpl; auto. apply meet_glb.
Defined.
