Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export TableCheck.

(** * Specialization of table checking on label algebras and embeddings. *)

Definition LAs := list packedLA.

Definition embed_table
           {F: Type -> Type -> Type}
           {M: LAMapApply F}
           (sem: SEM F)
           (elts: LAs)
           (answers: list (list Answer)) : Prop :=
  property_square_table (embeds_in sem) elts answers.

(** ** Special notation to put label algebras into a list. *)
Notation "[| x ; .. ; y |]" :=
  (cons (packLA x) .. (cons (packLA y) nil) ..) : table.

(** ** Tactics *)
Ltac solve_embed :=
  try solve [auto];
  try abstract solve [apply LAEmbedding_injection; auto];
  try abstract solve [eapply embeds_bottom1_eval_boolean; auto];
  try abstract solve [eapply embeds_bottom1_eval_boolean_contra; auto];
  try abstract solve [eapply LAEmbedding_injection_contra
                      with (sem := boolean_sem); eauto];
  try abstract solve [eapply LAEmbedding_injection_contra
                      with (sem := eval_sem); eauto].

Ltac solve_embed_table := split_property_table; solve_embed.

Ltac exploit_eq :=
  match goal with
    | [ Hm: forall l1 l2, l1 =L l2 <-> Lmap ?m l1 =L Lmap ?m l2,
        H1: Lmap ?m ?X = ?Y,
        H2: Lmap ?m ?Z = ?Y |- _ ] =>
      try solve
        [ let H := fresh in
          assert (Lmap m X =L Lmap m Z) as H by (rewrite H1; rewrite H2; reflexivity);
            apply Hm in H;
            compute in H;
            auto
        ]
    | [ Hm: forall a l1 l2,
              l1 ⊑[ a]l2 <-> Lmap ?m l1 ⊑[Amap ?m a] Lmap ?m l2,
        H1: Lmap ?m ?X = ?X',
        H2: Lmap ?m ?Y = ?Y' |- _ ] =>
      solve
      [ let H := fresh in
        assert (X' ⊑[Amap m bottom] Y')
          as H by
              (rewrite <- H1; rewrite <- H2;
               apply Hm; compute; auto);
        compute in H; auto
      ]
    | [ Hm: forall a l1 l2,
              l1 ⊑[ a]l2 <-> Lmap ?m l1 ⊑[Amap ?m a] Lmap ?m l2,
        H1: Lmap ?m ?X = ?X',
        H2: Lmap ?m ?Y = ?Y' |- _ ] =>
      solve
      [ let H := fresh in
        assert (X' ⊑[Amap m bottom] Y')
          as H by (compute; auto);
        rewrite <- H1 in H;
        rewrite <- H2 in H;
        apply Hm in H; compute in H; auto
      ]    | _ => idtac
  end.
