Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Three.
Require Export Four.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma injects_three_four: Three ↪ Four.
Proof.
exists ({| Lmap l := match l with
                       | Bottom3 => Bottom4
                       | Middle3 => LMiddle4
                       | Top3 => Top4
                     end
         ; Amap a := bottom |}).
rewrite LAinjection_iff. unfold LAinjection_prop.
splits; try solve [compute; auto].
* intros l1 l2. destruct l1; destruct l2; compute; auto.
* intros l1 l2. destruct l1; destruct l2; compute; auto.
* intros a l1 l2. destruct a; destruct l1; destruct l2; compute; auto.
Qed.
Hint Resolve injects_three_four.

Lemma not_boolean_embeds_four_three: ~(Four ==>[boolean_sem] Three).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
case_eq (Lmap m Bottom4); intro Heq1;
case_eq (Lmap m LMiddle4); intro Heq2; exploit_eq;
case_eq (Lmap m UMiddle4); intro Heq3; exploit_eq;
case_eq (Lmap m Top4); intro Heq4; exploit_eq.
Qed.
Hint Resolve not_boolean_embeds_four_three.

(** * Summary *)
Open Scope table.

Theorem injects_three_four_summary :
  property_square_table injects_in
     [| Three; Four |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_three_four_summary :
  embed_table boolean_sem
     [| Three; Four |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_three_four_summary :
  embed_table eval_sem
     [| Three; Four |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
