Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Three.
Require Export FourD.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma injects_three_fourD: Three ↪ FourD.
Proof.
exists ({| Lmap l := match l with
                       | Bottom3 => Bottom4D
                       | Middle3 => LMiddle4D
                       | Top3 => Top4D
                     end
         ; Amap a := bottom |}).
rewrite LAinjection_iff. unfold LAinjection_prop.
splits; try solve [compute; auto].
* intros l1 l2. destruct l1; destruct l2; compute; auto.
* intros l1 l2. destruct l1; destruct l2; compute; auto.
* intros a l1 l2. destruct a; destruct l1; destruct l2; compute; auto.
Qed.
Hint Resolve injects_three_fourD.

Lemma not_boolean_embeds_fourD_three: ~(FourD ==>[boolean_sem] Three).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
case_eq (Lmap m Bottom4D); intro Heq1;
case_eq (Lmap m LMiddle4D); intro Heq2; exploit_eq;
case_eq (Lmap m RMiddle4D); intro Heq3; exploit_eq;
case_eq (Lmap m Top4D); intro Heq4; exploit_eq.
Qed.
Hint Resolve not_boolean_embeds_fourD_three.

(** * Summary *)
Open Scope table.

Theorem injects_three_fourD_summary :
  property_square_table injects_in
     [| Three; FourD |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_three_fourD_summary :
  embed_table boolean_sem
     [| Three; FourD |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_three_fourD_summary :
  embed_table eval_sem
     [| Three; FourD |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
