Require Import MyTactics.
Require Import LabelAlgebra.
Require Import LabelSemantics.
Require Import BooleanSemantics.
Require Import EvaluationSemantics.
Require Export Two.
Require Export Three.
Require Import EmbedTableCheck.

(** * Intermediate lemmas. *)
Lemma injects_two_three: Two ↪ Three.
Proof.
exists ({| Lmap l := match l with
                       | Bottom2 => Bottom3
                       | Top2 => Top3
                     end
         ; Amap a := bottom |}).
rewrite LAinjection_iff. unfold LAinjection_prop.
splits; try solve [compute; auto].
* intros l1 l2. destruct l1; destruct l2; compute; auto.
* intros l1 l2. destruct l1; destruct l2; compute; auto.
* intros a l1 l2. destruct a; destruct l1; destruct l2; compute; auto.
Qed.
Hint Resolve injects_two_three.

Lemma not_boolean_embeds_three_two: ~(Three ==>[boolean_sem] Two).
Proof.
rewrite boolean_embeds_iff.
intros [m [Hm1 Hm2]].
case_eq (Lmap m Bottom3); intro Heq1;
case_eq (Lmap m Middle3); intro Heq2; exploit_eq;
case_eq (Lmap m Top3); intro Heq3; exploit_eq.
Qed.
Hint Resolve not_boolean_embeds_three_two.

(** * Summary *)
Open Scope table.

Theorem injects_two_three_summary :
  property_square_table injects_in
     [| Two; Three |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem boolean_embed_two_three_summary :
  embed_table boolean_sem
     [| Two; Three |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Theorem eval_embed_two_three_summary :
  embed_table eval_sem
     [| Two; Three |]
     [ [ ✓; ✓ ]
     ; [ ×; ✓ ]
     ].
Proof.
solve_embed_table.
Qed.

Close Scope table.
